/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <QUANTA/QUANTA.hxx>

/* This is a callback to handle any incoming data on a key. This method follows
the traditional style of firing callbacks to handle events. See below for the
new way.
*/
void incomingData(char* path, char* key, void* userData)
{
	QUANTAdb_client_c *client = (QUANTAdb_client_c*) userData;
	int size;
	char* data = client->get(path, key, &size);

	printf("CALLBACK FIRED    %s %s %s\n", path, key, data);
	if (data)
	  delete[] data;
}

/* This is contemporary way of handling incoming data on a key. This is an observer
class that is set up to listen for updates to the db client. This technique
follows the observer/observable (or observer/subject) design pattern- as used in
Java.
*/
class myObserver_c : public QUANTAmisc_observer_c 
{
	// You overload the update function to handle the incoming data.
	void update(QUANTAmisc_subject_c *client) {

		QUANTAdb_client_c *theClient = (QUANTAdb_client_c*) client;

		int size;
		char *path = theClient->getNotifyPathname();
		char *key = theClient->getNotifyKeyname();

		char* data = theClient->get(path, key, &size);

		printf("OBSERVER NOTIFIED %s %s %s\n", path, key, data);
		if (data)
		  delete[] data;
	}

};


int
main(int argc, char** argv)
{
	QUANTAinit();

	myObserver_c myObserver;

	QUANTAdb_client_c client;
	int status = 0;
	if (argc > 1)
		status = client.init(argv[1]);
	else
		status = client.init(NULL);
	if (status != QUANTAdb_client_c::OK) {
	  printf("ERROR initializing QUANTAdb_client_c! (code: %i)\n", status);
	}

	/* There are 2 ways to get a notification on when new data is arriving.
	   Register a callback using the trigger() method.
	   Or attach an observer to the client using the attach() method.
	*/	  
	if (argc == 3  && client.initSendStats(argv[2]) != QUANTAnet_perfMonitor_c::OK) {
		printf("Cannot connect to %s\n", argv[2]);
		exit(1);
	}
    if (argc <3) 

	client.trigger(incomingData, (void*) &client);
	client.attach(&myObserver);

	client.fetch("path1","key1");

	client.put("path1","key1", "hello", 6, QUANTAdb_client_c::UNRELIABLE_CONNECTION );

	printf("DONE PUT\n");
    client.commit("path1", "key1");
	client.remoteCommit("path1","key1");

	while(1) {
		int status = client.process();
        //Uncomment this for an instantaneous display of performance statistics
        //client.showStats("dbClient_monitoring","Some_local_DB_data");
     
        //Uncomment this for sending the statistics to a remote perfDaemon (the initSendStats must be done before this)
        /*if (client.sendStats("CAVE_avatar_on_clark", "New_avatar_entered") != QUANTAnet_perfMonitor_c::OK)
            printf("\nsendStats failed!\n");*/
		if (status == QUANTAdb_client_c::CONNECTION_TERMINATED) {
			client.reconnect();
		}
	}
	return 1;
}









