/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <QUANTA/QUANTA.hxx>
#include <QUANTA/QUANTAdb_sharedState_c.hxx>

/** Simple shared state class.
    This class encapsulates a very simple shared state. The state that will
    be shared is nothing more than a simple integer value. However, by
    using the shared state class, this class encapsulates the process of
    sharing its representation across the network. */
class MyState : public QUANTAdb_sharedState_c
{

 public:
  /** Constructor with parameters.
      The constructor takes the parameters required by its parent class.
      These values are passed along during initialization. They must be
      created before the MyState instance is created.
      @param client pointer to the QUANTAdb_client_c sharing the state.
      @param path character string for the path to this state.
      @param key character string for the state's key.
  */
  MyState(QUANTAdb_client_c *client, char *path, char *key)
    : QUANTAdb_sharedState_c(client, path, key) {}

  /// Virtual destructor.
  virtual ~MyState() {}

  /// Returns the state value.
  int getValue() { return _value; }

  /// Sets the state value.
  void setValue(const int& newValue) {
    // Use class method for setting the value.
    this->setState(newValue);
    // If sharing is enabled, update the state across the network.
    if (this->isSharingEnabled())
      this->packAndSendState();
  }

 protected:
  /** Packs the current state and sends it across the network.
      The subclass must implement how it should pack its own data for
      transmission across the network. Using the QUANTAnet_datapack_c
      instance from the parent class, the state is packed into a
      buffer before calling QUANTAdb_client_c::put().
  */
  virtual int packAndSendState() {
    // Make a buffer into which the state will be packed.
    char *buffer = new char[sizeof(int)];
    // Use a data packer to store the state.
    _packer.initPack( buffer, sizeof(int) );
    _packer.packInt( _value );
    // Get the filled buffer size.
    int realsize = _packer.getBufferFilledSize();
    // Place the updated state into the database.
    int result = _dbclient->put( _path, _key, buffer, realsize );
    if (result == QUANTAdb_client_c::OK)
      printf("MyState = %i\n", _value);
    // Cleanup the buffer memory.
    delete[] buffer;
    return result;
  }

  /** Unpacks the state from a buffer.
      When data is received from the network, the subclass must implement
      how that buffer will be unpacked. As with packing, the method uses
      the parent class's QUANTAnet_datapack_c instance.
  */
  virtual int unpackState(char* data, const int& size) {
    // If the size is 0, then there is no data.
    if (size <= 0)
      return 0;
    // Otherwise, initialize the data packer to unpack the state.
    _packer.initUnpack(data, size);
    int newVal;
    if (_packer.unpackInt(&newVal) == QUANTAnet_datapack_c::OK) {
      printf("MyState update = %i\n", newVal);
      this->setState(newVal);
      return QUANTAnet_datapack_c::OK;
    } else {
      printf("MyState ERROR: unable to unpack value.\n");
      return QUANTAnet_datapack_c::FAILED;
    }
  }

 private:
  /// Privately sets the state while hiding its implementation.
  void setState(const int& newValue) { _value = newValue; }

 private:
  /// Private state value.
  int _value;
};

int
main(int argc, char** argv)
{
  // Check the number of arguments.
  if (argc < 2) {
    printf("Usage: sharedState server [performance-server]\n");
    return 1;
  }

  bool initOk = false;
  // Initialize QUANTA and the database client.
  QUANTAinit();
  QUANTAdb_client_c* client = new QUANTAdb_client_c;
  if (client->init(argv[1]) == QUANTAdb_client_c::OK) {
    initOk = true;
  } else {
    printf("ERROR: Cannot connect to server %s\n", argv[1]);
  }
  if (argc == 3 &&
      client->initSendStats(argv[2]) != QUANTAnet_perfMonitor_c::OK) {
    printf("ERROR: Cannot connect to performance server %s\n", argv[2]);
    initOk = false;
  }

  // If initialization failed, clean up memory and exit.
  if (!initOk) {
    printf("\nUnable to properly initialize the database. Exiting...\n");
    delete client;
    return 1;
  }

  // Refresh the state from the server.
  MyState* myState = new MyState(client, "myPath", "myKey");
  // Update the value by refreshing with any data which may be on the network.
  myState->refresh();
  // Activate sharing.
  myState->enableSharing();

  int status = 0;
  bool doSetValue = false;
  while (1) {
    // Get the client process status.
    status = client->process();

    // Check to see if the connection has terminated.
    if (status == QUANTAdb_client_c::CONNECTION_TERMINATED) {
      client->reconnect();
    }

    // Randomly decide whether to set a new value for the state based
    // on the time.
    srand(time(0));
    doSetValue = rand() < (RAND_MAX * 0.10f);
    if (doSetValue) {
      // Set a randome value for the state.
      // We're using the standard method to set the value for the state
      // instance, and the class implementation will automatically share
      // the state across the network by updating the database.
      int newState = rand();
      myState->setValue(newState);
    }

    QUANTAsleep(1);
  }

  // Cleanup memory and exit.
  delete myState;
  delete client;
  return 0;
}
