/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include "QUANTAmisc_list_c.hxx"
#include "QUANTAmisc_stack_c.hxx"
#include "QUANTAmisc_queue_c.hxx"


//
// test suite for quanta adt classes
//
// only print to the terminal if an error occurs.
// a message stating the error count is dumped to the terminal for each test
//

class Integer;

static int error_count = 0;
static void testList(void);
static void testStack(void);
static void testQueue(void);




/**
* int wrapper class
**/
class Integer
{
	public:
		Integer(void) { _i = 0; }
		Integer(int i) { _i = i; }
		~Integer(void) {}
		
		int getInt(void) { return _i; }
		void setInt(int i) { _i = i; }

		bool operator ==(const Integer &i)
		{
			if(_i == i._i)
				return true;
			
			return false;
		}
		
		void toString(void)
		{
			printf("%d", _i);
		}

	private:
		int _i;
};


static void testList(void)
{
	error_count = 0;
	QUANTAmisc_list_c<Integer> list;

	if(list.size() != 0)
	{
		fprintf(stderr, "List Error: size should be zero; size = %d\n", list.size());
		error_count++;
	}

	if(!list.isEmpty())
	{
		printf("list Error: isEmpty() - list is not empty when it should contain zero nodes\n");
		error_count++;
	}

	if(list.isCurrent())
	{
		printf("list Error: isCurrent() - list is not empty when it should contain zero nodes\n");
		error_count++;
	}

	list.addAfter(Integer(55));
	list.addAfter(Integer(13));
	list.addAfter(Integer(75));
	list.removeCurrent();
	try {
		if(list.getCurrent().getInt() != 13)
		{
			fprintf(stderr, "List Error: removeCurrent(); current = %d\n", list.getCurrent().getInt());
			error_count++;
		}
	}	catch(QUANTAmisc_noCurrentException_c ex) {
		fprintf(stderr, "Exception caught.  %s\n", ex.getMessage());
	}

	if(list.size() != 2)
	{
		fprintf(stderr, "List Error: size should be two; size = %d\n", list.size());
		error_count++;
	}

	if(list.isEmpty())
	{
		printf("list Error: isEmpty() - list is empty when it should contain two nodes\n");
		error_count++;
	}

	if(!list.isCurrent())
	{
		printf("list Error: isCurrent() - list is not empty when it should contain two nodes\n");
		error_count++;
	}


	try {
		list.clear();
	} catch(QUANTAmisc_exception_c ex) {}

	if(list.size() != 0)
	{
		fprintf(stderr, "List Error: size should be zero; size = %d\n", list.size());
		error_count++;
	}

	list.addAfter(Integer(333));
	list.addBefore(Integer(5555));
	list.addBefore(Integer(123));
	list.addBefore(Integer(11));
	list.addBefore(Integer(50));
/*	
	// print out all the nodes in the list in sequencial order starting
	// with the first and ending with the last node.
	list.first();
	while(!list.isEmpty())
	{
		try {
				fprintf(stderr, "current = %d\n", list.getCurrent().getInt());
				list.next();
		}	catch(QUANTAmisc_noCurrentException_c ex) {
			break; // exit the loop if there are no more nodes in the list
		}
	}
*/
	try {
		if(list.getCurrent().getInt() != 50)
		{
			fprintf(stderr, "List Error: first(); current = %d\n", list.getCurrent().getInt());
			error_count++;
		}
	}	catch(QUANTAmisc_noCurrentException_c ex) {
		fprintf(stderr, "Exception caught.  %s\n", ex.getMessage());
	}


	try {
		list.next();
		list.next();
		list.previous();
	} catch (QUANTAmisc_noCurrentException_c ex) {}
	if(list.size() != 5)
	{
		fprintf(stderr, "List Error: size should be five; size = %d\n", list.size());
		error_count++;
	}

	try {
		if(list.getCurrent().getInt() != 11)
		{
			fprintf(stderr, "List Error: getCurrent(); current = %d\n", list.getCurrent().getInt());
			error_count++;
		}
	}	catch(QUANTAmisc_noCurrentException_c ex) {
		fprintf(stderr, "Exception caught.  %s\n", ex.getMessage());
	}
	
	list.last();
	list.removeCurrent();
	try {
		if(list.getCurrent().getInt() != 5555)
		{
			fprintf(stderr, "List Error: last()/removeCurrent(); current = %d\n", list.getCurrent().getInt());
			error_count++;
		}
	}	catch(QUANTAmisc_noCurrentException_c ex) {
		fprintf(stderr, "Exception caught.  %s\n", ex.getMessage());
	}
	
	if(!list.contains(123))
	{
		fprintf(stderr, "List Error: contains(); node not found\n");
		error_count++;
	}
	

	printf("List test complete. %d errors occured.\n", error_count);
}


static void testStack(void)
{
	error_count = 0;
	int temp_int = 0, id = 0;
	QUANTAmisc_stack_c<Integer> stack;

	if(!stack.isEmpty())
	{
		printf("Stack Error: isEmpty() - stack is not empty when it should contain zero nodes\n");
		error_count++;
	}
	
	id = 2;
	stack.push(Integer(id));
	stack.push(Integer(id += 2));
	stack.push(Integer(id += 2));
	stack.push(Integer(id += 2));

	if(stack.isEmpty())
	{
		printf("Stack Error: isEmpty() - stack is empty when it should contain nodes\n");
		error_count++;
	}

	try {
		stack.clear();
	} catch(QUANTAmisc_emptyStackException_c ex) {
		printf("Stack Error: clear() - stack is empty when it should contain nodes\n");
		error_count++;
	}

	id = 1;
	stack.push(Integer(id));
	stack.push(Integer(id += 2));
	stack.push(Integer(id += 2));
	stack.push(Integer(id += 2));

	if(stack.size() != 4)
	{
		fprintf(stderr, "Stack Error: size should be four; size = %d\n", stack.size());
		error_count++;
	}

	try {
		id = (stack.size() * 2) - 1; // start at an odd number since the ints added started at 1
		while(!stack.isEmpty())
		{
			if((temp_int = stack.peek().getInt()) != id)
			{
				fprintf(stderr, "Stack Error: peek() top node = %d\n", temp_int);
				error_count++;
			}

			if((temp_int = stack.pop().getInt()) != id)
			{
				fprintf(stderr, "Stack Error: pop() top node = %d\n", temp_int);
				error_count++;
			}
			
//			printf("node = %d\n", temp_int);

			id -= 2;
		}

		stack.clear();

		id = 2;
		stack.push(Integer(id));
		stack.push(Integer(id += 2));
		stack.push(Integer(id += 2));
		stack.push(Integer(id += 2));
		stack.push(Integer(id += 2));
		stack.push(Integer(id += 2));

		if(stack.size() != 6)
		{
			fprintf(stderr, "Stack Error: size should be six; size = %d\n", stack.size());
			error_count++;
		}

		id = (stack.size() * 2);
		while(!stack.isEmpty())
		{
			if((temp_int = stack.peek().getInt()) != id)
			{
				fprintf(stderr, "Stack Error: peek() top node = %d\n", temp_int);
				error_count++;
			}

			if((temp_int = stack.pop().getInt()) != id)
			{
				fprintf(stderr, "Stack Error: pop() top node = %d\n", temp_int);
				error_count++;
			}

//			printf("node = %d\n", temp_int);			

			id -= 2;
		}

	} catch(QUANTAmisc_emptyStackException_c ex) {}

	if(!stack.isEmpty())
	{
		printf("Stack Error: isEmpty() - stack contains nodes when it should be empty\n");
		error_count++;
	}

	if(stack.size() != 0)
	{
		fprintf(stderr, "Stack Error: size should be zero; size = %d\n", stack.size());
		error_count++;
	}


	printf("Stack test complete. %d errors occured.\n", error_count);
}


static void testQueue(void)
{
	error_count = 0;
	int temp_int = 0, id = 0;
	QUANTAmisc_queue_c<Integer> queue;

	if(!queue.isEmpty())
	{
		printf("Queue Error: isEmpty() - queue is not empty when it should contain zero nodes\n");
		error_count++;
	}

	id = 1;
	queue.enqueue(Integer(id));
	queue.enqueue(Integer(id += 2));
	queue.enqueue(Integer(id += 2));
	queue.enqueue(Integer(id += 2));
	queue.enqueue(Integer(id += 2));
	queue.enqueue(Integer(id += 2));
	queue.enqueue(Integer(id += 2));

	if(queue.isEmpty())
	{
		printf("Queue Error: isEmpty() - queue is empty when it should contain nodes\n");
		error_count++;
	}

	if(queue.size() != 7)
	{
		fprintf(stderr, "Queue Error: size should be seven; size = %d\n", queue.size());
		error_count++;
	}

	try {
		id = 1;
		while(!queue.isEmpty())
		{
			if((temp_int = queue.dequeue().getInt()) != id)
			{
				fprintf(stderr, "Queue Error: dequeue() front node = %d\n", temp_int);
				error_count++;
			}
//			printf("node = %d\n", temp_int);
			
			id += 2;
		}
	} catch(QUANTAmisc_emptyQueueException_c ex) { 
		printf("Exception caught: %s\n", ex.getMessage());
	}

	if(!queue.isEmpty())
	{
		printf("Queue Error: isEmpty() - queue is not empty when it should contain zero nodes\n");
		error_count++;
	}

	if(queue.size() != 0)
	{
		fprintf(stderr, "Queue Error: size should be zero; size = %d\n", queue.size());
		error_count++;
	}


	id = 9;
	queue.enqueue(Integer(id));
	queue.enqueue(Integer(id -= 3));
	queue.enqueue(Integer(id -= 3));
	queue.clear();

	if(!queue.isEmpty())
	{
		printf("Queue Error: isEmpty() - queue is not empty when it should contain zero nodes\n");
		error_count++;
	}

	if(queue.size() != 0)
	{
		fprintf(stderr, "Queue Error: size should be zero; size = %d\n", queue.size());
		error_count++;
	}


	id = 2;
	queue.enqueue(Integer(id));
	queue.enqueue(Integer(id += 2));
	queue.enqueue(Integer(id += 2));
	queue.enqueue(Integer(id += 2));
	queue.enqueue(Integer(id += 2));
	queue.enqueue(Integer(id += 2));
	queue.enqueue(Integer(id += 2));
	queue.enqueue(Integer(id += 2));
	queue.enqueue(Integer(id += 2));
	try {
		id = 2;
		while(!queue.isEmpty())
		{
			if((temp_int = queue.dequeue().getInt()) != id)
			{
				fprintf(stderr, "Queue Error: dequeue() front node = %d\n", temp_int);
				error_count++;
			}
//			printf("node = %d\n", temp_int);
			
			id += 2;
		}
	} catch(QUANTAmisc_emptyQueueException_c ex) { 
		printf("Exception caught: %s\n", ex.getMessage());
	}

	if(!queue.isEmpty())
	{
		printf("Queue Error: isEmpty() - queue is not empty when it should contain zero nodes\n");
		error_count++;
	}

	if(queue.size() != 0)
	{
		fprintf(stderr, "Queue Error: size should be zero; size = %d\n", queue.size());
		error_count++;
	}


	printf("Queue test complete. %d errors occured.\n", error_count);
}


int main()
{
	testList();
	testStack();
	testQueue();

	return 1;
}
