/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

/*******************************************************************************
 * QUANTA Condition Variable Class Demo
 *
 * NOTE: adapted from "Thread Time The Multithreaded Programming Guide"
 *******************************************************************************/
/*
 Class for thread condition variables.
 Condition variables are used in conjunction with
 mutexes to provide a way for threads to wait until
 a condition occurs.

 An example of waiting on a signal is:
 // Lock your mutex that is protecting someState.
 myMutex->lock();

 // Watch for your desired state to occur.
 while(someState != reached)
 {

	// Wait for a signal.
	myCondition->wait(myMutex);

	.... got the condition and the lock so now continue ....
 }

 myMutex->unlock();


 An example of sending the signal is:
 // Lock your mutex that is protecting someState.
 myMutex->lock();

 // Signal that the state has been reached.
 if (someState == reached) myCondition->signal();

 // Unlock your mutex so that the waiting thread can continue.
 myMutex->unlock();
*/

#include <QUANTA/QUANTA.hxx>

#define TERMINATE_COUNT 5

QUANTAts_condition_c condition;
QUANTAts_mutex_c mutex;
int count = 0;

// send a signal when a specific event occurs or condition is reached
void* producer(void* data)
{

 // Lock the mutex that is protecting count.
 mutex.lock();

 // incremental count to TERMINATE_COUNT
 // basically, stall the wake up call until some event has occured
 // in this case the wake up is not called until TERMINATE_COUNT is reached
// count = 0;
 while(count != TERMINATE_COUNT)
 {
    fprintf(stdout, "producer: count = %d\n", count);
    QUANTAsleep(1);
    count++; 
 }
 
 // Signal that the state has been reached.
 if(count == TERMINATE_COUNT) // this if/else shows 
    condition.signal();      // how an actual application might work when the 
 else                         // condition was meet in another function
 {
    fprintf(stderr, "producer: signal failed\n");
    exit(0);
 }

 // Unlock your mutex so that the waiting thread can continue.
 mutex.unlock();

 return NULL;
}


//  wait for a signal to allow this
void* consumer(void* data)
{
 // Lock the mutex that is protecting count.
 mutex.lock();

 // Watch for your desired state to occur.
 while(count < TERMINATE_COUNT)
 {
    // Wait for a signal.
    condition.wait(&mutex);
 }
 
 // ...got the condition and the lock so now continue...
 fprintf(stdout, "consumer: event occurred/condition meet... so continuing\n");
 count = 0;
 while(count != TERMINATE_COUNT)
 {
    fprintf(stdout, "consumer: count = %d\n", count);
    QUANTAsleep(1);
    count++; 
 }

 mutex.unlock();
 return NULL;
}


int main(int argc, char *argv[])
{
 QUANTAinit();
 QUANTAts_thread_c thread1, thread2;


 thread1.create(producer, (void*)NULL);
 thread2.create(consumer, (void*)NULL);


 // halt main process so all threads can finish
 thread1.join();
 thread2.join();
 
 QUANTAexit();
 return 1;
}
