/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <QUANTA/QUANTAdb_c.hxx>
#include <QUANTA/QUANTAnet_datapack_c.hxx>
#include <QUANTA/QUANTAnet_tcpReflector_c.hxx>
#include <QUANTA/QUANTAnet_udpReflector_c.hxx>
#include <QUANTA/keytool.hxx>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

const char QUANTAdb_server_c::PUT_MESG = 'p';
const char QUANTAdb_server_c::FETCH_MESG = 'f';
const char QUANTAdb_server_c::BLOCKING_FETCH_MESG = 'b';
const char QUANTAdb_server_c::COMMIT_MESG = 'c';
const int QUANTAdb_server_c::DEFAULT_TCP_REFLECTOR_PORT = 9500;
const int QUANTAdb_server_c::DEFAULT_UDP_REFLECTOR_PORT = 9550;
const int QUANTAdb_server_c::DEFAULT_MAX_CLIENTS = 64;
const int QUANTAdb_server_c::OK = 1;
const int QUANTAdb_server_c::FAILED = 0;
const int QUANTAdb_server_c::MEM_ALLOC_ERR = 3;
const int QUANTAdb_server_c::TOO_MANY_CLIENTS = 4;
const int QUANTAdb_server_c::KEY_DOES_NOT_EXIST =-1;

const int QUANTAdb_client_c::FAILED = 0;
const int QUANTAdb_client_c::OK = 1;
const int QUANTAdb_client_c::MEM_ALLOC_ERR = 2;
const int QUANTAdb_client_c::CONNECTION_TERMINATED = 3;
const int QUANTAdb_client_c::IGNORED = 4;
const int QUANTAdb_client_c::UNRELIABLE_CONNECTION = 5;
const int QUANTAdb_client_c::RELIABLE_CONNECTION = 6;
const int QUANTAdb_client_c::MAX_UDP_PACKET_SIZE = 1000;
const int QUANTAdb_client_c::KEY_DOES_NOT_EXIST =-1;


/*
This is the replacement for the key database that was a central part
of the earlier generation of QUANTA. In QUANTA this is no longer the case.
Here it is significantly simplified with fewer capabilities.

One of the simplifications is that it follows only a client server model and
all connections are reliable TCP connections.

Clients instantiate a QUANTAdb_client_c object to communicate with the
server.  The server is built from 2 main components the keyTool
database and the new
QUANTAnet_tcpReflector_c. The client is also built from the keyTool
database however it uses the QUANTAnet_tcpReflectorClient_c to speak
to the server.

Basically the server sits around and waits for incoming connections. The
server also listens for incoming messages from these connections. Some messages it
will reflect to all other connected clients. Some messages it will absorb to do
some processing.

Currently the messages that the server recognizes are:

PUT, FETCH, COMMIT

PUT
---

This message will tell the server to store a piece of data in its database
given the pathname, keyname and the data. When a server receives this message
it first stores it in its database and then forwards this same message to
all other connected clients so that the same database update may occur at
each respective client.

FORMAT:
p+pathsize+pathname+keysize+keyname+datasize+dataname
pathsize=int
pathname=char* array of chars
keysize=int
keyname=char*
datasize=int
dataname=char*

FETCH
-----

This message will tell the server to send the data stored in the server's
database at a specific pathname and keyname, back to the client.
This is an asynchronous operation meaning once the client has issued the fetch
command, it will continue doing its own processing.

Clients also recognize messages. Currently the only one the client recognizes
is the PUT message which follows the same format as the PUT message above.
On receipt of a PUT message the client will store new incoming data into its
own database.

FORMAT:
f+pathsize+pathname+keysize+keyname
pathsize=int
pathname=char*
keysize=int
keyname=char*

BLOCKING FETCH
--------------

This is the same as the fetch message, except that the client blocks till a reply from the server comes back

Clients also recognize messages. Currently the only one the client recognizes
is the PUT message which follows the same format as the PUT message above.
On receipt of a PUT message the client will store new incoming data into its
own database.

FORMAT:
fetch:
f+pathsize+pathname+keysize+keyname 
block fetch:
b+pathsize+pathname+keysize+keyname

pathsize=int
pathname=char*
keysize=int
keyname=char*


COMMIT
------

This tells the server to commit a specific key to the database.

FORMAT:
c+pathsize+pathname+keysize+keyname
pathsize=int
pathname=char*
keysize=int
keyname=char*

--------------------------------------------

Future things to add include:
DONE - if no connection exists or if connection broken keys should continue to work
DONE - autoreconnect
- lock local and remote
- smarter server so that it will deallocate memory that's not needed

*/

QUANTAdb_server_c::~QUANTAdb_server_c()
{
	if (keyToolManager) delete keyToolManager;
	if (server) delete server;
    if (udpreflector) delete udpreflector;
}

int QUANTAdb_server_c :: init(char *defaultDBName, int inPort, int maxCli, int udpReflectorPort)
{
	keyToolManager = new keyToolManager_c(defaultDBName);
	if (keyToolManager == NULL) return QUANTAdb_server_c::MEM_ALLOC_ERR;

	incomingPort = inPort;
	server = new QUANTAnet_tcpReflector_c();
	if (server == NULL) {
		delete keyToolManager;
		keyToolManager = NULL;
		return QUANTAdb_server_c::MEM_ALLOC_ERR;
	}
	int status = server->init(inPort, maxCli);
	if (status != QUANTAnet_tcpReflector_c::OK) {
		return QUANTAdb_server_c::FAILED;
	}
	server->interceptNewConnection(newConnectionHandler, this);
	server->intercept(newDataHandler, this);
    
    udpreflector = new QUANTAnet_udpReflector_c;
    if (udpreflector == NULL) 
        return QUANTAdb_server_c::MEM_ALLOC_ERR;
    udpreflector->setIncomingPort(udpReflectorPort);
    udpreflector->init();
    udpreflector->enableInstrumentation();
    udpreflector->intercept(udpDataHandler, this);
	return QUANTAdb_server_c::OK;
}

int QUANTAdb_server_c :: process()
{
	int status;

	status = server->checkForNewClients();
	if (status == QUANTAnet_tcpReflector_c::TOO_MANY_CLIENTS)
		return QUANTAdb_server_c::TOO_MANY_CLIENTS;
	
	status = server->process();
	if (status != QUANTAnet_tcpReflector_c::OK) return QUANTAdb_server_c::FAILED;
	// JASON - UDPreflector does not need to call this anymore
	// udpreflector->process();
	return QUANTAdb_server_c::OK;
}

void QUANTAdb_server_c :: newConnectionHandler(QUANTAnet_tcpReflectorClient_c *newClient, void* userData)
{
	//QUANTAdb_server_c *server = (QUANTAdb_server_c*) userData;
}

int QUANTAdb_server_c :: newDataHandler(QUANTAnet_tcpReflectorClient_c *newClient, char** buffer, int *bufferSize, void *userData)
{
	QUANTAdb_server_c *server = (QUANTAdb_server_c*) userData;

	// First byte of buffer is command code.
	char* buf = *buffer;

	QUANTAnet_datapack_c data;
	data.initUnpack(buf, *bufferSize);

	char command;
	data.unpackChar(&command);

	switch(command) {

		// QUANTAdb_server_c::PUT_MESG is for PUT command
	case QUANTAdb_server_c::PUT_MESG:
		server->handlePut(buf, *bufferSize);
		return QUANTAnet_tcpReflector_c::OK;
		// QUANTAdb_server_c::FETCH_MESG is for FETCH command
	case QUANTAdb_server_c::FETCH_MESG:
		server->handleFetch(newClient, buf, *bufferSize);

		// For a fetch the reply is a put sent only to the originating
		// client so the message is not forwarded to anyone else.
		return QUANTAnet_tcpReflector_c::SKIP_DISTRIBUTION;
	case QUANTAdb_server_c::COMMIT_MESG:
		server->handleCommit(newClient, buf, *bufferSize);

		// For a fetch the reply is a put sent only to the originating
		// client so the message is not forwarded to anyone else.
		return QUANTAnet_tcpReflector_c::SKIP_DISTRIBUTION;
		//QUANTAdb_server_c::BLOCKING_FETCH_MESG is for blockingFetch command
	case QUANTAdb_server_c::BLOCKING_FETCH_MESG:
		server->handleBlockingFetch(newClient, buf, *bufferSize);
		return QUANTAnet_tcpReflector_c::SKIP_DISTRIBUTION;
	default:
		printf("QUANTAdb_server_c: Unknown incoming command.\n");
		return QUANTAnet_tcpReflector_c::SKIP_DISTRIBUTION;
	}
}

int QUANTAdb_server_c :: udpDataHandler(QUANTAnet_udp_c *newClient, char** buffer, int *bufferSize, void *userData)
{
    QUANTAdb_server_c *server = (QUANTAdb_server_c*) userData;

    // First byte of buffer is command code.
    char* buf = *buffer;

    QUANTAnet_datapack_c data;
    data.initUnpack(buf, *bufferSize);

    char command;
    data.unpackChar(&command);

    switch(command) {
    
    // QUANTAdb_server_c::PUT_MESG is for PUT command
    case QUANTAdb_server_c::PUT_MESG:
        server->handlePut(buf, *bufferSize);
        return QUANTAnet_udpReflector_c::OK;
    
    // The db server right now only accepts the put command via udp, the fetch and commit are done using reliable connections 
    
    default:
        printf("QUANTAdb_server_c: Unknown incoming command.\n");
        return QUANTAnet_udpReflector_c::SKIP_DISTRIBUTION;                         }                                                                           }
        
void QUANTAdb_server_c :: handlePut(char *buf, int size)
{

	QUANTAnet_datapack_c data;

	data.initUnpack(buf,size);

	// Remove command
	char command;
	data.unpackChar(&command);

	// get pathsize
	int pathsize;
	data.unpackInt(&pathsize);

	// get pathname
	char* pathname =NULL;
	if (pathsize) {
		pathname = new char[pathsize];
		if (pathname == NULL) {
			printf("QUANTAdb_server_c::handlePut: Memory alloc error\n");
			return;
		}

		data.unpack(pathname, pathsize);
	}

	// get keysize
	int keysize;
	data.unpackInt(&keysize);

	// get keyname
	char* keyname = NULL;
	if (keysize) {
		keyname = new char[keysize];
		if (keyname == NULL) {
			printf("QUANTAdb_server_c::handlePut: Memory alloc error\n");
			if (pathsize) delete[] pathname;
			return;
		}

		data.unpack(keyname, keysize);
	}

	// get datasize
	int datasize;
	data.unpackInt(&datasize);

	char* rawdata = NULL;
	if (datasize) {
		// get data
		rawdata = new char[datasize];

		if (rawdata == NULL) {
			printf("QUANTAdb_server_c::handlePut: Memory alloc error\n");
			if (keysize) delete[] keyname;
			if (pathsize) delete[] pathname;
			return;
		}
		data.unpack(rawdata, datasize);
	}

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);
	if (aKey == NULL) {
#ifdef WIN32
		printf("QUANTAdb_server_c::handlePut: Cannot access key %s/%s\n",pathname, keyname);
#else
		printf("QUANTAdb_server_c::handlePut: Cannot access key %s\\%s\n",pathname, keyname);
#endif
		if (rawdata) delete[] rawdata;
		if (keysize) delete[] keyname;
		if (pathsize) delete[] pathname;
		return;
	}

	aKey->alloc(datasize);
	char *rawBuffer = aKey->getData();

	if (datasize)
		memcpy(rawBuffer, rawdata, datasize);

	if (datasize)
		delete[] rawdata;
	delete aKey;

	if (keysize)
		delete[] keyname;
	if (pathsize)
		delete[] pathname;

}

void QUANTAdb_server_c :: handleFetch(QUANTAnet_tcpReflectorClient_c *client, char *buf, int size)
{

	QUANTAnet_datapack_c data;

	data.initUnpack(buf,size);

	// Remove command
	char command;
	data.unpackChar(&command);

	// get pathsize
	int pathsize;
	data.unpackInt(&pathsize);

	// get pathname
	char* pathname = NULL;
	if (pathsize) {
		pathname = new char[pathsize];
		if (pathname == NULL) {
			printf("QUANTAdb_server_c::handleFetch: Memory alloc error\n");

			return;
		}

		data.unpack(pathname, pathsize);
	}

	// get keysize
	int keysize;
	data.unpackInt(&keysize);

	char* keyname = NULL;

	if (keysize) {
	// get keyname
		keyname = new char[keysize];
		if (keyname == NULL) {
			if (pathsize) delete[] pathname;
			printf("QUANTAdb_server_c::handleFetch: Memory alloc error\n");

			return;
		}

		data.unpack(keyname, keysize);
	}

	sendPutMessage(client, pathname, pathsize, keyname, keysize);

	if (keysize)
		delete[] keyname;
	if (pathsize)
		delete[] pathname;
}

void QUANTAdb_server_c :: handleBlockingFetch(QUANTAnet_tcpReflectorClient_c *client, char *buf, int size)
{

        QUANTAnet_datapack_c data;

        data.initUnpack(buf,size);

        // Remove command
        char command;
        data.unpackChar(&command);

        // get pathsize
        int pathsize;
        data.unpackInt(&pathsize);

        // get pathname
        char* pathname = NULL;
        if (pathsize) {
                pathname = new char[pathsize];
                if (pathname == NULL) {
                        printf("QUANTAdb_server_c::handleBlockingFetch: Memory alloc error\n");
                        return;
                }
                data.unpack(pathname, pathsize);
        }
        // get keysize
        int keysize;
        data.unpackInt(&keysize);
        char* keyname = NULL;
        if (keysize) {
            // get keyname
            keyname = new char[keysize];
            if (keyname == NULL) {
                        if (pathsize) delete[] pathname;
                        printf("QUANTAdb_server_c::handleBlockingFetch: Memory alloc error\n");

                        return;
                }
                data.unpack(keyname, keysize);
        }
    keyToolManager_c::status_t status;
    int answer, replyBufferSize;
    char* replyBuffer;
    QUANTAnet_datapack_c reply;
    //get the key
    keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);
    if (aKey == NULL) {
#ifdef WIN32
        printf("QUANTAdb_server_c::sendPutMessage: Cannot access key %s/%s\n",pathname, keyname);
#else
        printf("QUANTAdb_server_c::sendPutMessage: Cannot access key %s\\%s\n",pathname, keyname);
#endif
        return;
    }
    if (status == keyToolManager_c::KEY_COMPLETELY_NEW)
    {
        QUANTAnet_datapack_c packer;
        answer = KEY_DOES_NOT_EXIST;
        //fprintf(stderr, "\nNew key");
        replyBufferSize = packer.sizeof_int() + packer.sizeof_char();
        replyBuffer = new char[replyBufferSize+1];
        packer.initPack(replyBuffer, replyBufferSize);
        packer.packChar(QUANTAdb_server_c::BLOCKING_FETCH_MESG);
        packer.packInt(answer);
    }
    else
    {
        answer = QUANTAdb_server_c::OK;
        char* thedata = aKey->getData();
        int thedatasize = aKey->getDataSize();

        replyBufferSize = QUANTAnet_datapack_c::sizeof_char() + //command
        QUANTAnet_datapack_c::sizeof_int()+    //answer
        QUANTAnet_datapack_c::sizeof_int()+    //data follows
        QUANTAnet_datapack_c::sizeof_char(pathsize)+
        QUANTAnet_datapack_c::sizeof_int()+
        QUANTAnet_datapack_c::sizeof_char(keysize)+
        QUANTAnet_datapack_c::sizeof_int()+
        QUANTAnet_datapack_c::sizeof_char(thedatasize);

        replyBuffer = new char[replyBufferSize];
        if (replyBuffer == NULL) {
            delete aKey;
	    printf("QUANTAdb_server_c::sendPutMessage: Memory alloc error\n");
            return;
        }
        QUANTAnet_datapack_c reply;
        reply.initPack(replyBuffer,replyBufferSize);
        reply.packChar(QUANTAdb_server_c::BLOCKING_FETCH_MESG);
        reply.packInt(QUANTAdb_server_c::OK);
        reply.packInt(pathsize);
        if (pathsize)
            reply.pack(pathname, pathsize);
        reply.packInt(keysize);
        if (keysize)
            reply.pack(keyname, keysize);
        reply.packInt(thedatasize);

        if (thedatasize)
            reply.pack(thedata, thedatasize);
    } //end else
    //fprintf(stderr, "\nBlocking fetch -sending reply");
    client->write(replyBuffer, &replyBufferSize);
    //fprintf(stderr, "\nBlocking fetch -finished sending reply");
    delete[] replyBuffer;
    delete aKey;

        if (keysize)
                delete[] keyname;
        if (pathsize)
                delete[] pathname;
}


void QUANTAdb_server_c :: handleCommit(QUANTAnet_tcpReflectorClient_c *client, char *buf, int size)
{

	QUANTAnet_datapack_c data;

	data.initUnpack(buf,size);

	// Remove command
	char command;
	data.unpackChar(&command);

	// get pathsize
	int pathsize;
	data.unpackInt(&pathsize);

	// get pathname
	char* pathname = NULL;
	if (pathsize) {
		pathname = new char[pathsize];
		if (pathname == NULL) {
			printf("QUANTAdb_server_c::handleCommit: Memory alloc error\n");

			return;
		}

		data.unpack(pathname, pathsize);
	}

	// get keysize
	int keysize;
	data.unpackInt(&keysize);

	char* keyname = NULL;

	if (keysize) {
	// get keyname
		keyname = new char[keysize];
		if (keyname == NULL) {
			if (pathsize)
        delete[] pathname;
			printf("QUANTAdb_server_c::handleFetch: Memory alloc error\n");

			return;
		}

		data.unpack(keyname, keysize);
	}

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);

	if (aKey == NULL) {
		return;
	}

	aKey->commit();

	if (keysize)
		delete[] keyname;
	if (pathsize)
		delete[] pathname;
}

void QUANTAdb_server_c :: sendPutMessage(QUANTAnet_tcpReflectorClient_c *client, char* pathname, int pathsize, char* keyname, int keysize)
{
	// Generate a reply message to send back.

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);
	if (aKey == NULL) {
#ifdef WIN32
		printf("QUANTAdb_server_c::sendPutMessage: Cannot access key %s/%s\n",pathname, keyname);
#else
		printf("QUANTAdb_server_c::sendPutMessage: Cannot access key %s\\%s\n",pathname, keyname);
#endif
		return;
	}


	char* thedata = aKey->getData();
	int thedatasize = aKey->getDataSize();

	int replyBufferSize = QUANTAnet_datapack_c::sizeof_char()+
		QUANTAnet_datapack_c::sizeof_int()+
		QUANTAnet_datapack_c::sizeof_char(pathsize)+
		QUANTAnet_datapack_c::sizeof_int()+
		QUANTAnet_datapack_c::sizeof_char(keysize)+
		QUANTAnet_datapack_c::sizeof_int()+
		QUANTAnet_datapack_c::sizeof_char(thedatasize);

	char *replyBuffer = new char[replyBufferSize];

	if (replyBuffer == NULL) {
		delete aKey;
		printf("QUANTAdb_server_c::sendPutMessage: Memory alloc error\n");
		return;
	}

	QUANTAnet_datapack_c reply;

	reply.initPack(replyBuffer,replyBufferSize);
	reply.packChar(QUANTAdb_server_c::PUT_MESG);
	reply.packInt(pathsize);
	if (pathsize)
		reply.pack(pathname, pathsize);
	reply.packInt(keysize);
	if (keysize)
		reply.pack(keyname, keysize);
	reply.packInt(thedatasize);

	if (thedatasize)
		reply.pack(thedata, thedatasize);

	client->write(replyBuffer, &replyBufferSize);
	delete[] replyBuffer;
	delete aKey;
}

void QUANTAdb_server_c:: showStats(char* streamInfo, char* comment)
{
    if (server) server->showStats(streamInfo, comment);    
}

int QUANTAdb_server_c::logStats(char* streamInfo, char* comment, FILE* filePtr)
{
    return (server) ? server->logStats(streamInfo, comment, filePtr) : QUANTAnet_perfMonitor_c::FAILED;
}

int QUANTAdb_server_c::sendStats(char* streamInfo, char* comment)
{
    return (server) ? server->sendStats(streamInfo, comment) : QUANTAnet_perfMonitor_c::FAILED;
}

int QUANTAdb_server_c::initSendStats(char* monitorClientIP, int port)
{
    return (server) ? server->initSendStats(monitorClientIP,port) : QUANTAnet_perfMonitor_c::FAILED;
}

void QUANTAdb_server_c::exitSendStats()
{
    if (server) server->exitSendStats();
}


/*************************************************************************/


QUANTAdb_client_c :: QUANTAdb_client_c()
{
	client = NULL;
    udpClient = NULL;
	triggerCallback = NULL;
	callbackUserData = NULL;
}

QUANTAdb_client_c :: ~QUANTAdb_client_c()
{
	delete clientmutex;
	delete callbackmutex;
	if (client){
		client->close(); 
		delete client;
	}
    if (udpClient)
        delete udpClient;
	if (keyToolManager) {
		delete keyToolManager;
	}
}

int QUANTAdb_client_c :: init(char* servername, int serverport, char *dbname,  int udp_reflector_port)
{

	clientmutex = new QUANTAts_mutex_c;
	callbackmutex = new QUANTAts_mutex_c;

	if (clientmutex == NULL) {
		return QUANTAdb_client_c::MEM_ALLOC_ERR;
	}
	if (callbackmutex == NULL) {
		return QUANTAdb_client_c::MEM_ALLOC_ERR;
	}


	keyToolManager = new keyToolManager_c(dbname);

	if (keyToolManager == NULL) {
		return QUANTAdb_client_c::MEM_ALLOC_ERR;
	}

	itsServerPort = serverport;
    	udpReflectorPort = udp_reflector_port;
	itsServerName[0]='\0';

	if (servername == NULL)
		return QUANTAdb_client_c::OK;;

	strcpy(itsServerName, servername);
    udpClient = new QUANTAnet_udp_c;
    if (udpClient == NULL)
        return QUANTAdb_client_c::MEM_ALLOC_ERR;
	udpClient->init();
    udpClient->setSendAddress(itsServerName, udpReflectorPort);
    udpClient->enableInstrumentation();
    udpClient->makeNonBlocking();

    return reconnect();

	// Worry about autoreconnect later
}

int QUANTAdb_client_c :: reconnect(char* servername, int serverport)
{
	clientmutex->lock();
	char theserver[256];
	int theport;

	if (servername == NULL)
		strcpy(theserver, itsServerName);
	if (servername) {
		strcpy(theserver, servername);
		strcpy(itsServerName , servername);
	}
	if (serverport == 0)
		theport = itsServerPort;
	if (serverport) {
		theport =serverport;
		itsServerPort = theport;
	}

	if (theserver[0] == '\0') {
		clientmutex->unlock();
		return QUANTAdb_client_c :: IGNORED;
	}

	client = new QUANTAnet_tcpReflectorClient_c;
	if (client == NULL) {
		QUANTAusleep(1000);
		clientmutex->unlock();
		return QUANTAdb_client_c::MEM_ALLOC_ERR;
	}

	int status = client->connectToServer(theserver, theport);
	if (status == QUANTAnet_tcpReflectorClient_c :: FAILED){
		delete client;
		client = NULL;
		QUANTAusleep(1000);
		clientmutex->unlock();
		return QUANTAdb_client_c::FAILED;
	}
	printf("QUANTA DB client connected to: %s:%d\n",theserver,theport);

	clientmutex->unlock();
	return QUANTAdb_client_c::OK;

}

int QUANTAdb_client_c :: fetch(char* pathname, char* keyname)
{
	clientmutex->lock();
	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);

	int pathsize = 0;
	int keysize = 0;
	if (pathname)
		pathsize = strlen(pathname)+1;

	if (keyname == NULL) {
		printf("QUANTAdb_client_c::fetch: Cannot fetch a NULL keyname\n");
		clientmutex->unlock();
		return QUANTAdb_client_c::FAILED;
	}

	keysize = strlen(keyname)+1;

	int sendBufferSize = QUANTAnet_datapack_c::sizeof_char() +
		QUANTAnet_datapack_c::sizeof_int()+
		QUANTAnet_datapack_c::sizeof_char(pathsize)+
		QUANTAnet_datapack_c::sizeof_int()+
		QUANTAnet_datapack_c::sizeof_char(keysize);

	char *sendBuffer = new char[sendBufferSize];

	QUANTAnet_datapack_c send;
	send.initPack(sendBuffer,sendBufferSize);
	send.packChar(QUANTAdb_server_c::FETCH_MESG);
	send.packInt(pathsize);
	if (pathsize)
		send.pack(pathname, pathsize);
	send.packInt(keysize);
	if (keysize)
		send.pack(keyname, keysize);

	int statusc = QUANTAnet_tcpReflectorClient_c::FAILED;
	if (client) 
	statusc = client->write(sendBuffer, &sendBufferSize);

	delete[] sendBuffer;
	delete aKey;
	clientmutex->unlock();

	if (statusc == QUANTAnet_tcpReflectorClient_c::OK) {
		return QUANTAdb_client_c::OK;
	}
	return QUANTAdb_client_c::FAILED;
}

int QUANTAdb_client_c :: blockingFetch(char* pathname, char* keyname)
{
    clientmutex->lock();
    keyToolManager_c::status_t status;
    keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);

    int pathsize = 0;
    int keysize = 0;
    if (pathname)
        pathsize = strlen(pathname)+1;

    if (keyname == NULL) {
        printf("QUANTAdb_client_c::fetch: Cannot fetch a NULL keyname\n");
        clientmutex->unlock();
        return QUANTAdb_client_c::FAILED;
    }

    keysize = strlen(keyname)+1;

    int sendBufferSize = QUANTAnet_datapack_c::sizeof_char() +
        QUANTAnet_datapack_c::sizeof_int()+
        QUANTAnet_datapack_c::sizeof_char(pathsize)+
        QUANTAnet_datapack_c::sizeof_int()+
        QUANTAnet_datapack_c::sizeof_char(keysize);

    char *sendBuffer = new char[sendBufferSize];

    QUANTAnet_datapack_c send;
    send.initPack(sendBuffer,sendBufferSize);
    send.packChar(QUANTAdb_server_c::BLOCKING_FETCH_MESG);
    send.packInt(pathsize);
    if (pathsize)
        send.pack(pathname, pathsize);
    send.packInt(keysize);
    if (keysize)
        send.pack(keyname, keysize);

    int statusc = QUANTAnet_tcpReflectorClient_c::FAILED;
    if (!client)
        printf("\nClient is not connected");
    if (client)
        statusc = client->write(sendBuffer, &sendBufferSize);
    clientmutex->unlock();

    // Clean up memory involved in the send command.
    delete[] sendBuffer;
    delete aKey;

    //now wait for data
    int dataSize, replystatus;
    char* data;
    int DONE=0;

   /*
    Repeat this Loop till you get a reply to the blocking fetch request.
    This has to be done because at times the client might have old messages queued up from the server and has to clear them. So keep reading till you get a reply to this request
    */
    while(!DONE)
    {
        clientmutex->lock();
        if (client)
	    replystatus = client->read(&data, &dataSize, QUANTAnet_tcpReflectorClient_c::BLOCKING);
        if (replystatus == QUANTAnet_tcpReflectorClient_c :: OK) {
            QUANTAnet_datapack_c pack;
            char command;
            pack.initUnpack(data, dataSize);
            pack.unpackChar(&command);
            if (command == QUANTAdb_server_c::BLOCKING_FETCH_MESG)
            {
                DONE =1;
                clientmutex->unlock();
		return handleFetchReply(data, dataSize);
            }
            else
            {
                clientmutex->unlock();
                handleMessage(data, dataSize);
            }
	}

	if (replystatus == QUANTAnet_tcpReflectorClient_c :: CONNECTION_TERMINATED)
        {
	    delete client;
	    client = NULL;
            clientmutex->unlock();
	    return QUANTAdb_client_c::CONNECTION_TERMINATED;
	}
    }
    return QUANTAdb_client_c::OK;
}

int QUANTAdb_client_c :: handleFetchReply(char* buf, int bufsize)
{
    clientmutex->lock();
    QUANTAnet_datapack_c data;
    data.initUnpack(buf,bufsize);

    // Remove command
    char command;
    int answer;
    data.unpackChar(&command);
    data.unpackInt(&answer);
    if (answer == QUANTAdb_server_c::KEY_DOES_NOT_EXIST)
    {
        clientmutex->unlock();
        return QUANTAdb_client_c::KEY_DOES_NOT_EXIST;
    }

    //key exists - unpack the data and Store in the local db just as an incoming put
    // get pathsize
    int pathsize;
    data.unpackInt(&pathsize);

    char* pathname =NULL;

    if (pathsize) {
    // get pathname
        pathname = new char[pathsize];
        if (pathname == NULL) {
            printf("QUANTAdb_client_c::handleFetchReply: Memory alloc error\n");            clientmutex->unlock();
            return QUANTAdb_client_c::FAILED;
        }
        data.unpack(pathname, pathsize);
    }

    // get keysize
    int keysize;
    data.unpackInt(&keysize);

    char* keyname = NULL;

    if (keysize) {
        // get keyname
        keyname = new char[keysize];
        if (keyname == NULL) {
            if (pathsize)
              delete[] pathname;
            printf("QUANTAdb_client_c::handleFetchReply: Memory alloc error\n");            clientmutex->unlock();
            return QUANTAdb_client_c::FAILED;
        }

        data.unpack(keyname, keysize);
    }

    // get datasize
    int datasize;
    data.unpackInt(&datasize);

    char* rawdata = NULL;

    if (datasize) {
        // get data
        rawdata = new char[datasize];
        if (keyname == NULL) {
            if (pathsize)
              delete[] pathname;
            if (keysize)
              delete[] keyname;
            printf("QUANTAdb_client_c::handleFetchReply: Memory alloc error\n");            clientmutex->unlock();
            return QUANTAdb_client_c::FAILED;
        }

        data.unpack(rawdata, datasize);
    }

    keyToolManager_c::status_t status;
    keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);

    aKey->alloc(datasize);
    char *rawBuffer = aKey->getData();

    if (datasize) {
        memcpy(rawBuffer, rawdata, datasize);
        delete[] rawdata;
    }
    clientmutex->unlock();
    delete aKey;
    if (keysize)
        delete[] keyname;
    if (pathsize)
        delete[] pathname;
    return QUANTAdb_client_c::OK;
}


char* QUANTAdb_client_c :: get(char* pathname, char* keyname, int* datasize)
{

	if (keyname == NULL) {
		printf("QUANTAdb_client_c::get: Cannot get a NULL keyname\n");
		return NULL;
	}

	clientmutex->lock();

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);
	*datasize = aKey->getDataSize();
	if (*datasize == 0) {
		clientmutex->unlock();
		return NULL;
	}

	char* data = new char[*datasize];
	if (data == NULL) {
		clientmutex->unlock();
		return NULL;
	}

	char* keydata = aKey->getData();
	memcpy(data, keydata, *datasize);

	clientmutex->unlock();

	return data;
}

int QUANTAdb_client_c :: commit(char* pathname, char* keyname)
{
	if (keyname == NULL) {
		printf("QUANTAdb_client_c::commit: Cannot commit to a NULL keyname\n");
		return QUANTAdb_client_c::FAILED;
	}

	clientmutex->lock();

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);

	if (aKey == NULL) {
		clientmutex->unlock();

		return QUANTAdb_client_c::MEM_ALLOC_ERR;
	}

	aKey->commit();
	clientmutex->unlock();

	return QUANTAdb_client_c::OK;

}

int QUANTAdb_client_c :: remoteCommit(char* pathname, char* keyname)
{
	clientmutex->lock();

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);

	if (aKey == NULL) {
		clientmutex->unlock();
		return QUANTAdb_client_c::MEM_ALLOC_ERR;
	}

	int pathsize = 0;
	int keysize = 0;
	if (pathname)
		pathsize = strlen(pathname)+1;

	if (keyname == NULL) {
		printf("QUANTAdb_client_c::remoteCommit: Cannot remoteCommit a NULL keyname\n");
		clientmutex->unlock();

		return QUANTAdb_client_c::FAILED;
	}

	keysize = strlen(keyname)+1;

	int sendBufferSize = QUANTAnet_datapack_c::sizeof_char() +
		QUANTAnet_datapack_c::sizeof_int()+
		QUANTAnet_datapack_c::sizeof_char(pathsize)+
		QUANTAnet_datapack_c::sizeof_int()+
		QUANTAnet_datapack_c::sizeof_char(keysize);

	char *sendBuffer = new char[sendBufferSize];

	QUANTAnet_datapack_c send;
	send.initPack(sendBuffer,sendBufferSize);

	send.packChar(QUANTAdb_server_c::COMMIT_MESG);
	send.packInt(pathsize);
	if (pathsize)
		send.pack(pathname, pathsize);
	send.packInt(keysize);
	if (keysize)
		send.pack(keyname, keysize);

	int statusc = QUANTAnet_tcpReflectorClient_c::FAILED;
	if (client) 
		statusc = client->write(sendBuffer, &sendBufferSize);

	delete[] sendBuffer;
	delete aKey;

	clientmutex->unlock();

	if (statusc == QUANTAnet_tcpReflectorClient_c::OK) {
		return QUANTAdb_client_c::OK;
	}
	return QUANTAdb_client_c::FAILED;
}

int QUANTAdb_client_c :: put(char* pathname, char* keyname, char* data, int thedatasize, int typeOfConnection)
{

  // JASON: Force reliable connection until we can create a proper multithreaded QUANTAdb
  typeOfConnection = QUANTAdb_client_c::RELIABLE_CONNECTION;

	int pathsize = 0;
	int keysize = 0;
	if (pathname)
		pathsize = strlen(pathname)+1;

	if (keyname == NULL) {
		printf("QUANTAdb_client_c::put: Cannot put to a NULL keyname\n");
		return QUANTAdb_client_c::FAILED;
	}

	keysize = strlen(keyname)+1;

	clientmutex->lock();

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);

	if (aKey == NULL) {
		clientmutex->unlock();

		return QUANTAdb_client_c::MEM_ALLOC_ERR;
	}

	aKey->alloc(thedatasize);

	char* thedata = aKey->getData();

	if (thedatasize)
		memcpy(thedata, data, thedatasize);

	int sendBufferSize = QUANTAnet_datapack_c::sizeof_char() +
		QUANTAnet_datapack_c::sizeof_int()+
		QUANTAnet_datapack_c::sizeof_char(pathsize)+
		QUANTAnet_datapack_c::sizeof_int()+
		QUANTAnet_datapack_c::sizeof_char(keysize)+
		QUANTAnet_datapack_c::sizeof_int()+
		QUANTAnet_datapack_c::sizeof_char(thedatasize);

	char *sendBuffer = new char[sendBufferSize];
	if (sendBuffer == NULL) {
		delete aKey;
		clientmutex->unlock();
		return QUANTAdb_client_c::MEM_ALLOC_ERR;
	}

	QUANTAnet_datapack_c send;

	send.initPack(sendBuffer,sendBufferSize);
	send.packChar(QUANTAdb_server_c::PUT_MESG);
	send.packInt(pathsize);
	if (pathsize)
		send.pack(pathname, pathsize);
	send.packInt(keysize);

	if (keysize)
		send.pack(keyname, keysize);
	send.packInt(thedatasize);

	if (thedatasize)
		send.pack(thedata, thedatasize);
        
    if (typeOfConnection == QUANTAdb_client_c::RELIABLE_CONNECTION)
    {
	    if (client) 
		    client->write(sendBuffer, &sendBufferSize);
    }
    else            //Unreliable udp is to be used
    {
        if (thedatasize <= QUANTAdb_client_c::MAX_UDP_PACKET_SIZE)
        {
            if (udpClient)
                udpClient->send(sendBuffer, sendBufferSize); 
        }
        else
        {
            printf("\nPacket size exceeds 1 KB, by-passing UDP connection and sending it via TCP\n");
            if (client)
               client->write(sendBuffer, &sendBufferSize);
        }
    }
	clientmutex->unlock();

	delete[] sendBuffer;
	delete aKey;

	return QUANTAdb_client_c::OK;
}

int QUANTAdb_client_c :: process()
{
	int status;
	char* data, *udpData;
	int dataSize;
    int udpDataSize = QUANTAdb_client_c::MAX_UDP_PACKET_SIZE;
    int udpReturnSize;
	clientmutex->lock();

    if (udpClient)
    {
        udpData = new char[udpDataSize+1];
        udpReturnSize = udpClient->receive(udpData, udpDataSize);
        if (udpReturnSize > 0)
            handleMessage(udpData, udpDataSize);
        delete[] udpData;
    }
    if (client)
		status = client->read(&data, &dataSize, QUANTAnet_tcpReflectorClient_c::NON_BLOCKING);
	clientmutex->unlock();

	if (client == NULL)
    return QUANTAdb_client_c::CONNECTION_TERMINATED;

	if (status == QUANTAnet_tcpReflectorClient_c :: OK) {
		handleMessage(data, dataSize);
		delete data;
		return QUANTAdb_client_c::OK;
	}

	clientmutex->lock();
	if (status == QUANTAnet_tcpReflectorClient_c :: CONNECTION_TERMINATED) {
		delete client;
		client = NULL;
		clientmutex->unlock();

		return QUANTAdb_client_c::CONNECTION_TERMINATED;
	}

	clientmutex->unlock();

	return QUANTAdb_client_c::OK;

}

void QUANTAdb_client_c :: handleMessage(char* data, int dataSize)
{
	QUANTAnet_datapack_c thepack;
	thepack.initUnpack(data,dataSize);

	char command;
	thepack.unpackChar(&command);

	switch(command) {
	case QUANTAdb_server_c::PUT_MESG:
		handleIncomingPut(data, dataSize);
		break;
	default:
		printf("QUANTAdb_client_c: Unknown incoming command.\n");
	}
}

void QUANTAdb_client_c :: handleIncomingPut(char* buf, int bufsize)
{
	clientmutex->lock();

	QUANTAnet_datapack_c data;

	data.initUnpack(buf,bufsize);

	// Remove command
	char command;
	data.unpackChar(&command);

	// get pathsize
	int pathsize;
	data.unpackInt(&pathsize);

	char* pathname =NULL;

	if (pathsize) {
	// get pathname
		pathname = new char[pathsize];
		if (pathname == NULL) {
			printf("QUANTAdb_client_c::handleIncomingPut: Memory alloc error\n");
			clientmutex->unlock();

			return;
		}
		data.unpack(pathname, pathsize);
	}

	// get keysize
	int keysize;
	data.unpackInt(&keysize);

	char* keyname = NULL;

	if (keysize) {
		// get keyname
		keyname = new char[keysize];
		if (keyname == NULL) {
			if (pathsize)
        delete[] pathname;
			printf("QUANTAdb_client_c::handleIncomingPut: Memory alloc error\n");
			clientmutex->unlock();

			return;
		}

		data.unpack(keyname, keysize);
	}

	// get datasize
	int datasize;
	data.unpackInt(&datasize);

	char* rawdata = NULL;

	if (datasize) {
		// get data
		rawdata = new char[datasize];
		if (keyname == NULL) {
			if (pathsize)
        delete[] pathname;
			if (keysize)
        delete[] keyname;
			printf("QUANTAdb_client_c::handleIncomingPut: Memory alloc error\n");
			clientmutex->unlock();

			return;
		}

		data.unpack(rawdata, datasize);
	}

	keyToolManager_c::status_t status;
	keyToolKey_c *aKey = keyToolManager->getKey(pathname, keyname, &status);

	if (aKey == NULL) {
		if (pathsize)
      delete[] pathname;
		if (keysize)
      delete[] keyname;
		if (datasize)
      delete[] rawdata;
#ifdef WIN32
		printf("QUANTAdb_client_c::handleIncomingPut:  Cannot access key %s\\%s\n",pathname, keyname);
#else
		printf("QUANTAdb_client_c::handleIncomingPut:  Cannot access key %s/%s\n",pathname, keyname);
#endif
		clientmutex->unlock();

		return;
	}

	aKey->alloc(datasize);
	char *rawBuffer = aKey->getData();

	if (datasize) {
		memcpy(rawBuffer, rawdata, datasize);
		delete[] rawdata;
	}

	delete aKey;


	notifyPathname = pathname;
	notifyKeyname = keyname;

	clientmutex->unlock();

	notify();

	if (triggerCallback) {
		callbackmutex->lock();
		triggerCallback(pathname, keyname, callbackUserData);
		callbackmutex->unlock();
	}
	if (keysize) 
		delete[] keyname;

	if (pathsize)
		delete[] pathname;


}

void QUANTAdb_client_c :: trigger(void (*callback) (char* pathname, char* keyname, void *userData), void* userData)
{
	clientmutex->lock();
	triggerCallback = callback;
	callbackUserData = userData;
	clientmutex->unlock();

}

void QUANTAdb_client_c::showStats(char* streamInfo, char* comment)
{
    if (client) client->showStats(streamInfo, comment);    
}

int QUANTAdb_client_c::logStats(char* streamInfo, char* comment, FILE* filePtr)
{
    return (client) ? client->logStats(streamInfo, comment, filePtr) : QUANTAnet_perfMonitor_c::FAILED;
}

int QUANTAdb_client_c::sendStats(char* streamInfo, char* comment)
{
    return (client) ? client->sendStats(streamInfo, comment) : QUANTAnet_perfMonitor_c::FAILED;
}

int QUANTAdb_client_c::initSendStats(char* monitorClientIP, int port)
{
    return (client) ? client->initSendStats(monitorClientIP,port) : QUANTAnet_perfMonitor_c::FAILED;
}

void QUANTAdb_client_c::exitSendStats()
{
    if (client) client->exitSendStats();
}
