/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <QUANTA/QUANTAnet_barrierTcp_c.hxx>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

/* 
 * TCP Barrier for sync. swaplock in a distributed environment
 */


QUANTAnet_barrierTcp_c::QUANTAnet_barrierTcp_c(char *key_id, int block_count, unsigned int port, int verbose)
 : QUANTAnet_barrierBase_c(key_id, block_count, port, verbose)
{
	// init the server
	server = new QUANTAnet_tcpServer_c();
	if(server->init(_port) < 0)
	{
		printf("QUANTAnet_barrierTcp_c cannot init port %d", _port);
		exit(EXIT_FAILURE);
	}
	create(); // start the thread cycle; calls run()
}


QUANTAnet_barrierTcp_c::~QUANTAnet_barrierTcp_c(void)
{
	_nodes.clear();
	delete _key_id;
	delete [] _recvline;
	delete [] command;
	delete [] arg;
	delete server;
}


int QUANTAnet_barrierTcp_c::validateNode(void)
{
	if(!strcasecmp(_key_id, command))
	{
		// the ids are a match
		if(_verbose)
			fprintf(stdout, "tcpbarrier: VAILD NODE: master key_id = %s, node key_id = %s\n", _key_id, command);
		return 1;
	}

	// failure to match
	if(_verbose)
		fprintf(stdout, "tcpbarrier: INVAILD NODE: port = %d, master key_id = %s, node key_id = %s\n", _port, _key_id, command);
	return 0;
}


// special commands: reset, key_id, block_count
//
void QUANTAnet_barrierTcp_c::checkSpecialCommands(void)
{
	if(!strcasecmp(command, "key_id"))
		_key_id = arg;
	else if(!strcasecmp(command, "block_count"))
		_block_count = atoi(arg);
	else if(!strcasecmp(command, "wait_count"))
		_wait_count = atoi(arg);
}


/* Check to see if the current node is one that we have not encountered
 * before.
 */
void QUANTAnet_barrierTcp_c::checkNode(void)
{
	char name[256];
	QUANTAnet_tcpClient_c *tmp_node;

	tmp_node = server->checkForNewConnections();
	if(tmp_node)
	{
		if(_verbose)
		{
			QUANTAnet_socketbase_c::getSelfIP(name);
			fprintf(stdout, "Node %d connected, %s:%u\n", _nodes.size(), name, tmp_node->getSelfPort());
		}
		_nodes.addAfter(tmp_node);
//		printf("node count = %d\n", _nodes.size());
	}
}


// Reply to all conencted nodes
void QUANTAnet_barrierTcp_c::broadcast(void)
{
	int status;
	int dataSize;
	char message[BROADCAST_DATA_SIZE];

	if(_nodes.size() == 0)
		return;

	strcpy(message,"BARRIER_RELEASE");
	dataSize = BROADCAST_DATA_SIZE;

	_nodes.first();
	while(!_nodes.isEmpty())
	{
		try
		{
			status = ((QUANTAnet_tcpClient_c *)_nodes.getCurrent())->write(message, &dataSize, QUANTAnet_tcpClient_c::NON_BLOCKING);
			if(status != QUANTAnet_tcpClient_c::OK)
			{
				// this should never happen cause dead nodes are
				// erased if the barrier cannot read from them
				if(_verbose)
					printf("QUANTAnet_barrierTcp_c::broadcast() - Writing BARRIER_RELEASE to node error\n");
			}

			_nodes.next();
		}
		catch(QUANTAmisc_noCurrentException_c ex)
		{
			break; // exit the loop if there are no more nodes in the list
		}
	}
}


void QUANTAnet_barrierTcp_c::run()
{
	int status;
	int dataSize;

	while(1)
	{
		checkNode();

		_nodes.first();
		while(!_nodes.isEmpty())
		{
			try
			{
				dataSize = TRANSMITTED_DATA_SIZE;
				status = ((QUANTAnet_tcpClient_c *)_nodes.getCurrent())->read(_recvline, &dataSize, QUANTAnet_tcpClient_c::NON_BLOCKING);
				if(status == QUANTAnet_tcpClient_c::OK)
				{
					sscanf(_recvline, "%s%s", command, arg);
					// check this nodes key_id to make sure it matches
					if(!validateNode())
					{
						// the key_ids did not match; deny this node
						// but, first check if the incoming message is a special command
						checkSpecialCommands();
						continue;
					}

					_wait_count++;
					if(_wait_count == _block_count)
					{
						broadcast();
						_wait_count = 0;
					}
				}
				else if(status != QUANTAnet_tcpClient_c::NON_BLOCKING_HAS_NO_DATA)
				{
					_nodes.removeCurrent();
					break;
				}
			
				_nodes.next();
			}
			catch(QUANTAmisc_noCurrentException_c ex)
			{
				break; // exit the loop if there are no more nodes in the list
			}
		}

		QUANTAusleep(10000);
	}
}


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


QUANTAnet_barrierNodeTcp_c::QUANTAnet_barrierNodeTcp_c(char *host, unsigned int port)
 : QUANTAnet_barrierNodeBase_c(host, port)
{
	client = new QUANTAnet_tcpClient_c();
	if(client->connectToServer(_host, _port) < 0)
	{
		printf("ERROR: Failed to connect to host: %s,  port: %i\n", _host, _port);
		exit(EXIT_FAILURE);
	}
}


QUANTAnet_barrierNodeTcp_c::~QUANTAnet_barrierNodeTcp_c(void)
{
	delete client;
}


void QUANTAnet_barrierNodeTcp_c::wait(char *key_id)
{
	int status, dataSize;
	char sendBuffer[TRANSMITTED_DATA_SIZE];
	char receiveBuffer[BROADCAST_DATA_SIZE];

	// - lock the local mutex
	// - send a message to the barrier that this node is blocking, waiting for
	// a return message to unlock the mutex and continue thread execution
	//

	sprintf(sendBuffer, key_id);
	dataSize = TRANSMITTED_DATA_SIZE;
	status = client->write(sendBuffer, &dataSize, QUANTAnet_tcpClient_c::NON_BLOCKING);
	if(status != QUANTAnet_tcpClient_c::OK)
	{
//		client->showStatus(status, dataSize);
	}

	// message sent to barrier, now lock the local mutex
	_mutex.lock();

	dataSize = BROADCAST_DATA_SIZE;
	status = client->read(receiveBuffer, &dataSize, QUANTAnet_tcpClient_c::BLOCKING);
	if(status == QUANTAnet_tcpClient_c::OK)
	{
//		printf("Data received: %s\n", receiveBuffer);
		_mutex.unlock();
	}
	else if(status != QUANTAnet_tcpClient_c::NON_BLOCKING_HAS_NO_DATA)
	{
//		client->showStatus(status, dataSize);
	}
}
