/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <QUANTA/QUANTAnet_barrierUdp_c.hxx>

#ifndef WIN32
#include <sys/errno.h>
#endif

/* 
 * UDP Barrier for sync. swaplock in a distributed environment
 */

// TODO:
// - allow a node to get the current node count from a barrier

QUANTAnet_barrierUdp_c::QUANTAnet_barrierUdp_c(char *key_id, int block_count, unsigned int port, int verbose)
 : QUANTAnet_barrierBase_c(key_id, block_count, port, verbose)
{
	_udpSocket = new QUANTAnet_udp_c();
	if(_udpSocket->init(_port) == 0)
	{
		printf("QUANTAnet_barrierUdp_c cannot init port %d", _port);
		exit(EXIT_FAILURE);
	}
	create(); // start the thread cycle; calls run()
}


QUANTAnet_barrierUdp_c::~QUANTAnet_barrierUdp_c(void)
{
	delete _udpSocket;
	_nodes.clear();
}


int QUANTAnet_barrierUdp_c::validateNode(void)
{
	if(!strcasecmp(_key_id, command))
	{
		// the ids are a match
		if(_verbose)
			fprintf(stdout, "udpbarrier: VAILD NODE: master key_id = %s, node key_id = %s\n", _key_id, command);
		return 1;
	}

	// failure to match
	if(_verbose)
		fprintf(stdout, "udpbarrier: INVAILD NODE: port = %d, master key_id = %s, node key_id = %s\n", _port, _key_id, command);
	return 0;
}



// special commands: reset, key_id, block_count
//
void QUANTAnet_barrierUdp_c::checkSpecialCommands(void)
{
	if(!strcasecmp(command, "key_id"))
		_key_id = arg;
	else if(!strcasecmp(command, "block_count"))
		_block_count = atoi(arg);
	else if(!strcasecmp(command, "wait_count"))
		_wait_count = atoi(arg);
}


/* Check to see if the current client is one that we have not encountered
 * before.
 */
void QUANTAnet_barrierUdp_c::checkNode(void)
{
	QUANTAnet_udp_c *tmp_node;

	// check this nodes key_id to make sure it matches
	if(!validateNode())
	{
		// the key_ids did not match; deny this node
		return;
	}

	// the node is valid; clone a new node and put it in the array
	tmp_node = _udpSocket->clone();
	tmp_node->copyReceiveAddressToSendAddress();
	_nodes.addAfter(tmp_node);
//	printf("node count = %d\n", _nodes.size());
	
	// increment number of waiting nodes
	_wait_count++;
	if(_wait_count == _block_count)
	{
		broadcast();
		_wait_count = 0;
	}
}


// Reply to all conencted clients
void QUANTAnet_barrierUdp_c::broadcast(void)
{
	char mesg[MAX_LINE];

	_nodes.first();
	while(!_nodes.isEmpty())
	{
		try
		{
			sprintf(mesg,"udpbarrier: BARRIER RELEASED!  TOTAL NODES: %d  KEY_ID: %s\n", _wait_count, _key_id);
			((QUANTAnet_udp_c *)_nodes.getCurrent())->send(mesg, strlen(mesg) + 1, QUANTAnet_udp_c::NON_BLOCKING);

			_nodes.next();
		}
		catch(QUANTAmisc_noCurrentException_c ex)
		{
			break; // exit the loop if there are no more nodes in the list
		}
	}
	
	_nodes.clear();
}


void QUANTAnet_barrierUdp_c::run()
{
	while(1)
	{
		while(_udpSocket->receive(_recvline, MAX_LINE, QUANTAnet_udp_c::BLOCKING) > 0)
		{
			sscanf(_recvline, "%s%s", command, arg);
			checkNode();
			checkSpecialCommands();
		}
#ifdef WIN32
		if(errno != WSAEWOULDBLOCK)
		{
			printf("RECEIVE ERROR\n");
		}
#else
		if(errno != EWOULDBLOCK)
		{
			printf("RECEIVE ERROR\n");
		}
#endif
	}
}


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


QUANTAnet_barrierNodeUdp_c::QUANTAnet_barrierNodeUdp_c(char *host, unsigned int port)
 : QUANTAnet_barrierNodeBase_c(host, port)
{
	_mainUdpSocket = new QUANTAnet_udp_c();
	_mainUdpSocket->init();
	if(_mainUdpSocket->setSendAddress(_host, _port) == 0)
	{
		printf("ERROR: Failed to connect to host: %s,  port: %i\n", _host, _port);
		exit(EXIT_FAILURE);
	}
}


QUANTAnet_barrierNodeUdp_c::~QUANTAnet_barrierNodeUdp_c(void)
{
	delete _mainUdpSocket;
}


void QUANTAnet_barrierNodeUdp_c::wait(char *key_id)
{
	int n;

	// - lock the local mutex
	// - send a message to the barrier that this node is blocking, waiting for
	// a return message to unlock the mutex and continue thread execution
	//
	n = strlen(key_id) + 1;

	// send message to the barrier
	if(_mainUdpSocket->send(key_id, n) != n)
	{
		printf("barrier node: sendto error\n");
		return;
	}

	// message sent to barrier, now lock the local mutex
	_mutex.lock();
	n = _mainUdpSocket->receive(_recvline, MAX_LINE, QUANTAnet_udp_c::BLOCKING);
	if(n < 0)
	{
		if(n == QUANTAnet_udp_c::NON_BLOCKING_HAS_NO_DATA)
			printf("Nonblocking has no data\n");
		else
			printf("Client receive error\n");
	}
	else
	{
//		printf("RECV SIZE %d ",n);
//		printf(_recvline);
		_mutex.unlock();
	}

}


void QUANTAnet_barrierNodeUdp_c::sendError(char *params)
{
	printf("***Error, QUANTAnet_barrierNodeUdp_c::sendCommand%s" \
		" : array out of bounds; MAX BYTE SIZE == %i\n\n", params, MAX_LINE);
	exit(0);
}


void QUANTAnet_barrierNodeUdp_c::sendCommand(char *command, int arg)
{
	char message[MAX_LINE];

	sprintf(message, "%s %d", command, arg);
	if(strlen(message) > MAX_LINE)
		sendError("(char *, int)");
	_mainUdpSocket->send(message, strlen(message) + 1);
}


void QUANTAnet_barrierNodeUdp_c::sendCommand(char *command, float arg)
{
	char message[MAX_LINE];
 
	sprintf(message, "%s %f", command, arg);
	if(strlen(message) > MAX_LINE)
		sendError("(char *, float)");
	_mainUdpSocket->send(message, strlen(message) + 1);
}


void QUANTAnet_barrierNodeUdp_c::sendCommand(char *command, char *arg)
{
	char message[MAX_LINE];

	sprintf(message, "%s %s", command, arg);
	if(strlen(message) > MAX_LINE)
		sendError("(char *, char *)");
	_mainUdpSocket->send(message, strlen(message) + 1);
}


void QUANTAnet_barrierNodeUdp_c::setBarrierBlockCount(int block_count)
{
	sendCommand("block_count", block_count);
}


void QUANTAnet_barrierNodeUdp_c::setBarrierWaitCount(int wait_count)
{
	sendCommand("wait_count", wait_count);
}


void QUANTAnet_barrierNodeUdp_c::setBarrierKeyID(char *key_id)
{
	sendCommand("key_id", key_id);
}
