/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

/* File: QUANTAnet_fecClient_c.cxx
   Description: The implementation of QUANTAnet_fecClient_c class
*/

#include <QUANTAnet_fecClient_c.hxx>
#ifdef WIN32
#include <QUANTA/QUANTAconfig_win32.hxx>
#else /* !WIN32 */
#include <QUANTA/QUANTAconfig.hxx>
#endif /* WIN32 */

#include <iostream.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>

#ifndef WIN32
#include <sys/time.h>
#endif /* !WIN32 */

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#define DEBUG

QUANTAnet_fecClient_c::QUANTAnet_fecClient_c()
{
	srcUDPSocket = new QUANTAnet_udp_c;
	fecUDPSocket = new QUANTAnet_udp_c;
	fecEngine = new QUANTAnet_fecEngine_c();
	srcUDPSocket->init();
	fecUDPSocket->init();
}

QUANTAnet_fecClient_c::~QUANTAnet_fecClient_c()
{
	for(unsigned int i=0; i<_numSourcePack; i++)
		delete srcData[i];
	for(unsigned int j=0; j<_numTotalPack; j++)
		delete totalData[j];
	delete srcData;
	delete totalData;
	delete srcPackPool;
	delete fecPackPool;
	delete srcUDPSocket;
	delete fecUDPSocket;
	fecEngine->release(fecMatrix);
	delete fecEngine;
}

// Set number of source pack in the whole block
void QUANTAnet_fecClient_c::setNumSourcePack(unsigned int numSourcePack)
{
	_numSourcePack = numSourcePack;
}

// Set number of packets in the whole block
void QUANTAnet_fecClient_c::setNumTotalPack(unsigned int numTotalPack)
{
	_numTotalPack = numTotalPack;
}

// Set the size of payload
void QUANTAnet_fecClient_c::setPayloadSize(unsigned int payloadSize)
{
	_payloadSize = payloadSize;
}

// Get number of source pack in the whole block
unsigned int QUANTAnet_fecClient_c::getNumSourcePack()
{
	return _numSourcePack;
}

// Get number of packets in the whole block
unsigned int QUANTAnet_fecClient_c::getNumTotalPack()
{
	return _numTotalPack;
}

// Get the size of payload
unsigned int QUANTAnet_fecClient_c::getPayloadSize()
{
	return _payloadSize;
}

// Set hostname and port number for QUANTAnet_udp_c object of source packet
int QUANTAnet_fecClient_c::setSrcSendAddress(char* host, unsigned short port)
{
	return srcUDPSocket->setSendAddress(host,port);
}

// Set hostname and port number for QUANTAnet_udp_c object of FEC packet
int QUANTAnet_fecClient_c::setFecSendAddress(char* host, unsigned short port)
{
	return fecUDPSocket->setSendAddress(host,port);
}

// Get QUANTAnet_udp_c pointer to the source packet socket 
QUANTAnet_udp_c* QUANTAnet_fecClient_c::getSrcUDPSocket()
{
	if(srcUDPSocket != NULL)
		return srcUDPSocket;
	else
		return NULL;
}

// Get QUANTAnet_udp_c pointer to the FEC packet socket
QUANTAnet_udp_c* QUANTAnet_fecClient_c::getFecUDPSocket()
{
	if(fecUDPSocket != NULL)
		return fecUDPSocket;
	else
		return NULL;
}

/* Initialize. The following parameters are needed: numSourcePack:number of source pack in the
	block, numTotalPack:number of packets in the block, payloadSize:size of payload,
	numRowofRecvBuff: number of rows of receiving buffer on the server site, host:server's name,
	srcPort:source port number, fecPort:FEC port number */
void QUANTAnet_fecClient_c::init(unsigned int numSourcePack, unsigned int numTotalPack, unsigned int payloadSize, char*
	host, unsigned short srcPort, unsigned short fecPort)
{
	setNumSourcePack(numSourcePack);
	setNumTotalPack(numTotalPack);
	setPayloadSize(payloadSize);
	
	if(setSrcSendAddress(host, srcPort) == 0)
		cerr << "<QUANTAnet_fecClient_c::init>setSrcSendAddress on srcUDPsocket failed!" << endl;
	if(setFecSendAddress(host, fecPort) == 0)
		cerr << "<QUANTAnet_fecClient_c::init>setFecSendAddress on fecUDPsocket failed!" << endl;
	fecMatrix = fecEngine->init(_numSourcePack, _numTotalPack);
	if(fecMatrix == NULL)
		cerr << "<QUANTAnet_fecClient_c::init>fecEngine init failed!" << endl;
	
	srcData = new char*[_numSourcePack];
	totalData = new char*[_numTotalPack];
	
	for(unsigned int i=0; i<_numSourcePack; i++)
	{
		if((srcData[i] = new char[_payloadSize]) == NULL)
			cerr << "<QUANTAnet_fecClient_c::init>allocating memery failed!" << endl;
	}
	
	for(unsigned int j=0; j<_numTotalPack; j++)
	{
		if((totalData[j] = new char[_payloadSize]) == NULL)
			cerr << "<QUANTAnet_fecClient_c::init>allocating memery failed!" << endl;
	}
	
	offsite = 2*sizeof(unsigned short)+sizeof(unsigned int);
	len = offsite + _payloadSize;
printf("len: %d, offsite: %d, _payloadSize: %d\n", len, offsite, _payloadSize);
	srcPackPool = new char[len];
	fecPackPool = new char[len];
	bzero(srcPackPool, len);
	bzero(fecPackPool, len);
	
	_countBlock = 0;
	_countSourcePack = 0;
	_countFecPack = 0;
	_countFecBlock = 0;
	_sendRate = 1.0;
}

void QUANTAnet_fecClient_c::setSendRate(double rate)
{
	_sendRate = rate;
}

double QUANTAnet_fecClient_c::getSendRate()
{
	return _sendRate;
}	

void QUANTAnet_fecClient_c::streamingSend(void(*updateFunc)(void*, int))
{
	int timeBtwBlocks, sent;
	unsigned int i;
	struct timeval start, now;
	int notFinishedYet = 1;
	FILE *fp = fopen("//tmp//fec_send.log", "w");
	
	if (fp==NULL) printf("log file open failed\n");

	printf("starting........\n");

	timeBtwBlocks = (_payloadSize * _numSourcePack * 8) / (int)_sendRate;
	#ifdef DEBUG
		printf("Interval between blocks is %d\n", timeBtwBlocks);
	#endif

	gettimeofday(&start, NULL);

	char *data = (char *)malloc(_numSourcePack*_payloadSize);
	while(notFinishedYet)	
	{
		gettimeofday(&now, NULL);
		if (USEC(&start, &now) > 60000000)
			notFinishedYet = 0;
		else if (USEC(&start, &now) < timeBtwBlocks * _countFecBlock)
		{
			usleep(100);
		}
		else 
		{
			double sendTime;
			sendTime = (double)now.tv_usec*0.000001 + now.tv_sec; 
			fprintf(fp, "%d\t%f\n", _countFecBlock, sendTime);
			updateFunc(data, _numSourcePack*_payloadSize);
			for(i=0; i<_numSourcePack; i++)
			{
				srcPack.srcPackNum = i;
				srcPack.BlockNum = _countFecBlock;
				srcPack.dataSize = _payloadSize;
				#ifdef DEBUG
				fprintf(fp, "send src pack %d in block %d\n", srcPack.srcPackNum, srcPack.BlockNum);
				#endif
				memcpy(srcPackPool, &srcPack, offsite);
				memcpy(&srcPackPool[offsite], data+i*_payloadSize, _payloadSize);
				sent = srcUDPSocket->send(srcPackPool, len);
				if (sent != len)
					cerr << "<QUANTAnet_fecClient_c::send> Source sending error!" << endl;
		
				//put src data into the array for FEC encoding
				memcpy(srcData[i], data, _payloadSize);
 	  	}

			for (i=0; i<_numTotalPack; i++)
				fecEngine->encode(fecMatrix, (void**)srcData, (void*)totalData[i], i, _payloadSize, _numSourcePack);

			for (i=0; i<_numTotalPack-_numSourcePack; i++)
			{
				fecPack.fecPackNum = i;
				fecPack.BlockNum = _countFecBlock;
				fecPack.dataSize = _payloadSize;
				#ifdef DEBUG
				fprintf(fp, "send fec pack %d in %d block\n", fecPack.fecPackNum, fecPack.BlockNum);		
				#endif	
				memcpy(fecPackPool, &fecPack, offsite);
				memcpy(&fecPackPool[offsite], totalData[i+_numSourcePack], _payloadSize);
				if (fecUDPSocket->send(fecPackPool, len)!=len)
					cerr << "<QUANTAnet_fecClient_c::send> FEC sending error!" << endl;
			}
			_countFecBlock ++;
		}
	}
	free(data);
	fclose(fp);
}
	

		
		
		
		
		
		
		
