/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <QUANTA/QUANTAnet_http_c.hxx>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

// Added for WINCE
#include <time.h>
 
#ifdef _WIN32_WCE
#include <tempnam.h>
#include <file.h>
#endif

const int QUANTAnet_http_c::AUTO_LOAD = 0;
const int QUANTAnet_http_c::CHECK_BUT_DONT_LOAD = 3;
const int QUANTAnet_http_c::FORCED_LOAD = 1;
const int QUANTAnet_http_c::NO_TIME_OUT = -1;
const int QUANTAnet_http_c::INBUFFER = INBUFFER_CONST;
const int QUANTAnet_http_c::BUFFER = 1024;
 
#ifdef WIN32
static double rint(double x)
{
    double floored_value = floor(x);
    double ceiled_value = ceil(x);
    double floored_diff = x - floor(x);
    double ceiled_diff = ceil(x) - x;

    return (floored_diff >= ceiled_diff) ? ceiled_value : floored_value;
}
#endif

void QUANTAnet_http_c::setRemoteSite(char *site)
{
  strcpy(remoteSite, site);
}

void QUANTAnet_http_c::setRemotePath(char *path)
{
  strcpy(remotePath, path);
}

void QUANTAnet_http_c::setRemoteFile(char *file)
{
  strcpy(remoteFile, file);
}
        
void QUANTAnet_http_c::setLocalPath(char *path)
{
  strcpy(localPath, path);
}

void QUANTAnet_http_c::setLocalFile(char *file)
{
  strcpy(localFile, file);
}

int QUANTAnet_http_c::indexIntoString(char **strlist, int len, char *searchstr)
{
	for(int i=0; i < len; i++) {
		if (strcmp(strlist[i], searchstr) == 0) return i;
	}
	return -1;
}

// If date 1 < date 2 return -1.
// If date 1 > date 2 return 1.
// If date 1 == date 2 return 0.
int QUANTAnet_http_c::compareDate(int date1, int month1, int year1,
		int hour1, int minute1, int second1,
		int date2, int month2, int year2,
		int hour2, int minute2, int second2)
{
	if (year1 < year2) return -1;
	if (year1 > year2) return 1;

	if (month1 < month2) return -1;
	if (month1 > month2) return 1;

	if (date1 < date2) return -1;
	if (date1 > date2) return 1;

	if (hour1 < hour2) return -1;
	if (hour1 > hour2) return 1;

	if (minute1 < minute2) return -1;
	if (minute1 > minute2) return 1;

	if (second1 < second2) return -1;
	if (second1 > second2) return 1;

	return 0;
}


// Given a HTTP 1.0 Date string convert it into component values.
int QUANTAnet_http_c::webDateConvert(char *instring,
		 int *date,
		 int *month,
		 int *year,
		 int *hour,
		 int *minute,
		 int *second)
{

	// HTTP 1.0 has 3 possible formats. I suspect RFC850 will be
	// obsoleted since it encodes year as 2 digits.

	const int RFC1123 = 1;
	const int ASC_TIME = 2;
	const int RFC850 = 3;
	int dateFormat = 0;

	char *monthString[12] = {"Jan","Feb","Mar","Apr","May","Jun",
		"Jul","Aug","Sep","Oct","Nov","Dec"};

	/* NOT USED. KEEP here becoz it shows the example formats.
	char *rfc1123Day[7] = {"Mon,","Tue,","Wed,","Thu,","Fri,","Sat,","Sun,"};
	//Mon, 11 Jan 1999 02:34:32 GMT

	char *rfc850Day[7] = {"Monday,","Tuesday,","Wednesday,","Thursday,",
			   "Friday,","Saturday,","Sunday,"};
	//Monday, 12-Feb-96 02:34:32 GMT

	char *asctimeDay[7] = {"Mon","Tue","Wed","Thu","Fri","Sat","Sun"};
	//Mon Jan 2 02:34:32 1997
	*/

	char *temp;
	int len;

	temp = strtok(instring," ");

	// Grab day of week to figure out the date format.
	len = strlen(temp);

	if (len == 4) dateFormat = RFC1123;
	if (len == 3) dateFormat = ASC_TIME;
	if (len > 4) dateFormat = RFC850;
	if (dateFormat == 0) return -1;

	switch(dateFormat) {
	case RFC1123:
		*date = atoi(strtok(NULL," "));
		*month = indexIntoString(monthString,12,strtok(NULL," ")) + 1;
		*year = atoi(strtok(NULL," "));
		*hour = atoi(strtok(NULL,":"));
		*minute = atoi(strtok(NULL,":"));
		*second = atoi(strtok(NULL," "));
		break;
	case ASC_TIME:
		*month = indexIntoString(monthString,12,strtok(NULL," ")) + 1;
		*date = atoi(strtok(NULL," "));
		*hour = atoi(strtok(NULL,":"));
		*minute = atoi(strtok(NULL,":"));
		*second = atoi(strtok(NULL," "));
		*year = atoi(strtok(NULL," "));
		break;
	case RFC850:
		*date = atoi(strtok(NULL,"-"));
		*month = indexIntoString(monthString,12,strtok(NULL,"-")) + 1;
		*year = atoi(strtok(NULL," ")) + 1900;
		*hour = atoi(strtok(NULL,":"));
		*minute = atoi(strtok(NULL,":"));
		*second = atoi(strtok(NULL," "));
		break;
	}
	return 1;
}

// Given a filename return its modification time in HTTP-date format
int QUANTAnet_http_c::timeDateOfFile(char* filename, char *out, 
		 int *date,
		 int *month,
		 int *year,
		 int *hour,
		 int *minute,
		 int *second)
{

	struct stat buf;

	if (stat(filename, &buf) != 0) return -1;

	time_t tim;
	tim = buf.st_mtime; //tim = buf.st_mtim.tv_sec;

	struct tm *gmt;
	gmt = gmtime(&tim);

	strftime(out,10000,"%a, %d %b %Y %T GMT",gmt);

	*date = gmt->tm_mday;
	*month = gmt->tm_mon + 1;
	*year = gmt->tm_year + 1900;
	*hour = gmt->tm_hour;
	*minute = gmt->tm_min;
	*second = gmt->tm_sec;

	return 0;
}

int QUANTAnet_http_c::grabWEBFile()
{
	// Figure out date and time stamp on local file.
	char dts[128];
	char localTmp[128];
	sprintf(localTmp,"%s/%s", localPath, localFile);

	// Returns 0 if file found, -1 if not found.
	int lDate, lMonth, lYear, lHour, lMinute, lSecond;

	int fileStatus = timeDateOfFile(localTmp , dts,
					&lDate, &lMonth, &lYear,
					&lHour, &lMinute, &lSecond);
	int status;

	// If status is not found or if mode is FORCED_LOAD then load it.
	if ((fileStatus == -1) || (itsMode == FORCED_LOAD)) {
		lDate = lMonth = lYear = lHour = lMinute = lSecond = 0;

		if (remotePath[0]) 
		sprintf(sendBuffer,"GET /%s/%s HTTP/1.0\n\n", remotePath, remoteFile);
		else
		// If remote path is a null string "" then ignore it in the path to
		// send to WEB server.
		sprintf(sendBuffer,"GET /%s HTTP/1.0\n\n", remoteFile);

	}
	else {
		// Otherwise use time stamp.

		if (remotePath[0])
		sprintf(sendBuffer,"GET /%s/%s HTTP/1.0\nIf-Modified-Since: %s\n\n", remotePath, remoteFile, dts);
		else
		// If remote path is a null string "" then ignore it in the path to
		// send to WEB server.
		sprintf(sendBuffer,"GET /%s HTTP/1.0\nIf-Modified-Since: %s\n\n",  remoteFile, dts);
	}

	printf("Sending: %s\n",sendBuffer);

	QUANTAnet_tcpClient_c client;
	client.setTimeOut(timeOutPeriod);

	// Connect to server on port 80.
	if (client.connectToServer(remoteSite,80) < 0) return -2;

	// Send command to server.
	int bs;
	bs = strlen(sendBuffer);
	client.write(sendBuffer, &bs, QUANTAnet_tcpClient_c::BLOCKING);

    //for performance monitoring

    char selfIp[256], remoteIp[256];

    QUANTAnet_socketbase_c::getSelfIP(selfIp);
    client.getRemoteIP(remoteIp);
    pmonitor.setIPs(selfIp, remoteIp);
    pmonitor.setPorts(client.getSelfPort(), client.getRemotePort());
    pmonitor.incrementDataSent(bs);
    
	// Read initial data from server.
	bs = INBUFFER;
	int sockstatus = client.read(receiveBuffer,&bs, QUANTAnet_tcpClient_c::BLOCKING);

	if (sockstatus == QUANTAnet_tcpClient_c::TIMED_OUT){
		client.close();

		if (fileStatus == 0) 
		return -1;
		else return -11;
	}

    //for performance monitoring
    
    pmonitor.incrementDataRead(bs);    
	char dummy[128];

	// Grab the status. If it's 200 it means incoming good data.
	// If 304 then it means file on server is not modified so no download.
	sscanf(receiveBuffer,"%s %d", dummy, &status);

	char *runningBuf;
	int fileSize;
	int skipped, bytesToDump;

	if (status == 200) {

		// If not forced load then must check last modified date of incoming data.
		// We need to do this because sending the "If-Modified-Since" directive
		// does not seem to get a proper response from all servers. So we
		// need to do a manual check ourselves too.
		if (itsMode != FORCED_LOAD) {

			// First find the Last-modified field.
			runningBuf = strstr(receiveBuffer, "Last-modified: ");

// Following code is commented out by Yongjoo Cho since it seems it doesn't
// have any meaning at all...
/*			if (runningBuf == NULL)
				runningBuf - strstr(receiveBuffer,"Last-Modified: ");*/

			// If found then check its date. If not found then assume
			// a load by default.
			if (runningBuf) {

				// Look for end of the Last-Modified line.
				char *toEndOfLine = strstr(runningBuf,"\r\n");

				// Copy that date substring of the Last-Modified line.
				char grabDate[256];
				int len = strlen("Last-Modified: ");
				int lenToCopy = toEndOfLine - (runningBuf + len);

				memcpy(grabDate, runningBuf + len,
				       lenToCopy);

				grabDate[lenToCopy] = '\0'; // End the string properly.

				int mDate, mMonth, mYear,
					mHour, mMinute, mSecond;
				// Convert the date into raw numbers.
				webDateConvert(grabDate, &mDate, &mMonth, &mYear, &mHour,
					       &mMinute, &mSecond);

				// Now compare the remote date with the local date.
				// If the local date is > -1 that means local date
				// is newer so no download.

				printf("Local\n");
				printf("Date: %d/%d %d, Time: %d:%d:%d\n",
				       lMonth, lDate, lYear, lHour, lMinute, lSecond);
				printf("Remote\n");
				printf("Date: %d/%d %d, Time: %d:%d:%d\n",
				       mMonth, mDate, mYear, mHour, mMinute, mSecond);

				if (compareDate(lDate, lMonth, lYear,
						lHour, lMinute, lSecond,
						mDate, mMonth, mYear,
						mHour, mMinute, mSecond) > -1) {
					client.close();
					return 304;
				}
				// Otherwise assume download will continue.

			}
	      
		}

		// First check if you are only interested in seeing if you are older or not.
		if (itsMode == CHECK_BUT_DONT_LOAD){
			client.close();
			return -5;
		}

		// Ok we seriously want to grab the file so continue.... 

		// Grab the content length.
		runningBuf = strstr(receiveBuffer,"Content-length: ");
		if (runningBuf == NULL) 
			runningBuf = strstr(receiveBuffer,"Content-Length: ");

		if (runningBuf == NULL) {
			client.close();
			return -6;
		}

		sscanf(runningBuf,"%s %d", dummy, &fileSize);

		printf("File size: %d bytes\n", fileSize);

		// Skip past the header data which is delimited
		// by two carriage returns and new lines
		runningBuf = strstr(receiveBuffer,"\r\n\r\n");
		runningBuf+=4;

		FILE *tempFile;
		char *tempFileName;
		tempFileName = tempnam(localPath,"_QUANTAnet_http_c");
		tempFile = fopen(tempFileName,"wb");
		if (!tempFile) {
			client.close();
			return -4;
		}

		// Dump the stuff in buffer so far.
		skipped = runningBuf - receiveBuffer;
		bytesToDump = min(fileSize, INBUFFER - skipped);
		fwrite(runningBuf, bytesToDump, 1, tempFile);

		// Figure out how much is left to grab and dump.
		int bytesLeft;
		bytesLeft = fileSize - bytesToDump;

		while(bytesLeft) {
			bytesToDump = min(bytesLeft, INBUFFER);
			sockstatus = client.read(receiveBuffer,&bytesToDump, QUANTAnet_tcpClient_c::BLOCKING);
			if (sockstatus == QUANTAnet_tcpClient_c::TIMED_OUT) {
				client.close();
				unlink(tempFileName);
				return -1;
			}

            //for performance monitoring
             pmonitor.incrementDataRead(bytesToDump); 
			// Need to check for error condition in read for timeouts.

			bytesLeft -= bytesToDump;
			fwrite(receiveBuffer, bytesToDump, 1, tempFile);

			printf("");
			printf("Loaded: %g%%",rint(float(fileSize - bytesLeft)/ float(fileSize) * 100.0));
			fflush(stdout);
 
			// If n is zero then server has broken the connection usually
			// meaning the data is done transmitting.
			if (sockstatus == QUANTAnet_tcpClient_c::CONNECTION_TERMINATED) break;
		}
		printf("\n");
		fclose(tempFile);
		int renameStatus = rename(tempFileName,localTmp);
		if (renameStatus == -1) {
			status= -3;
			unlink(tempFileName);
		}

		delete tempFileName;

	} else {
		client.close();
	}

	return status;

}

void  QUANTAnet_http_c::showStats(char* streamInfo, char* comment)
{
    pmonitor.showStats(streamInfo, comment);    
}

int  QUANTAnet_http_c::logStats(char* streamInfo, char* comment, FILE* filePtr)
{
    return pmonitor.logStats(streamInfo, comment, filePtr);
}

int QUANTAnet_http_c::sendStats(char* streamInfo, char* comment)
{
    return pmonitor.sendStats(streamInfo, comment);    
}

int QUANTAnet_http_c::initSendStats(char* monitorClientIP, int port)
{
    return pmonitor.initSendStats(monitorClientIP, port);
}

void QUANTAnet_http_c::exitSendStats()
{
    pmonitor.exitSendStats();
}



