/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <QUANTA/QUANTAnet_mcast_c.hxx>
#include <QUANTA/perfmon.hxx>

#include <stdio.h>
#include <string.h>
#ifndef WIN32
#include <sys/socket.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#endif /* !WIN32 */

#ifdef _WIN32_WCE
#include <errno.h>
#endif

const int QUANTAnet_mcast_c::maxTTL = 128;
const int QUANTAnet_mcast_c::OK = 1;
const int QUANTAnet_mcast_c::FAILED = 0;

#ifdef sun
long gethostid(void);
#endif

#ifdef WIN32
#define MAXHOSTNAME   128   /* Maximum length of a DNS hostname */

// This function is excerpted from "Windows Sockets Network Programming"
// by Bob Quinn and Dave Shute published from Addison-Wesley.
static long gethostid(void)
{
	char szLclHost [MAXHOSTNAME];
	LPHOSTENT lpstHostent;
	SOCKADDR_IN stLclAddr;
	SOCKADDR_IN stRmtAddr;
	int nAddrSize = sizeof(SOCKADDR);
	SOCKET hSock;
	int nRet;

	/* Init local address (to zero) */
	stLclAddr.sin_addr.s_addr = INADDR_ANY;

	/* Get the local hostname */
	nRet = gethostname(szLclHost, MAXHOSTNAME); 
	if (nRet != SOCKET_ERROR) {
		/* Resolve hostname for local address */
		lpstHostent = gethostbyname((LPSTR)szLclHost);
		if (lpstHostent)
			stLclAddr.sin_addr.s_addr = *((u_long FAR*) (lpstHostent->h_addr));
	} 

	/* If still not resolved, then try second strategy */
	if (stLclAddr.sin_addr.s_addr == INADDR_ANY) {
		/* Get a UDP socket */
		hSock = socket(AF_INET, SOCK_DGRAM, 0);
		if (hSock != INVALID_SOCKET)  {
			/* Connect to arbitrary port and address (NOT loopback) */
			stRmtAddr.sin_family = AF_INET;
			stRmtAddr.sin_port   = htons(IPPORT_ECHO);
			stRmtAddr.sin_addr.s_addr = inet_addr("128.127.50.1");
			nRet = connect(hSock, (LPSOCKADDR)&stRmtAddr,
						   sizeof(SOCKADDR));
			if (nRet != SOCKET_ERROR) {
				/* Get local address */
				getsockname(hSock, (LPSOCKADDR)&stLclAddr, 
							(int FAR*)&nAddrSize);
			}
			closesocket(hSock);   /* we're done with the socket */
		}
	}
	return (stLclAddr.sin_addr.s_addr);
}
#endif

QUANTAnet_mcast_c :: QUANTAnet_mcast_c ()
{
	sockfd = 0;
	enableInstr = 0;
}

void QUANTAnet_mcast_c::enableInstrumentation()
{
	enableInstr = 1;
}

void QUANTAnet_mcast_c::disableInstrumentation()
{
	enableInstr = 0;
}

void QUANTAnet_mcast_c::printInfo()
{
	printf("\nQUANTAPLUS_MCAST_C Info:\n");

	char name[256];

	QUANTAnet_socketbase_c::getSelfIP(name);
	printf("Self Address: %s Port: %u\n",name,
	       getSelfPort());

	getSendIP(name);
	printf("Multicast Send Address: %s Port: %u\n",
	       name,
	       getSendPort());

	getReceiveIP(name);
	printf("Receive Address: %s Port: %u\n\n",
	       name,
	       getReceivePort());
}

void QUANTAnet_mcast_c::close()
{
#ifdef WIN32
  ::closesocket(sockfd);
#else
  ::close(sockfd);
#endif
}

int QUANTAnet_mcast_c :: init(char *QUANTAnet_mcast_cGroup, int port, int mttl)
{
	/* create what looks like an ordinary UDP socket */
/* socket function in Windows is returns a unsigned value. Thus, checking 
   whether the result is less than 0 cannot give us accurate result of the
   function call. We have to compare the result with INVALID_SOCKET, instead. */
#ifdef WIN32 
	if ((sockfd=socket(AF_INET,SOCK_DGRAM,0)) == INVALID_SOCKET) {
		printf("QUANTAnet_mcast_c::init: Error in Socket opening.\n");
		return QUANTAnet_mcast_c::FAILED;
	}
#else
	if ((sockfd=socket(AF_INET,SOCK_DGRAM,0)) < 0) {
		printf("QUANTAnet_mcast_c::init: Error in Socket opening.\n");
		return QUANTAnet_mcast_c::FAILED;
	}
#endif

	/* set up destination address */
	memset(&receiveAddress,0,sizeof(receiveAddress));
	receiveAddress.sin_family=AF_INET;
	receiveAddress.sin_addr.s_addr=htonl(INADDR_ANY); /* N.B.: differs from sender */
	receiveAddress.sin_port=htons(port);

	memset(&sendAddress,0,sizeof(sendAddress));
	sendAddress.sin_family=AF_INET;
	sendAddress.sin_addr.s_addr=inet_addr(QUANTAnet_mcast_cGroup);
	sendAddress.sin_port=htons(port);
     
     
	/* bind to receive address */
	if (bind(sockfd,(struct sockaddr *) &receiveAddress,sizeof(receiveAddress)) < 0) {
		printf("QUANTAnet_mcast_c::init: Error in bind.\n");
		return QUANTAnet_mcast_c::FAILED;
	}

	char mcloop = 0;
	/* Disable loopback of QUANTAnet_mcast_c packets so that you don't
	 * get a copy of what you sent.
	 */
	if (setsockopt(sockfd, IPPROTO_IP, IP_MULTICAST_LOOP, &mcloop, sizeof(mcloop)) < 0){
		printf("QUANTAnet_mcast_c::init: setsockopt QUANTAnet_mcast_c cannot set loopback.\n");
	}

	ttl = char(mttl);
	char QUANTAnet_mcast_c_ttl = ttl;

	if (setsockopt(sockfd, IPPROTO_IP, IP_MULTICAST_TTL, &QUANTAnet_mcast_c_ttl, 
		sizeof(QUANTAnet_mcast_c_ttl)) < 0)	{
		printf("QUANTAnet_mcast_c::init: setsockopt QUANTAnet_mcast_c cannot set Time To Live.");
	}

	/* use setsockopt() to request that the kernel join a multicast group */
	mreq.imr_multiaddr.s_addr=inet_addr(QUANTAnet_mcast_cGroup);
	mreq.imr_interface.s_addr=htonl(INADDR_ANY);
// Winsock requires the 4th parameter as const char*, and Visual C++ shows
// an error that says passing different type of parameter.
#if (defined(WIN32)) || (defined(sun))
	if (setsockopt(sockfd,IPPROTO_IP,IP_ADD_MEMBERSHIP,(char*) &mreq,sizeof(mreq)) < 0) {
		printf("QUANTAnet_mcast_c::init: setsockopt cannot add membership.\n");
		return QUANTAnet_mcast_c::FAILED;
	}
#else
	if (setsockopt(sockfd,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq)) < 0) {
		printf("QUANTAnet_mcast_c::init: setsockopt cannot add membership.\n");
		return QUANTAnet_mcast_c::FAILED;
	}
#endif

	return QUANTAnet_mcast_c::OK;

}

int QUANTAnet_mcast_c :: send(char *message, int length) 
{

	char* newmsg = message;
	int newlength = length;
	if (enableInstr) {
		newmsg = _quanta_allocMemForPerformanceMonitoredMesg(length, &newlength);
		if (newmsg == NULL) return 0;

		_quanta_prependPerformanceMonitoringHeader(newmsg, message, length);
	}

	int n = sendto(sockfd, newmsg, newlength, 0, (sockaddr *)&sendAddress,
		       sizeof(sendAddress));

	if (n > 0) incrementDataSent(n);
	if (enableInstr) {
		delete newmsg;
		int realsize = n-_quanta_getSizeOfPerformanceMonitoringHeader();
		if (realsize > 0)
			return realsize;
		else
			return 0;
	}
	else return n;

	/*
	int n;

	n = sendto(sockfd,message,size,0,(struct sockaddr *) &sendAddress,sizeof(sendAddress));
	if (n>0) incrementDataSent(n);
	return n;
	*/
}

int QUANTAnet_mcast_c :: receive(char *message, int length)
{
	char *newmsg = message;
	int newlength = length;

	// if instrumentation enabled
	if (enableInstr) {
		newmsg = _quanta_allocMemForPerformanceMonitoredMesg(length, &newlength);
		if (newmsg == NULL) return 0;
	}

    #if (defined(WIN32) || defined(sun))
        int clilen;
    #else
        socklen_t clilen;
    #endif

	clilen = sizeof(receiveAddress);
	int n;

	n = recvfrom(sockfd, newmsg, newlength, 0, (struct sockaddr *) &receiveAddress,
		     & clilen);
	if (n > 0) {
		incrementDataRead(n);

		// if instrumentation enabled
		if (enableInstr) {
			double remTimeStamp, latency;
			int realSize;
			_quanta_unprependPerformanceMonitoringHeader(&remTimeStamp, &latency, newmsg, n, message, &realSize);

			delete newmsg;

			setInstantLatency(latency);
			return realSize;
		}
	} else {
#ifdef WIN32
		if (n == SOCKET_ERROR) errno = WSAGetLastError();
#endif
		if (enableInstr) delete newmsg;
	}
	
	return n;

	/*
	int  addrlen=sizeof(receiveAddress);
	int n;

	n =recvfrom(sockfd,message,bufferSize,0, (struct sockaddr *) &receiveAddress,&addrlen);
	if (n>0) incrementDataRead(n);

	return n;
	*/

}


void QUANTAnet_mcast_c :: makeNonBlocking() 
{
	if (sockfd < 0) {
		printf("QUANTAnet_mcast_c::MakeNonBlocking: No connection open yet.\n");
	}
#ifdef __sgi
	fcntl(sockfd, F_SETFL, FNDELAY);
#endif
#ifdef linux
	fcntl(sockfd, F_SETFL, O_NONBLOCK);
#endif
#ifdef sun
	fcntl(sockfd, F_SETFL, O_NDELAY);
#endif
#ifdef WIN32
    u_long nonBlockingMode = 1L;

    ioctlsocket(sockfd, FIONBIO, &nonBlockingMode);
#endif
}

unsigned long QUANTAnet_mcast_c :: getReceiveIP() {
	return receiveAddress.sin_addr.s_addr;
}

void QUANTAnet_mcast_c :: getReceiveIP(char *clientIP) {
	ipNumToString(getReceiveIP(), clientIP);
}

unsigned long QUANTAnet_mcast_c :: getSendIP() {
	return sendAddress.sin_addr.s_addr;
}

void QUANTAnet_mcast_c :: getSendIP(char *clientIP) {
	ipNumToString(getSendIP(), clientIP);
}

unsigned short  QUANTAnet_mcast_c :: getSelfPort() {
	struct sockaddr_in name;
    #if (defined(WIN32) || defined(sun))
        int size;
    #else
        socklen_t size;
    #endif
	size = sizeof(name);
	getsockname (sockfd, (sockaddr *)&name, &size);
	return ntohs(name.sin_port);
}


unsigned short QUANTAnet_mcast_c :: getSendPort() {
	return ntohs(sendAddress.sin_port);
}

unsigned short QUANTAnet_mcast_c :: getReceivePort() {
	return ntohs(receiveAddress.sin_port);
}

