/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <QUANTAnet_rbudpReceiver_c.hxx>

#include <stdarg.h>

// If you want to output debug info on terminals when running, put
//  fprintf(stderr, __VA_ARGS__);\
//  fprintf(stderr, "\n");\
// into TRACE_DEBUG definition.
// If you want to log the output into a log file, uncomment DEBUG, and put
//  fprintf(log, __VA_ARGS__);\
//  fprintf(log, "\n");\
// into TRACE_DEBUG definition.

//#define DEBUG

/*#define TRACE_DEBUG(...) do { \
  fprintf(stderr, __VA_ARGS__);\
  fprintf(stderr, "\n");\
  }  while(0)
*/

inline void TRACE_DEBUG( char *format, ...)
{
	va_list arglist;
	va_start( arglist, format);
	vfprintf( stderr, format, arglist);
	fprintf( stderr, "\n");
	va_end(arglist);
}

void QUANTAnet_rbudpReceiver_c::receive(void * buffer, int bufSize, int packetSize)
{
	int done = 0;
	struct timeval curTime;
	
	gettimeofday(&curTime, NULL);
	initReceiveRudp(buffer, bufSize, packetSize);	
	initErrorBitmap();
	while (!done)
	{
		// receive UDP packetsq		TRACE_DEBUG("receiving UDP packets");
		reportTime(curTime);
			
		udpReceiveReadv();  
	
		reportTime(curTime);

		gettimeofday(&curTime, NULL);
		TRACE_DEBUG("Current time: %d %ld", curTime.tv_sec, curTime.tv_usec);
		
		if (updateHashTable() == 0)
			done = 1;
		
		TRACE_DEBUG("Error: %d, Loss rate: %f", updateHashTable(), (double)updateHashTable()/(double)totalNumberOfPackets);
		
		// Send back Error bitmap
		if (writen(tcpSockfd, errorBitmap, sizeofErrorBitmap) != sizeofErrorBitmap)

                {
                        perror("tcp send");
                        exit(1);
                }			
	}
	free(errorBitmap);
	free(hashTable);
}	

void QUANTAnet_rbudpReceiver_c::udpReceiveReadv()
{
	int done;
	long seqno, packetno;
	struct timeval timeout;
	fd_set rset;
	int maxfdpl;
	done = 0; packetno=0;
	
	timeout.tv_sec = 100;
	timeout.tv_usec = 0;
	#define QMAX(x, y) ((x)>(y)?(x):(y))
	maxfdpl = QMAX(udpSockfd, tcpSockfd) + 1;
	FD_ZERO(&rset);
	while (!done)
	{
		// These two FD_SET cannot be put outside the while, don't why though
		FD_SET(udpSockfd, &rset);
		FD_SET(tcpSockfd, &rset);
		select(maxfdpl, &rset, NULL, NULL, &timeout);

		// receiving a packet
		if (FD_ISSET(udpSockfd, &rset))
		{
			// set seqno to expected sequence number
			seqno = hashTable[packetno];
			//fprintf(log, "expect %d\t", seqno);
			iovRecv[1].iov_base = (char *)mainBuffer+(seqno*payloadSize);
			iovRecv[1].iov_len = payloadSize;
			if (recvmsg(udpSockfd, &msgRecv, 0) < 0)
			{
				perror("recvmsg error\n");
				exit(1);
			}
			
			// make sure it is the expected packet
			if (recvHeader.seq == seqno)
			{
				updateErrorBitmap(recvHeader.seq);
				// next expected packet
				packetno ++;
			}
			// if some packets lost before this packet, move this packet backward
			else if (recvHeader.seq > seqno)
			{
				bcopy(iovRecv[1].iov_base, (char *)mainBuffer+(recvHeader.seq*payloadSize) , iovRecv[1].iov_len);
				updateErrorBitmap(recvHeader.seq);
				// missed some packets, next expected packet
				do packetno ++;
				while((recvHeader.seq != hashTable[packetno]) || (packetno > remainNumberOfPackets));
				if (packetno > remainNumberOfPackets)
				{	
					printf("recv error\n");
				}
				packetno ++;
			}
			//TRACE_DEBUG("received %d", recvHeader.seq);
		}
		//receive end of UDP signal
		else if (FD_ISSET(tcpSockfd, &rset))
		{
			done = 1;
			TRACE_DEBUG("received TCP signal");
                	readn(tcpSockfd, (char *)&endOfUdp, sizeof(struct _endOfUdp));
		}
		else // time out
		{
			done = 1;
		}
	}		
}

void QUANTAnet_rbudpReceiver_c::udpReceive()
{
	int done, actualPayloadSize;
	long seqno;
	struct timeval start;
	char *msg = (char *) malloc(packetSize);	
	struct timeval timeout;
	fd_set rset;
	int maxfdpl;
	done = 0; seqno = 0;
	
	timeout.tv_sec = 200;
	timeout.tv_usec = 0;
	#define QMAX(x, y) ((x)>(y)?(x):(y))
	maxfdpl = QMAX(udpSockfd, tcpSockfd) + 1;
	FD_ZERO(&rset);
	gettimeofday(&start, NULL);
	while (!done)
	{
		// These two FD_SET cannot be put outside the while, don't why though
		FD_SET(udpSockfd, &rset);
		FD_SET(tcpSockfd, &rset);
		select(maxfdpl, &rset, NULL, NULL, &timeout);

		// receiving a packet
		if (FD_ISSET(udpSockfd, &rset))
		{
			int temp = recv(udpSockfd, msg, packetSize, 0);
			bcopy(msg, &recvHeader, sizeof(struct _rbudpHeader));
			seqno = recvHeader.seq;

			// If the packet is the last one, 
			if (seqno < totalNumberOfPackets - 1)
			{
				actualPayloadSize = payloadSize;
			}
			else
			{
				actualPayloadSize = lastPayloadSize; 
			}

			bcopy(msg+headerSize, (char *)mainBuffer+(seqno*payloadSize) , actualPayloadSize);

			updateErrorBitmap(seqno);
		}
		//receive end of UDP signal
		else if (FD_ISSET(tcpSockfd, &rset))
		{
			done = 1;
                	readn(tcpSockfd, (char *)&endOfUdp, sizeof(struct _endOfUdp));
		}
		else // time out
		{
			done = 1;
		}
	}		
	free(msg);

}

int QUANTAnet_rbudpReceiver_c::getfile(char * origFName, char * destFName, int packetSize)
{
  // Send getfile message.
   if (writen(tcpSockfd, origFName, SIZEOFFILENAME) != SIZEOFFILENAME)
   {
     {
       perror("tcp send");
       exit(1);
     }			
   }

   int filesize;
   int n = readn(tcpSockfd, (char *)&filesize, sizeof(filesize));
   if (n < 0) {
     printf("read error.\n");
     return(FAILED);
   }

   filesize = ntohl(filesize);
   printf("The size of the file is %d.\n", filesize);
   
   int fd = open(destFName, O_RDWR|O_CREAT|O_TRUNC, 0666);
   ftruncate(fd, filesize);
   
   char *buf;
   buf = (char *)mmap(NULL, filesize, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
   if (buf == MAP_FAILED) {
     printf("mmap failed.\n");
     return FAILED;
   }

   receive(buf, filesize, packetSize);

   munmap(buf, filesize);
   ::close(fd);
   
   return SUCCESS; 
}

/* sendRate: Mbps */
int QUANTAnet_rbudpReceiver_c::initReceiveRudp(void* buffer, int bufSize, int pSize)
{
	int i;
	mainBuffer = (char *)buffer;
	dataSize = bufSize;
	payloadSize = pSize;
	headerSize = sizeof(struct _rbudpHeader);
	packetSize = payloadSize + headerSize;
	isFirstBlast = 1;

	if (dataSize % payloadSize == 0)
	{
		totalNumberOfPackets = dataSize/payloadSize;
		lastPayloadSize = payloadSize;
	}
	else
	{
		totalNumberOfPackets = dataSize/payloadSize + 1; /* the last packet is not full */
		lastPayloadSize = dataSize - payloadSize * (totalNumberOfPackets - 1);
	}
	remainNumberOfPackets = totalNumberOfPackets;
	sizeofErrorBitmap = totalNumberOfPackets/8 + 2;
	errorBitmap = (char *)malloc(sizeofErrorBitmap);
	hashTable = (long *)malloc(totalNumberOfPackets * sizeof(long));
	
	TRACE_DEBUG("totalNumberOfPackets: %d", totalNumberOfPackets);

	if (errorBitmap == NULL)
	{
		printf("malloc errorBitmap failed\n");
		return (-1);
	}
	if (hashTable == NULL)
	{
		printf("malloc hashTable failed\n");
		return (-1);
	}

	/* Initialize the hash table */
	for (i=0; i<totalNumberOfPackets; i++)
	{
		hashTable[i] = i;
	}
	return 0;
}


void QUANTAnet_rbudpReceiver_c::init(char *remoteHost)
{
	int n;

#ifdef DEBUG
	log = fopen("rbudprecv.log","w");
#endif

	passiveUDP(remoteHost);

	if (!hasTcpSock) 
	{
		TRACE_DEBUG("try to connect the sender via TCP ...");
		n = connectTCP(remoteHost);
		if (n < 0)
		{
			printf("connecting TCP failed, make sure the sender has been started\n");
			exit(1);
		}
		TRACE_DEBUG("tcp connected.");
	}
       
	msgRecv.msg_name = NULL;
	msgRecv.msg_namelen = 0;
	msgRecv.msg_iov = iovRecv;
	msgRecv.msg_iovlen = 2;
	
	iovRecv[0].iov_base = reinterpret_cast<char*>(&recvHeader);
	iovRecv[0].iov_len = sizeof(struct _rbudpHeader);
}
	

void QUANTAnet_rbudpReceiver_c::close()
{
	if (!hasTcpSock)
		::close(tcpSockfd);
	::close(udpSockfd);
#ifdef DEBUG
	fclose(log);
#endif
}

