/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <QUANTAnet_rbudpSender_c.hxx>

#include <stdarg.h>

// If you want to output debug info on terminals when running, put
//  fprintf(stderr, __VA_ARGS__);\
//  fprintf(stderr, "\n");\
// into TRACE_DEBUG definition.
// If you want to log the output into a log file, uncomment DEBUG, and put
//  fprintf(log, __VA_ARGS__);\
//  fprintf(log, "\n");\
// into TRACE_DEBUG definition.

//#define DEBUG

/*#define TRACE_DEBUG(...) do { \
  fprintf(stderr, __VA_ARGS__);\
  fprintf(stderr, "\n");\
  }  while(0)
*/

inline void TRACE_DEBUG( char *format, ...)
{
	va_list arglist;
	va_start( arglist, format);
	vfprintf( stderr, format, arglist);
	fprintf(stderr, "\n");
	va_end(arglist);
}

void QUANTAnet_rbudpSender_c::send(void * buffer, int bufSize, int sendRate, int packetSize)
{
	int done = 0;
	struct timeval curTime;
	double srate;
	gettimeofday(&curTime, NULL);
	initSendRudp(buffer, bufSize, sendRate, packetSize);	
	while (!done)
	{
		// blast UDP packets
		TRACE_DEBUG("sending UDP packets");
		reportTime(curTime);
		udpSendWritev();

		srate = (double) remainNumberOfPackets * payloadSize * 8 / (double) reportTime(curTime);	
		TRACE_DEBUG("real sending rate in this send is %f", srate);
		
		// send end of UDP signal
		TRACE_DEBUG("send to socket %d an end signal.", tcpSockfd); 
		writen(tcpSockfd, (char *)&endOfUdp, sizeof(endOfUdp));
		endOfUdp.round ++;

		reportTime(curTime);
		gettimeofday(&curTime, NULL);
		TRACE_DEBUG("Current time: %d %ld", curTime.tv_sec, curTime.tv_usec);

		// receive error list
		TRACE_DEBUG("waiting for error bitmap");
		int n = readn(tcpSockfd, errorBitmap, sizeofErrorBitmap);
	        if (n<0)
	        {
        	        perror("read");
                	exit(1);
	        }
		
		if ((unsigned char)errorBitmap[0] == 1)
                {
			done = 1;
                        remainNumberOfPackets = 0;
			TRACE_DEBUG("done.");
                }
		else
		{
			remainNumberOfPackets = updateHashTable();
		}
		
                if (isFirstBlast)
                {
                        isFirstBlast = 0;
                        double lossRate = (double)remainNumberOfPackets / (double)totalNumberOfPackets;
//                        if (remainNumberOfPackets > 0)
//                                usecsPerPacket = (int) ((double)usecsPerPacket / (1.0 - lossRate - 0.05));
			TRACE_DEBUG("loss rate: %f", lossRate);
                        TRACE_DEBUG("usecsPerPacket updated to %d", usecsPerPacket);
                }
	}
	free(errorBitmap);
	free(hashTable);
}	

void QUANTAnet_rbudpSender_c::udpSendWritev()
{
	int i, done, actualPayloadSize;
	struct timeval start, now;

	done = 0; i = 0;
	gettimeofday(&start, NULL);
	while(!done)
	{
		gettimeofday(&now, NULL);
		if (USEC(&start, &now) < usecsPerPacket * i)
		{
		// busy wait or sleep 
	//		usleep(1);
		}
		else
		{
                        // last packet is probably smaller than regular packets 
                        // we have to pad the last packet, make it the same length 
                        // as regular packets 
			if (hashTable[i] < totalNumberOfPackets - 1)
			{
				actualPayloadSize = payloadSize;
			}
			else
			{
				actualPayloadSize = lastPayloadSize; 
			}
			sendHeader.seq = hashTable[i];
			
			iovSend[1].iov_base = (char*)mainBuffer+(sendHeader.seq*payloadSize);
			iovSend[1].iov_len = actualPayloadSize;
			//TRACE_DEBUG("sent %d, %d %d", sendHeader.seq, now.tv_sec, now.tv_usec);
			if (sendmsg(udpSockfd, &msgSend, 0) < 0)
                        {
                                perror("sendmsg");
                                exit(1);
                        }
			i++;
			if (i >= remainNumberOfPackets)
				done = 1;
		}
	}
}

void QUANTAnet_rbudpSender_c::udpSend()
{
	int i, done, actualPayloadSize;
	struct timeval start, now;
	char *msg = (char *) malloc(packetSize);	

	done = 0; i = 0;
	gettimeofday(&start, NULL);
	while(!done)
	{
		gettimeofday(&now, NULL);
		if (USEC(&start, &now) < usecsPerPacket * i)
		{
		// busy wait or sleep 
	//		usleep(1);
		}
		else
		{
                        // last packet is probably smaller than regular packets 
                        // we have to pad the last packet, make it the same length 
                        // as regular packets 
			if (hashTable[i] < totalNumberOfPackets - 1)
			{
				actualPayloadSize = payloadSize;
			}
			else
			{
				actualPayloadSize = lastPayloadSize; 
			}
			sendHeader.seq = hashTable[i];
			bcopy(&sendHeader, msg, headerSize);
			bcopy((char*)((char*)mainBuffer+(sendHeader.seq*payloadSize)), msg+headerSize, actualPayloadSize);
			
//			TRACE_DEBUG("sent %d", sendHeader.seq);
                        if (sendto(udpSockfd, msg, actualPayloadSize + headerSize, 0, (const struct sockaddr *)&udpServerAddr, sizeof(udpServerAddr)) < 0 )
                        {
                                perror("sendto");
                                exit(1);
                        }
			i++;
			if (i >= remainNumberOfPackets)
				done = 1;
		}
	}
	free (msg);
}

int QUANTAnet_rbudpSender_c::sendfile(int sendRate, int packetSize)
{
  // Receive the getfile message
  char fname[SIZEOFFILENAME];
  int n = readn(tcpSockfd, fname, SIZEOFFILENAME);
  if (n < 0) {
    printf("read error.\n");
    return(FAILED);
  }

  printf("Send file %s\n", fname);

  struct stat filestat;
  if (stat(fname, &filestat) < 0) {
    printf("stat error.\n");
    return(FAILED);
  }

  int filesize = filestat.st_size;
  printf("The size of the file is %d\n", filesize);

  int nfilesize = htonl(filesize);

  // Send the file size to the receiver.
   if (writen(tcpSockfd, (char *)&nfilesize, sizeof(nfilesize)) != sizeof(nfilesize))
   {
     {
       printf("tcp send failed.\n");
       return(FAILED);
     }			
   }

   int fd = open(fname, O_RDONLY);
   if (fd < 0) {
     printf("open file failed.\n");
     return(FAILED);
   }
   char *buf = (char *)mmap(NULL, filesize, PROT_READ, MAP_SHARED, fd, 0);

   send(buf, filesize, sendRate, packetSize);
   
   munmap(buf, filesize);
   ::close(fd);

   return(SUCCESS);
}

/* sendRate: Kbps */
int QUANTAnet_rbudpSender_c::initSendRudp(void* buffer, int bufSize, int sRate, int pSize)
{
	int i;
	mainBuffer = (char *)buffer;
	dataSize = bufSize;
	sendRate = sRate;
	payloadSize = pSize;
	headerSize = sizeof(struct _rbudpHeader);
	packetSize = payloadSize + headerSize;
	usecsPerPacket = 8 * payloadSize * 1000 / sendRate;
	isFirstBlast = 1;

	if (dataSize % payloadSize == 0)
	{
		totalNumberOfPackets = dataSize/payloadSize;
		lastPayloadSize = payloadSize;
	}
	else
	{
		totalNumberOfPackets = dataSize/payloadSize + 1; /* the last packet is not full */
		lastPayloadSize = dataSize - payloadSize * (totalNumberOfPackets - 1);
	}

	remainNumberOfPackets = totalNumberOfPackets;
	sizeofErrorBitmap = totalNumberOfPackets/8 + 2;
	errorBitmap = (char *)malloc(sizeofErrorBitmap);
	hashTable = (long *)malloc(totalNumberOfPackets * sizeof(long));
	
	endOfUdp.round = 0;
	strncpy(endOfUdp.end, "EEE", 3);
	
	TRACE_DEBUG("totalNumberOfPackets: %d", totalNumberOfPackets);
	TRACE_DEBUG("usecsPerPacket: %d", usecsPerPacket);
	
	if (errorBitmap == NULL)
	{
		printf("malloc errorBitmap failed\n");
		return (-1);
	}
	if (hashTable == NULL)
	{
		printf("malloc hashTable failed\n");
		return (-1);
	}

	/* Initialize the hash table */
	for (i=0; i<totalNumberOfPackets; i++)
	{
		hashTable[i] = i;
	}
	return 0;
}


void QUANTAnet_rbudpSender_c::init(char *remoteHost)
{
  openSockets(remoteHost);
  listenAndInit();
}
	
void QUANTAnet_rbudpSender_c::openSockets(char *remoteHost)
{
#ifdef DEBUG
	log = fopen("rbudpsend.log", "w");
#endif
  connectUDP(remoteHost);
  if (!hasTcpSock) {
    initTCPServer();
  }
}

void QUANTAnet_rbudpSender_c::listenAndInit()
{
  if(!hasTcpSock) {
    listenTCPServer();
  }

#if defined (__sgi)	
//	msgSend.msg_name = (caddr_t)&udpServerAddr.sin_addr;
        msgSend.msg_name = (caddr_t)&udpServerAddr;
#else
//        msgSend.msg_name = (void *)&udpServerAddr.sin_addr;
        msgSend.msg_name = reinterpret_cast<char*>(&udpServerAddr);
#endif 
	msgSend.msg_namelen = sizeof(udpServerAddr);
	msgSend.msg_iov = iovSend;
	msgSend.msg_iovlen = 2;
	
	iovSend[0].iov_base = reinterpret_cast<char*>(&sendHeader);
	iovSend[0].iov_len = sizeof(struct _rbudpHeader);
}

void QUANTAnet_rbudpSender_c::close()
{
  if (!hasTcpSock) {
    ::close(tcpSockfd);
    ::close(listenfd);
  }
  ::close(udpSockfd);
#ifdef DEBUG
  fclose(log);
#endif
}

