/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <QUANTA/QUANTAnet_remoteFileIO64_c.hxx>
#include <QUANTA/QUANTAts_thread_c.hxx>
#include <sys/stat.h>
#include <math.h>

#ifdef WIN32
#include <stdlib.h>
#define MAXPATHLEN	MAX_PATH
#else
#include <sys/param.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

#if defined(linux) 
#define fseek64 fseek
#define ftell64 ftell
#endif
/*
Data stream format from client to server:

Upload:
'u' + 
filenamelen[int] + 
filename +
datasize[int]
data

Download:
'd' +
filenamelen[int] +
filename

Check Remote Timestamp:
'c' +
filenamelen[int] +
filename

Data Stream from server to client:

From server to client- in response to download message:
'd' +
(filesize[int], time-stamp[double]) +
data

From server to client in response to check for remote timestamp:
'c' + 
date[6 integers], timestamp[double]

From Server to Client (For all commands)

File Not found:  'n'
Invalid command: 't'

*/

struct ServerData
{
    char* neighborServerName;
	QUANTAnet_extendedTcpClient_c* client;
	QUANTAnet_tcpServer_c* server;
	int store;
	int port;

	~ServerData() { if (client) { client->close(); delete client; } }
};

/* Minimized client version of remote file IO client that can
   only download.
*/
class QUANTAnet_ServerClient64_c {
	public:

	/// Status ok
	static const int OK/* = 1*/;

	/// Status failed
	static const int FAILED/* = 0*/;

	/// Cannot connect to remote server.
	static const int CANNOT_CONNECT/* = 2*/;

	/// Cannot open local file for reading.
	static const int CANNOT_READ_LOCAL_FILE/* = 3*/;

	/// CAnnot write to local file.
	static const int CANNOT_WRITE_LOCAL_FILE/* = 3*/;

	/// Received invalid command from server.
	static const int INVALID_COMMAND/* = 4*/;

	/// Memory allocation error.
	static const int MEM_ALLOC_ERR/* = 5*/;

	/// Remote file error.
	static const int FILE_ERROR/* = 6*/;

	/// Remote file error.
	static const int DOWNLOAD/* = 7*/;

	QUANTAnet_ServerClient64_c(QUANTAnet_tcpServer_c* Server,
						       int ServerIP = 0, int ServerPort = 0);
	~QUANTAnet_ServerClient64_c();

	/// Set port number to connect to.
	void setRemotePort(int port = QUANTAnet_remoteFileIOServer64_c::DEFAULT_PORT) {
	    neighborPortNumber = port;
	}

	/** Download a file from the remote server. If the remote file is not found or
	    a network error occurs, this function will return with a failed.
	    @param ipAddress is IP address of remote server.
	    @param localFileName is local file write to.
	    @param remoteFileName is remote file to get.
	    @return OK, FAILED, CANNOT_WRITE_LOCAL_FILE, INVALID_COMMAND, CANNOT_CONNECT
	*/
	int initDownload(char *ipAddress, char* localFilename, char* remoteFileName);
	int readBlock(char*, int nBytes);
    int getFileSize(int64& fileSize) { return client->readInt64(fileSize, QUANTAnet_tcpClient_c::BLOCKING); }

	int getRemoteDateTime(char *ipAddress, char* remoteFileName, int *remDay, 
		        int *remMon, int *remYear, int* remHour, int* remMin, int* remSec);

private:
	int neighborPortNumber;
	int store;
	int serverIPNumber;
	int serverPortNumber;
	char neighborServerName[MAXPATHLEN];
	QUANTAnet_tcpServer_c* server;
	QUANTAnet_extendedTcpClient_c* client;
};

const int QUANTAnet_ServerClient64_c::OK = 1;
const int QUANTAnet_ServerClient64_c::FAILED = 0;
const int QUANTAnet_ServerClient64_c::CANNOT_CONNECT = 2;
const int QUANTAnet_ServerClient64_c::CANNOT_READ_LOCAL_FILE = 3;
const int QUANTAnet_ServerClient64_c::CANNOT_WRITE_LOCAL_FILE = 3;
const int QUANTAnet_ServerClient64_c::INVALID_COMMAND = 4;
const int QUANTAnet_ServerClient64_c::MEM_ALLOC_ERR = 5;
const int QUANTAnet_ServerClient64_c::FILE_ERROR = 6;
const int QUANTAnet_ServerClient64_c::DOWNLOAD = 7;

QUANTAnet_ServerClient64_c::QUANTAnet_ServerClient64_c(QUANTAnet_tcpServer_c* Server,
													   int ServerIP, int ServerPort)
{
	client = NULL;
	server = Server;
	serverIPNumber = ServerIP;
	serverPortNumber = ServerPort;
	neighborPortNumber = QUANTAnet_remoteFileIOServer64_c::DEFAULT_PORT;
}

QUANTAnet_ServerClient64_c::~QUANTAnet_ServerClient64_c()
{
	if (client) {
		client->close();
		delete client;
	}
}

int QUANTAnet_ServerClient64_c::getRemoteDateTime(char *ipAddress, 
						char* remoteFileName, int *remDay, int *remMon, 
						int *remYear, int* remHour, int* remMin, int* remSec)
{
	QUANTAnet_extendedTcpClient_c *client = new QUANTAnet_extendedTcpClient_c;

	if (client == NULL || server == NULL) {
		*remDay = -1; *remMon = -1; *remYear = -1;
		*remHour = -1; *remMin = -1; *remSec = -1;
		return QUANTAnet_ServerClient64_c::MEM_ALLOC_ERR;
	}

	int status = client->connectToServer(ipAddress, neighborPortNumber);
	if (status <0) {
		*remDay = -1; *remMon = -1; *remYear = -1;
		*remHour = -1; *remMin = -1; *remSec = -1;
		return QUANTAnet_ServerClient64_c::CANNOT_CONNECT;
	}

	status = client->writeChar('t', QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) 
		return QUANTAnet_ServerClient64_c::FAILED;

	// Send filename length
	int filenamelen = strlen(remoteFileName) + 1;
	status = client->writeInt(filenamelen, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) 
		return QUANTAnet_ServerClient64_c::FAILED;

	// Send filename
	int nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) 
		return QUANTAnet_ServerClient64_c::FAILED;

	status = client->writeInt((serverIPNumber) ? serverIPNumber : QUANTAnet_socketbase_c::getSelfIP(), QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) 
		return QUANTAnet_ServerClient64_c::FAILED;

	// now sends out server's port number
	status = client->writeInt((serverPortNumber) ? serverPortNumber : server->getSelfPort(), QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) 
		return QUANTAnet_ServerClient64_c::FAILED; 

	// Now wait for response.
	char transferMode = ' ';
	status = client->readChar(transferMode, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) 
		return QUANTAnet_ServerClient64_c::FAILED;

	switch (transferMode) {
		case 'i': // invalid command
		case 'n': // File not exist
			*remDay = -1; *remMon = -1; *remYear = -1;
			*remHour = -1; *remMin = -1; *remSec = -1;
			return QUANTAnet_ServerClient64_c::FILE_ERROR;

		case 't': // retrieving date/time
		{
			int remdate[6];
			QUANTAnet_datapack_c packer;

			nbytes = packer.sizeof_int() * 6;
			char* read_buffer = new char[nbytes];
			status = client->read((char*) read_buffer, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
	        packer.initUnpack(read_buffer,nbytes);  
			for (int i = 0; i < 6; i++) 
				packer.unpackInt(&(remdate[i]));
	        delete read_buffer;                     

			if (status != QUANTAnet_tcpClient_c::OK) 
				return QUANTAnet_ServerClient64_c::FILE_ERROR;

			*remDay = remdate[0]; *remMon = remdate[1]; *remYear = remdate[2];
			*remHour = remdate[3]; *remMin = remdate[4]; *remSec = remdate[5];
			if ((remdate[0] == -1) && (remdate[1] == -1) && (remdate[2] == -1) &&
				(remdate[3] == -1) && (remdate[4] == -1) && (remdate[5] == -1)) {
				return QUANTAnet_ServerClient64_c::FILE_ERROR;
			}
			return QUANTAnet_ServerClient64_c::OK;
		}

		default:
			break;
	}
	return QUANTAnet_ServerClient64_c::FAILED;
}

int QUANTAnet_ServerClient64_c::initDownload(char *ipAddress, 
											char* localFileName, char* remoteFileName)
{
	QUANTAnet_datapack_c packer;
	
	client = new QUANTAnet_extendedTcpClient_c;
	if (client == NULL || server == NULL) return QUANTAnet_ServerClient64_c::MEM_ALLOC_ERR;

	int status = client->connectToServer(ipAddress, neighborPortNumber);
	if (status <0) 
		return QUANTAnet_ServerClient64_c::CANNOT_CONNECT;

	char transferMode = 'o';
	status = client->writeChar(transferMode, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) 
		return QUANTAnet_ServerClient64_c::FAILED;

	// Send filename length
	int filenamelen = strlen(remoteFileName) + 1;

	status = client->writeInt(filenamelen, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) 
		return QUANTAnet_ServerClient64_c::FAILED;

	// Send filename
	int nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) 
		return QUANTAnet_ServerClient64_c::FAILED;

	// now sends out server's IP number
	status = client->writeInt((serverIPNumber) ? serverIPNumber : QUANTAnet_socketbase_c::getSelfIP(), QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) 
		return QUANTAnet_ServerClient64_c::FAILED;

	// now sends out server's port number
	status = client->writeInt((serverPortNumber) ? serverPortNumber : server->getSelfPort(), QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) 
		return QUANTAnet_ServerClient64_c::FAILED;

	// Now wait for response.
	transferMode =' ';
	status = client->readChar(transferMode, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) 
		return QUANTAnet_ServerClient64_c::FAILED;

	switch (transferMode) {
		case 'i': // invalid command
		case 'n': // File not exist
			return QUANTAnet_ServerClient64_c::FAILED;

		case 'o': // Download 
			return QUANTAnet_ServerClient64_c::DOWNLOAD;

		default:
			return QUANTAnet_ServerClient64_c::INVALID_COMMAND;
	}
}

int QUANTAnet_ServerClient64_c::readBlock(char* dataBlock, int nBytes)
{
	if (client) {
		// Read from file and write to socket in chunks
		int status = client->read(dataBlock, &nBytes, QUANTAnet_tcpClient_c::BLOCKING);
		if (status != QUANTAnet_tcpClient_c::OK) 
			return QUANTAnet_ServerClient64_c::FAILED;
	
		return QUANTAnet_ServerClient64_c::OK;
	}
	return QUANTAnet_ServerClient64_c::FAILED;
}

/****************************************************************/

const int QUANTAnet_remoteFileIOServer64_c::DEFAULT_PORT = 8000;
const int QUANTAnet_remoteFileIOServer64_c::THREADED = 1;
const int QUANTAnet_remoteFileIOServer64_c::NON_THREADED = 0;
const int QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE = 30000;

int QUANTAnet_remoteFileIOServer64_c::init(int inPort, char* neighborServer, 
										   int neighborPort, int storeLocally)
{
	if (neighborServer) strcpy(neighborServerName, neighborServer);
	else neighborServerName[0] = '\0';
	store = storeLocally;
	neighborPortNumber = neighborPort;

	return server.init(inPort);
}

void QUANTAnet_remoteFileIOServer64_c::process(int threaded)
{
	QUANTAnet_extendedTcpClient_c *newClient = server.checkForNewConnections();
	if (newClient) {
		ServerData* data = new ServerData;

		data->client = newClient;
		data->neighborServerName = &neighborServerName[0];
		data->store = store;
		data->port = neighborPortNumber;
		data->server = &server;

		if (threaded== QUANTAnet_remoteFileIOServer64_c::THREADED) {
			QUANTAts_thread_c newThread; // = new QUANTAnet_thread_c();
			newThread.create(threadHandler, (void*) data);
		} 
		else threadHandler((void*) data);
	}
}

#define CLOSE_REMOVE_AND_RETURN 	if (readFile) fclose(readFile); \
									if (writeFile) fclose(writeFile); \
									if (downloadClient) delete downloadClient; \
									delete data; \
									return NULL; 

void* QUANTAnet_remoteFileIOServer64_c::threadHandler(void* inData)
{
	ServerData* data = (ServerData*) inData;
	if (data == NULL) return NULL;

	QUANTAnet_datapack_c packer;
	QUANTAnet_extendedTcpClient_c *newClient = ((ServerData *) data)->client;
	char* neighborServerName = ((ServerData*) data)->neighborServerName;
	int store = ((ServerData*) data)->store;
	int neighborPort = ((ServerData*) data)->port;
	QUANTAnet_tcpServer_c *server = ((ServerData *) data)->server;
	int nbytes;

	// Check if upload or download
	char transferMode;
	int status = newClient->readChar(transferMode, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {
	    delete data; return NULL;
	}

	// Check filename size
	int filenamelen;
	status = newClient->readInt(filenamelen, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) { delete data; return NULL; }

	// Check filename
	char* filename = new char[filenamelen];
	if (filename == NULL) { delete data; return NULL;}
	status = newClient->read(filename, &filenamelen, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) { delete filename; delete data; return NULL; }
	filename[filenamelen - 1]='\0';

	char dataBlock[QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE];

	// If handling check for return date/time stamp for the request from the client
	if (transferMode == 'c') {
		int thedate[6];

		// If file does not exist time stamp will return -1s.
		status = QUANTAnet_remoteFileIOClient64_c::getLocalDateTime(filename,
									  &thedate[0],
									  &thedate[1],
									  &thedate[2],
									  &thedate[3],
									  &thedate[4],
									  &thedate[5]);

		// If file does not exist locally, then check neighbor server if neighbor exists.
		if (status != QUANTAnet_remoteFileIOClient64_c::OK && neighborServerName[0]) {
			if (QUANTAnet_socketbase_c::getSelfIP() == getHostIPNumber(neighborServerName) 
							&& server->getSelfPort() == neighborPort) { // trying to connect to itself
#ifdef DBG_MSG
				printf("trying to connecting to itself. server processing 'c'\n");
#endif
				status = newClient->writeChar('n', QUANTAnet_tcpClient_c::BLOCKING);
				delete filename; delete data; return NULL;
			}
							    
		    QUANTAnet_ServerClient64_c client(server);

			client.setRemotePort(neighborPort);
			status = client.getRemoteDateTime(neighborServerName, filename, 
			                          &thedate[0], &thedate[1],
									  &thedate[2], &thedate[3],
									  &thedate[4], &thedate[5]);
			if (status != QUANTAnet_ServerClient64_c::OK) {
				// Send back command 'n'
				status = newClient->writeChar('n', QUANTAnet_tcpClient_c::BLOCKING);
				delete filename; delete data; return NULL;
			}
		}

		// Send back command 'c'
		status = newClient->writeChar('c', QUANTAnet_tcpClient_c::BLOCKING);
		if (status != QUANTAnet_tcpClient_c::OK) {
			delete filename; delete data; return NULL;
		}

		// Send back date/time stamp of file - also add the local time stamp to the stream -for latency calculations
		nbytes = 6 * packer.sizeof_int()+ packer.sizeof_double();
		char* send_buffer = new char[nbytes];
        double timeStamp = newClient->getTimeInSecs();

        packer.initPack(send_buffer,nbytes);
		for (int i = 0; i < 6; i++)
            packer.packInt(thedate[i]);
        packer.packDouble(timeStamp);
		status = newClient->write((char*) send_buffer, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
		delete send_buffer;

		delete filename;
		delete data;
		return NULL;
	}

	// If handling check for return date/time stamp for the request from a server
	if (transferMode == 't') {
		// Get server IP
		int serverIP;
		status = newClient->readInt(serverIP, QUANTAnet_tcpClient_c::BLOCKING);
		if (status != QUANTAnet_tcpClient_c::OK) { delete filename; delete data; return NULL; }

		// get port number 
		int serverPort;
		newClient->readInt(serverPort, QUANTAnet_tcpClient_c::BLOCKING);
		if (status != QUANTAnet_tcpClient_c::OK) { delete filename; delete data; return NULL; }

/*		// Check whether we are getting back the request this server sent before.
		if (serverIP == QUANTAnet_socketbase_c::getSelfIP() && serverPort == server->getSelfPort()) { 
#ifdef DBG_MSG
			printf("cycle detected. server processing 't'\n");
#endif
			newClient->writeChar('n', QUANTAnet_tcpClient_c::BLOCKING);
			delete filename; delete data; return NULL;
		}*/
		
		int thedate[6];

		// If file does not exist time stamp will return -1s.
		status = QUANTAnet_remoteFileIOClient64_c::getLocalDateTime(filename,
									  &thedate[0],
									  &thedate[1],
									  &thedate[2],
									  &thedate[3],
									  &thedate[4],
									  &thedate[5]);

		// If file does not exist locally, then check neighbor server if neighbor exists.
		if (status != QUANTAnet_remoteFileIOClient64_c::OK && neighborServerName[0]) {
			if (serverIP == getHostIPNumber(neighborServerName) 
							&& serverPort == neighborPort) { // detected cycle
#ifdef DBG_MSG
				printf("Cycle detected. Server processing 't'\n");
#endif
				status = newClient->writeChar('n', QUANTAnet_tcpClient_c::BLOCKING);
				delete filename; delete data; return NULL;
			}

		    QUANTAnet_ServerClient64_c client(server, serverIP, serverPort);

			client.setRemotePort(neighborPort);
			status = client.getRemoteDateTime(neighborServerName, filename, 
			                          &thedate[0], &thedate[1],
									  &thedate[2], &thedate[3],
									  &thedate[4], &thedate[5]);
			if (status != QUANTAnet_ServerClient64_c::OK) {
				newClient->writeChar('n', QUANTAnet_tcpClient_c::BLOCKING);
				delete filename; delete data; return NULL;
			}
		}

		// Send back command 't'
		status = newClient->writeChar('t', QUANTAnet_tcpClient_c::BLOCKING);
		if (status != QUANTAnet_tcpClient_c::OK) {
			delete filename; delete data; return NULL;
		}

		// Send back date/time stamp of file
		nbytes = 6*packer.sizeof_int();
		char* send_buffer = new char[nbytes];
        packer.initPack(send_buffer,nbytes);
		for (int i = 0; i < 6; i++) packer.packInt(thedate[i]);
		status = newClient->write((char*) send_buffer, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
		delete send_buffer;

		delete filename;
		delete data;
		return NULL;
	}

	// If handling upload
	if (transferMode == 'u') {
		// Get datasize
		int64 dataSize;

		status = newClient->readInt64(dataSize, QUANTAnet_tcpClient_c::BLOCKING);
		if (status != QUANTAnet_tcpClient_c::OK) { delete filename; delete data; return NULL; }

		FILE *writeFile = fopen(filename,"wb");
		if (writeFile == NULL) {
			delete filename; delete data; return NULL;
		}

		// Read from file and write to socket in 8K chunks
		int64 numChunks = dataSize / QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE;
		for (int64 i = 0; i < numChunks; i++) {
			nbytes = QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE;
			status = newClient->read(dataBlock, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
			if (status != QUANTAnet_tcpClient_c::OK) {
				fclose(writeFile); delete filename; delete data; return NULL;
			}

			fwrite(dataBlock, nbytes, 1, writeFile);
		}
		
		int remain = (int) (dataSize - (int64) ((int64) numChunks * (int64) QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE));
		if (remain) {
			nbytes = remain;
			status = newClient->read(dataBlock, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
			if (status != QUANTAnet_tcpClient_c::OK) {
				fclose(writeFile); delete filename; delete data; return NULL;
			}

			nbytes = remain;
			fwrite(dataBlock, nbytes, 1, writeFile);
		}
		fclose(writeFile);
		delete filename;
		delete data;
		return NULL;
	}

	// If handling download... 
	if (transferMode == 'd') {
		int64 fileSize;

		FILE *readFile = NULL, *writeFile = NULL;
		QUANTAnet_ServerClient64_c* downloadClient = NULL;
	
		readFile = fopen(filename, "rb");
		if (readFile == NULL) {
			if (neighborServerName[0]) {
				if (QUANTAnet_socketbase_c::getSelfIP() == getHostIPNumber(neighborServerName) 
								&& server->getSelfPort() == neighborPort) { // trying to connect to itself
#ifdef DBG_MSG
					printf("Trying to connecting to itself. server processing 'd'\n");
#endif
					status = newClient->writeChar('n', QUANTAnet_tcpClient_c::BLOCKING);
					delete filename; delete data; return NULL;
				}

				downloadClient = new QUANTAnet_ServerClient64_c(server);
				downloadClient->setRemotePort(neighborPort);

				if (downloadClient == NULL) {
					delete filename; delete data; return NULL;
				}
				if (downloadClient->initDownload(neighborServerName, filename, filename) !=
					                                  QUANTAnet_ServerClient64_c::DOWNLOAD) {
					newClient->writeChar('i', QUANTAnet_tcpClient_c::BLOCKING);
					delete filename; delete downloadClient; 
					delete data; return NULL;
				}

				if (store) writeFile = fopen(filename, "wb");
			}
			else { 
				status = newClient->writeChar('n', QUANTAnet_tcpClient_c::BLOCKING);
				delete filename; delete data; return NULL; 
			}
		}

		delete filename;

		status = newClient->writeChar('d', QUANTAnet_tcpClient_c::BLOCKING);
		if (status != QUANTAnet_tcpClient_c::OK) {
			CLOSE_REMOVE_AND_RETURN
		}

		// Find length of file.
		if (readFile) { // if file exists locally
			fseek64(readFile, 0L, SEEK_END);
			fileSize = ftell64(readFile);
			fseek64(readFile , 0L, SEEK_SET);
		}
		else { // get the length of file from the neighbor server
			status = downloadClient->getFileSize(fileSize);
			if (status != QUANTAnet_ServerClient64_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}

		// Send length of file along with the timestamp, for latency calculations
        nbytes = packer.sizeof_int64() + packer.sizeof_double();
        char* sendBuffer = new char[nbytes];
        packer.initPack(sendBuffer, nbytes);
        double timeStamp = newClient->getTimeInSecs();
        packer.packInt64(fileSize);            
        packer.packDouble(timeStamp);
 
        status = newClient->write((char*) sendBuffer, &nbytes, QUANTAnet_tcpClient_c :: BLOCKING);
		if (status != QUANTAnet_tcpClient_c::OK) {
			CLOSE_REMOVE_AND_RETURN
		}

		// Send file
		int64 numChunks = fileSize / (int64) QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE;
		for (long i = 0; i < numChunks; i++) {
			nbytes = QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE;
			if (readFile) fread(dataBlock, 1, nbytes, readFile);
			else {
				if (downloadClient->readBlock(dataBlock, nbytes) != QUANTAnet_ServerClient64_c::OK) {
					CLOSE_REMOVE_AND_RETURN
				}
				if (writeFile) fwrite(dataBlock, nbytes, 1, writeFile);
			}
			status = newClient->write(dataBlock, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
			if (status != QUANTAnet_tcpClient_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}

		int remain = (int) (fileSize - ((int64) numChunks * (int64) QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE));
		if (remain) {
			nbytes = remain;
			if (readFile) fread(dataBlock, 1, nbytes, readFile);
			else {
				if (downloadClient->readBlock(dataBlock, nbytes) != QUANTAnet_ServerClient64_c::OK) {
					CLOSE_REMOVE_AND_RETURN
				}
				if (writeFile) fwrite(dataBlock, nbytes, 1, writeFile);
			}
			status = newClient->write(dataBlock, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
			if (status != QUANTAnet_tcpClient_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}
		
		CLOSE_REMOVE_AND_RETURN
	}

	// If handling download for the request from a server... 
	if (transferMode == 'o') {
		int64 fileSize;
		FILE *readFile = NULL, *writeFile = NULL;
		QUANTAnet_ServerClient64_c* downloadClient = NULL;

		// Get server name's length
		int serverIP;
		if (newClient->readInt(serverIP, QUANTAnet_tcpClient_c::BLOCKING) != QUANTAnet_tcpClient_c::OK) { 
			delete filename; delete data; return NULL; 
		}

		// get port number 
		int serverPort;
		if (newClient->readInt(serverPort, QUANTAnet_tcpClient_c::BLOCKING) != QUANTAnet_tcpClient_c::OK) { 
			delete filename; delete data; return NULL; 
		}
		
		// Check whether we are getting back the request this server sent before.
/*		if (serverIP == QUANTAnet_socketbase_c::getSelfIP() && serverPort == server->getSelfPort()) { 
			printf("cycle detected\n");
			newClient->writeChar('i', QUANTAnet_tcpClient_c::BLOCKING);
			delete data;
			return NULL;
		}*/

		readFile = fopen(filename, "rb");
		if (readFile == NULL) {
			if (neighborServerName[0]) {
				if (serverIP == getHostIPNumber(neighborServerName) 
								&& serverPort == neighborPort) { // detected cycle
#ifdef DBG_MSG
					printf("Cycle detected in server processing 'o'\n");
#endif
					status = newClient->writeChar('n', QUANTAnet_tcpClient_c::BLOCKING);
					delete filename; delete data; return NULL;
				}

				downloadClient = new QUANTAnet_ServerClient64_c(server, serverIP, serverPort);
				downloadClient->setRemotePort(neighborPort);

				if (downloadClient == NULL) {
					newClient->writeChar('n', QUANTAnet_tcpClient_c::BLOCKING);
					delete filename; delete data; return NULL;
				}
				if (downloadClient->initDownload(neighborServerName, filename, filename) !=
					QUANTAnet_ServerClient64_c::DOWNLOAD) {
					newClient->writeChar('n', QUANTAnet_tcpClient_c::BLOCKING);
					delete filename; delete downloadClient; 
					delete data; return NULL;
				}

				if (store) writeFile = fopen(filename, "wb");
			}
			else { 
				newClient->writeChar('n', QUANTAnet_tcpClient_c::BLOCKING);
				delete filename; delete data; return NULL; 
			}
		}

		delete filename;

		if (newClient->writeChar('o', QUANTAnet_tcpClient_c::BLOCKING) != QUANTAnet_tcpClient_c::OK) {
			CLOSE_REMOVE_AND_RETURN
		}

		// Find length of file.
		if (readFile) { // if file exists locally
			fseek64(readFile, 0L, SEEK_END);
			fileSize = ftell64(readFile);
			fseek64(readFile , 0L, SEEK_SET);
		}
		else { // get the length of file from the neighbor server
			if (downloadClient->getFileSize(fileSize) != QUANTAnet_ServerClient64_c::OK) {
				newClient->writeChar('n', QUANTAnet_tcpClient_c::BLOCKING);
				CLOSE_REMOVE_AND_RETURN
			}
		}

		// Send length of file
		if (newClient->writeInt64(fileSize, QUANTAnet_tcpClient_c::BLOCKING) != QUANTAnet_tcpClient_c::OK) {
			CLOSE_REMOVE_AND_RETURN
		}

		// Send file
		int64 numChunks = fileSize / (int64) QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE;
		for (int64 i = 0; i < numChunks; i++) {
			nbytes = QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE;
			if (readFile) fread(dataBlock, 1, nbytes, readFile);
			else {
				if (downloadClient->readBlock(dataBlock, nbytes) != QUANTAnet_ServerClient64_c::OK) {
					CLOSE_REMOVE_AND_RETURN
                                }
				if (writeFile) fwrite(dataBlock, nbytes, 1, writeFile);
			}
			status = newClient->write(dataBlock, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
			if (status != QUANTAnet_tcpClient_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}

		int remain = (int) (fileSize - ((int64) numChunks * (int64) QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE));
		if (remain) {
			nbytes = remain;
			if (readFile) fread(dataBlock, 1, nbytes, readFile);
			else {
				if (downloadClient->readBlock(dataBlock, nbytes) != QUANTAnet_ServerClient64_c::OK) {
					CLOSE_REMOVE_AND_RETURN
				}
				if (writeFile) fwrite(dataBlock, nbytes, 1, writeFile);
			}
			status = newClient->write(dataBlock, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
			if (status != QUANTAnet_tcpClient_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}
		
		CLOSE_REMOVE_AND_RETURN
	}

	return NULL;
}
#undef CLOSE_REMOVE_AND_RETURN 	

int QUANTAnet_remoteFileIOServer64_c::getHostIPNumber(char* hostNameOrIP)
{
	struct hostent* hostEnt;

	unsigned ip = inet_addr(hostNameOrIP);
	if ((ip != INADDR_NONE) || (strcmp(hostNameOrIP, "255.255.255.255") == 0))
		return (int) ip;
	else {
		hostEnt = gethostbyname(hostNameOrIP);
		if (hostEnt) 
			return *((int*) hostEnt->h_addr);
		else {
		    printf("Invalid address is given: %s\n", hostNameOrIP);
			return 0;
		}
	}
}

/******************************************************/

const int QUANTAnet_remoteFileIOClient64_c::OK = 1;
const int QUANTAnet_remoteFileIOClient64_c::FAILED = 0;
const int QUANTAnet_remoteFileIOClient64_c::CANNOT_CONNECT = 2;
const int QUANTAnet_remoteFileIOClient64_c::CANNOT_READ_LOCAL_FILE = 3;
const int QUANTAnet_remoteFileIOClient64_c::CANNOT_WRITE_LOCAL_FILE = 3;
const int QUANTAnet_remoteFileIOClient64_c::INVALID_COMMAND = 4;
const int QUANTAnet_remoteFileIOClient64_c::MEM_ALLOC_ERR = 5;
const int QUANTAnet_remoteFileIOClient64_c::FILE_ERROR = 6;
const int QUANTAnet_remoteFileIOClient64_c::EARLIER = 7;
const int QUANTAnet_remoteFileIOClient64_c::LATER = 8;
const int QUANTAnet_remoteFileIOClient64_c::SAME = 9;

QUANTAnet_remoteFileIOClient64_c::QUANTAnet_remoteFileIOClient64_c()
{
	remotePort = QUANTAnet_remoteFileIOServer64_c::DEFAULT_PORT;
}

void QUANTAnet_remoteFileIOClient64_c::setRemotePort(int port) {remotePort = port;}

int QUANTAnet_remoteFileIOClient64_c::getRemoteDateTime(char *ipAddress, char* remoteFileName, int *remDay, int *remMon, int *remYear, int* remHour, int* remMin, int* remSec)
{
	QUANTAnet_extendedTcpClient_c *client = new QUANTAnet_extendedTcpClient_c;
	QUANTAnet_datapack_c packer;
	char *read_buffer;
	int nbytes;

	if (client == NULL) return QUANTAnet_remoteFileIOClient64_c::MEM_ALLOC_ERR;

	int status = client->connectToServer(ipAddress, remotePort);
	if (status <0) {
		return QUANTAnet_remoteFileIOClient64_c::CANNOT_CONNECT;
	}

	status = client->writeChar('c', QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteFileIOClient64_c::FAILED;}
    
    //Increment the data sent
    pmonitor.incrementDataSent(sizeof(char));

	// Send filename length
	int filenamelen = strlen(remoteFileName) + 1;
	nbytes = packer.sizeof_int();

	status = client->writeInt(filenamelen, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteFileIOClient64_c::FAILED;}

    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);

	// Send filename
	nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteFileIOClient64_c::FAILED;}

    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);

	// Now wait for response.
	char transferMode =' ';
	status = client->readChar(transferMode, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteFileIOClient64_c::FAILED;}

    //Increment the data read
    nbytes = sizeof(char);
    pmonitor.incrementDataRead(nbytes);

	if (transferMode == 'c') {
		int remdate[6];

		nbytes = packer.sizeof_int() * 6 + packer.sizeof_double();
		read_buffer = new char[nbytes];
        double remoteTimeStamp;
		status = client->read((char*) read_buffer, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
        packer.initUnpack(read_buffer,nbytes);  
		for (int i = 0; i < 6; i++) 
			packer.unpackInt(&(remdate[i]));
        packer.unpackDouble(&remoteTimeStamp);
        delete read_buffer;                     

		if (status != QUANTAnet_tcpClient_c::OK) {
			client->close(); delete client;	return QUANTAnet_remoteFileIOClient64_c::FILE_ERROR;
		}

       //Increment the data read
        pmonitor.incrementDataRead(nbytes);

        //Calculations for performance monitoring
        double localTime = client->getTimeInSecs();
        double latency = fabs(localTime - remoteTimeStamp);
        pmonitor.setInstantLatency(latency);

       //set IP and port info for the performance monitoring statistics        
        char selfIp[256], remoteIp[256];
        
        QUANTAnet_socketbase_c::getSelfIP(selfIp);
        client->getRemoteIP(remoteIp);
        pmonitor.setIPs(selfIp, remoteIp);
        pmonitor.setPorts(client->getSelfPort(), client->getRemotePort());

		*remDay = remdate[0]; *remMon = remdate[1]; *remYear = remdate[2];
		*remHour = remdate[3]; *remMin = remdate[4]; *remSec = remdate[5];
		if ((remdate[0] == -1) && (remdate[1] == -1) && (remdate[2] == -1) &&
		    (remdate[3] == -1) && (remdate[4] == -1) && (remdate[5] == -1)) {
			client->close(); delete client;	return QUANTAnet_remoteFileIOClient64_c::FILE_ERROR;
		}
		return QUANTAnet_remoteFileIOClient64_c::OK;
	}
	else {
		int result = QUANTAnet_remoteFileIOClient64_c::FAILED;

		if (transferMode == 'i')
			result = QUANTAnet_remoteFileIOClient64_c::INVALID_COMMAND;
		else if (transferMode == 'n')
			result = QUANTAnet_remoteFileIOClient64_c::FILE_ERROR;

		*remDay = -1; *remMon = -1; *remYear = -1;
		*remHour = -1; *remMin = -1; *remSec = -1;
		return result;
	}
	return QUANTAnet_remoteFileIOClient64_c::FAILED;
}

int QUANTAnet_remoteFileIOClient64_c::upload(char *ipAddress, char* localFileName, char* remoteFileName)
{
	QUANTAnet_extendedTcpClient_c *client = new QUANTAnet_extendedTcpClient_c;
	QUANTAnet_datapack_c packer;

	if (client == NULL) return QUANTAnet_remoteFileIOClient64_c::MEM_ALLOC_ERR;

	int status = client->connectToServer(ipAddress, remotePort);
	if (status <0) {
		return QUANTAnet_remoteFileIOClient64_c::CANNOT_CONNECT;
	}

	char transferMode = 'u';

	status = client->writeChar(transferMode, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteFileIOClient64_c::FAILED;}

    //Increment the data sent
    int nbytes = sizeof(char);
    pmonitor.incrementDataSent(nbytes);

	// Send filename length
	int filenamelen = strlen(remoteFileName) + 1;
	status = client->writeInt(filenamelen, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteFileIOClient64_c::FAILED;}

    //increment data sent
    nbytes = packer.sizeof_int();
    pmonitor.incrementDataSent(nbytes);

	// Send filename
	nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteFileIOClient64_c::FAILED;}

    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);

	// Send data size
	// Find length of file.
	FILE *readFile = fopen(localFileName, "rb");

	if (readFile==NULL) {
		client->close(); delete client;
		return QUANTAnet_remoteFileIOClient64_c::CANNOT_READ_LOCAL_FILE;
	}

	long fileSize;

	// get file size
	fseek64(readFile, 0, SEEK_END);
	fileSize = ftell64(readFile);
	fseek64(readFile, 0L, SEEK_SET);

	status = client->writeInt64(fileSize, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; fclose(readFile); return QUANTAnet_remoteFileIOClient64_c::FAILED;}

    //Increment the data sent
    nbytes = packer.sizeof_int64();
    pmonitor.incrementDataSent(nbytes);

	// Send data
	char dataBlock[QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE];

	int64 numChunks = fileSize / QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE;

	for (int64 i = 0; i < numChunks; i++) {
		fread(dataBlock, 1, QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE, readFile);
		nbytes = QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE;
		status = client->write(dataBlock, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
		if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; fclose(readFile); return QUANTAnet_remoteFileIOClient64_c::FAILED;}

      //Increment the data sent
        pmonitor.incrementDataSent(nbytes);
	}
	int remain = (int) (fileSize - (int64) ((int64) numChunks * (int64) QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE));
	if (remain) {
		fread(dataBlock, 1, remain, readFile);
		status = client->write(dataBlock, &remain, QUANTAnet_tcpClient_c::BLOCKING);
		if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; fclose(readFile); return QUANTAnet_remoteFileIOClient64_c::FAILED;}

        //Increment the data sent
        pmonitor.incrementDataSent(remain);
	}

    //set IP and port info for the performance monitoring statistics        
    char selfIp[256], remoteIp[256];
    QUANTAnet_socketbase_c::getSelfIP(selfIp);
    client->getRemoteIP(remoteIp);
    pmonitor.setIPs(selfIp, remoteIp);
    pmonitor.setPorts(client->getSelfPort(), client->getRemotePort());
   
	fclose(readFile);
	client->close();
	delete client;
	return QUANTAnet_remoteFileIOClient64_c::OK;
}


int QUANTAnet_remoteFileIOClient64_c::download(char *ipAddress, char* localFileName, char* remoteFileName)
{
	QUANTAnet_extendedTcpClient_c *client = new QUANTAnet_extendedTcpClient_c;
	QUANTAnet_datapack_c packer;

	if (client == NULL) return QUANTAnet_remoteFileIOClient64_c::MEM_ALLOC_ERR;

	int status = client->connectToServer(ipAddress, remotePort);
	if (status <0) {
		return QUANTAnet_remoteFileIOClient64_c::CANNOT_CONNECT;
	}

	char transferMode = 'd';
	int nbytes;
	status = client->writeChar(transferMode, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteFileIOClient64_c::FAILED;}

    //Increment the data sent
    nbytes = sizeof(char);
    pmonitor.incrementDataSent(nbytes);

	// Send filename length
	int filenamelen = strlen(remoteFileName) + 1;
	status = client->writeInt(filenamelen, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteFileIOClient64_c::FAILED;}

    //Increment the data sent
	nbytes = packer.sizeof_int();
    pmonitor.incrementDataSent(nbytes);

	// Send filename
	nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteFileIOClient64_c::FAILED;}
	
     //Increment the data sent
    pmonitor.incrementDataSent(nbytes);	

	transferMode =' ';
	status = client->readChar(transferMode, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteFileIOClient64_c::FAILED;}

    //Increment the data read
    nbytes = sizeof(char);
    pmonitor.incrementDataRead(nbytes);

	if (transferMode == 'd') {
		int64 dataSize;
        double remoteTimeStamp;

		nbytes = packer.sizeof_int64() + packer.sizeof_double();
        char* read_buffer = new char[nbytes];
        status = client->read((char*) read_buffer, &nbytes, QUANTAnet_tcpClient_c :: BLOCKING);

		if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteFileIOClient64_c::FAILED;}

        packer.initUnpack(read_buffer,nbytes);
        packer.unpackInt64(&dataSize);   
        packer.unpackDouble(&remoteTimeStamp);
        delete read_buffer; 
        
        //Increment the data read
        pmonitor.incrementDataRead(nbytes);

        //Calculations for performance monitoring
        double localTime = client->getTimeInSecs();
        double latency = fabs(localTime - remoteTimeStamp);
        pmonitor.setInstantLatency(latency);

		FILE *writeFile = fopen(localFileName,"wb");
		if (writeFile == NULL) {
			client->close(); delete client;
			return QUANTAnet_remoteFileIOClient64_c::CANNOT_WRITE_LOCAL_FILE;
		}

		// Read from file and write to socket in chunks
		int64 numChunks = dataSize / (int64) QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE;
		char dataBlock[QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE];

		for (int64 i = 0; i < numChunks; i++) {
			nbytes = QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE;
			status = client->read(dataBlock, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
			if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; fclose(writeFile); return QUANTAnet_remoteFileIOClient64_c::FAILED;}

            //Increment the data read
            pmonitor.incrementDataRead(nbytes);

			fwrite(dataBlock, nbytes, 1, writeFile);
		}

		int remain = (int) (dataSize - ((int64) numChunks * (int64) QUANTAnet_remoteFileIOServer64_c::TRANSFERSIZE));
		if (remain) {
			nbytes = remain;
			status = client->read(dataBlock, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
			if (status != QUANTAnet_tcpClient_c::OK) {client->close(); delete client; fclose(writeFile); return QUANTAnet_remoteFileIOClient64_c::FAILED;}
            //Increment the data read
            pmonitor.incrementDataRead(nbytes);
			nbytes = remain;
			fwrite(dataBlock, nbytes, 1, writeFile);
		}
        //set IP and port info for the performance monitoring statistics
        char selfIp[256], remoteIp[256];        
        QUANTAnet_socketbase_c::getSelfIP(selfIp);
        client->getRemoteIP(remoteIp);
        pmonitor.setIPs(selfIp, remoteIp);
        pmonitor.setPorts(client->getSelfPort(), client->getRemotePort());

		fclose(writeFile);
		client->close();
		delete client;
		return QUANTAnet_remoteFileIOClient64_c::OK;
	} 
	else {
		int result;

		if (transferMode == 'i') 
			result = QUANTAnet_remoteFileIOClient64_c::INVALID_COMMAND;
		else if (transferMode == 'n')
			result = QUANTAnet_remoteFileIOClient64_c::FILE_ERROR;

		client->close();
		delete client;
		return result;
	}
}

// Given a filename return its modification time
int QUANTAnet_remoteFileIOClient64_c::getLocalDateTime(char* filename,
		 int *day,
		 int *month,
		 int *year,
		 int *hour,
		 int *minute,
		 int *second)
{

	struct stat buf;

	*day = *month = *year = *hour = *minute = *second = -1;

	if (stat(filename, &buf) != 0) return QUANTAnet_remoteFileIOClient64_c::FILE_ERROR;

	time_t tim;

#ifdef __sgi                      
    tim = buf.st_mtim.tv_sec; 
#else                      
    tim = buf.st_mtime;       
#endif                            
                                  
	struct tm *gmt;
	gmt = gmtime(&tim);

	*day = gmt->tm_mday;
	*month = gmt->tm_mon + 1;
	*year = gmt->tm_year + 1900;
	*hour = gmt->tm_hour;
	*minute = gmt->tm_min;
	*second = gmt->tm_sec;

	return QUANTAnet_remoteFileIOClient64_c::OK;
}

int QUANTAnet_remoteFileIOClient64_c::compareDateTime(int day1, int month1, int year1,
						      int hour1, int minute1, int second1,
						      int day2, int month2, int year2,
						      int hour2, int minute2, int second2)
{
	if (year1 < year2) return QUANTAnet_remoteFileIOClient64_c::EARLIER;
	if (year1 > year2) return QUANTAnet_remoteFileIOClient64_c::LATER;

	if (month1 < month2) return QUANTAnet_remoteFileIOClient64_c::EARLIER;
	if (month1 > month2) return QUANTAnet_remoteFileIOClient64_c::LATER;

	if (day1 < day2) return QUANTAnet_remoteFileIOClient64_c::EARLIER;
	if (day1 > day2) return QUANTAnet_remoteFileIOClient64_c::LATER;

	if (hour1 < hour2) return QUANTAnet_remoteFileIOClient64_c::EARLIER;
	if (hour1 > hour2) return QUANTAnet_remoteFileIOClient64_c::LATER;

	if (minute1 < minute2) return QUANTAnet_remoteFileIOClient64_c::EARLIER;
	if (minute1 > minute2) return QUANTAnet_remoteFileIOClient64_c::LATER;

	if (second1 < second2) return QUANTAnet_remoteFileIOClient64_c::EARLIER;
	if (second1 > second2) return QUANTAnet_remoteFileIOClient64_c::LATER;

	return QUANTAnet_remoteFileIOClient64_c::SAME;
}

void QUANTAnet_remoteFileIOClient64_c::showStats(char* streamInfo, char* comment)        
{
    pmonitor.showStats(streamInfo, comment);    
}

int QUANTAnet_remoteFileIOClient64_c::logStats(char* streamInfo, char* comment, FILE* filePtr)
{
    return pmonitor.logStats(streamInfo, comment, filePtr);
}


int QUANTAnet_remoteFileIOClient64_c::sendStats(char* streamInfo, char* comment)
{
    return pmonitor.sendStats(streamInfo, comment);    
}

int QUANTAnet_remoteFileIOClient64_c::initSendStats(char* monitorClientIP, int port)
{
    return pmonitor.initSendStats(monitorClientIP, port);
}

void QUANTAnet_remoteFileIOClient64_c::exitSendStats()
{
    pmonitor.exitSendStats();
}


