/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <QUANTA/QUANTAnet_remoteParallelFileIO32_c.hxx>
#include <QUANTA/QUANTAts_thread_c.hxx>
#include <QUANTA/QUANTAnet_datapack_c.hxx>

#ifdef WIN32
#include <time.h>
#include <stdlib.h>

#define MAXPATHLEN	MAX_PATH
#else
#include <sys/param.h>
#include <netdb.h>
#include <netinet/in.h>
#endif

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

#ifdef sun
#define INADDR_NONE 0xffffffff
#endif

/*
Data stream format from client to server:

Upload:
'u' +
filenamelen[int] +
filename +
datasize[int]
data

Download:
'd' +
filenamelen[int] +
filename

From server to client- in response to download message:
'd' +
filesize[int]
data

*/

class ServerData
{
public:
    char* neighborServerName;
	QUANTAnet_extendedParallelTcpClient_c* client;
	QUANTAnet_extendedParallelTcpServer_c* server;
	int store;
	int port;
	int numberOfSockets;
    ServerData() {client = NULL; server = NULL;}
	virtual ~ServerData() { if (client) { delete client; } }
};

/* Minimized client version of remote file IO client that can
   only download.
*/
class QUANTAnet_ParallelServerClient32_c {
	public:

	/// Status ok
	static const int OK/* = 1*/;

	/// Status failed
	static const int FAILED/* = 0*/;

	/// Cannot connect to remote server.
	static const int CANNOT_CONNECT/* = 2*/;

	/// Cannot open local file for reading.
	static const int CANNOT_READ_LOCAL_FILE/* = 3*/;

	/// CAnnot write to local file.
	static const int CANNOT_WRITE_LOCAL_FILE/* = 3*/;

	/// Received invalid command from server.
	static const int INVALID_COMMAND/* = 4*/;

	/// Memory allocation error.
	static const int MEM_ALLOC_ERR/* = 5*/;

	/// Remote file error.
	static const int FILE_ERROR/* = 6*/;

	/// Remote file error.
	static const int DOWNLOAD/* = 7*/;

	QUANTAnet_ParallelServerClient32_c(QUANTAnet_extendedParallelTcpServer_c* Server,
						               int ServerIP = 0, int ServerPort = 0);
	~QUANTAnet_ParallelServerClient32_c();

	/// Set port number to connect to.
	void setRemotePort(int port = QUANTAnet_remoteParallelFileIOServer32_c::DEFAULT_PORT) {
	    neighborPortNumber = port;
	}

	/** Download a file from the remote server. If the remote file is not found or
	    a network error occurs, this function will return with a failed.
	    @param ipAddress is IP address of remote server.
	    @param localFileName is local file write to.
	    @param remoteFileName is remote file to get.
	    @return OK, FAILED, CANNOT_WRITE_LOCAL_FILE, INVALID_COMMAND, CANNOT_CONNECT
	*/
	int initDownload(char *ipAddress, char* localFilename, char* remoteFileName, int numOfSockets);
	int readBlock(char*, int nBytes);
	int getFileSize(int& fileSize) { return client->readInt(fileSize); }
	int getRemoteDateTime(char *ipAddress, char* remoteFileName, int *remDay, 
		        int *remMon, int *remYear, int* remHour, int* remMin, int* remSec);

private:
	int neighborPortNumber;
	int store;
	int serverIPNumber;
	int serverPortNumber;
	char neighborServerName[MAXPATHLEN];
	QUANTAnet_extendedParallelTcpServer_c* server;
	QUANTAnet_extendedParallelTcpClient_c* client;
};

const int QUANTAnet_ParallelServerClient32_c::OK = 1;
const int QUANTAnet_ParallelServerClient32_c::FAILED = 0;
const int QUANTAnet_ParallelServerClient32_c::CANNOT_CONNECT = 2;
const int QUANTAnet_ParallelServerClient32_c::CANNOT_READ_LOCAL_FILE = 3;
const int QUANTAnet_ParallelServerClient32_c::CANNOT_WRITE_LOCAL_FILE = 3;
const int QUANTAnet_ParallelServerClient32_c::INVALID_COMMAND = 4;
const int QUANTAnet_ParallelServerClient32_c::MEM_ALLOC_ERR = 5;
const int QUANTAnet_ParallelServerClient32_c::FILE_ERROR = 6;
const int QUANTAnet_ParallelServerClient32_c::DOWNLOAD = 7;

QUANTAnet_ParallelServerClient32_c::QUANTAnet_ParallelServerClient32_c(QUANTAnet_extendedParallelTcpServer_c* Server,
																	   int ServerIP, int ServerPort)
{
	client = NULL;
	server = Server;
	serverIPNumber = ServerIP;
	serverPortNumber = ServerPort;
	neighborPortNumber = QUANTAnet_remoteParallelFileIOServer32_c::DEFAULT_PORT;
}

QUANTAnet_ParallelServerClient32_c::~QUANTAnet_ParallelServerClient32_c()
{
	if (client) delete client;
}

int QUANTAnet_ParallelServerClient32_c::getRemoteDateTime(char *ipAddress, 
						char* remoteFileName, int *remDay, int *remMon, 
						int *remYear, int* remHour, int* remMin, int* remSec)
{
	QUANTAnet_extendedParallelTcpClient_c *client = new QUANTAnet_extendedParallelTcpClient_c;
	QUANTAnet_datapack_c packer;

	if (client == NULL || server == NULL) {
		*remDay = -1; *remMon = -1; *remYear = -1;
		*remHour = -1; *remMin = -1; *remSec = -1;
		return QUANTAnet_ParallelServerClient32_c::MEM_ALLOC_ERR;
	}

	int status = client->connectToServer(ipAddress, neighborPortNumber, 1);
	if (status == QUANTAnet_extendedParallelTcpClient_c::FAILED) {
		*remDay = -1; *remMon = -1; *remYear = -1;
		*remHour = -1; *remMin = -1; *remSec = -1;
		return QUANTAnet_ParallelServerClient32_c::CANNOT_CONNECT;
	}

	status = client->writeChar('t');
	if (status != QUANTAnet_extendedParallelTcpClient_c::OK) 
		return QUANTAnet_ParallelServerClient32_c::FAILED;

	// Send filename length
	int filenamelen = strlen(remoteFileName) + 1;
	status = client->writeInt(filenamelen);
	if (status != QUANTAnet_extendedParallelTcpClient_c::OK) 
		return QUANTAnet_ParallelServerClient32_c::FAILED;

	// Send filename
	int nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes);
	if (status != QUANTAnet_parallelTcpClient_c::OK) 
		return QUANTAnet_ParallelServerClient32_c::FAILED;

	status = client->writeInt((serverIPNumber) ? serverIPNumber : QUANTAnet_socketbase_c::getSelfIP());
	if (status != QUANTAnet_parallelTcpClient_c::OK) 
		return QUANTAnet_ParallelServerClient32_c::FAILED;

	// now sends out server's port number
	status = client->writeInt((serverPortNumber) ? serverPortNumber : client->getSelfPort());
	if (status != QUANTAnet_parallelTcpClient_c::OK) 
		return QUANTAnet_ParallelServerClient32_c::FAILED; 

	// Now wait for response.
	char transferMode = ' ';
	status = client->readChar(transferMode);
	if (status != QUANTAnet_parallelTcpClient_c::OK) 
		return QUANTAnet_ParallelServerClient32_c::FAILED;

	switch (transferMode) {
		case 'i': // invalid command
		case 'n': // File not exist
			*remDay = -1; *remMon = -1; *remYear = -1;
			*remHour = -1; *remMin = -1; *remSec = -1;
			return QUANTAnet_ParallelServerClient32_c::FILE_ERROR;

		case 't': // retrieving date/time
		{
			int remdate[6];

			nbytes = packer.sizeof_int() * 6;
			char* read_buffer = new char[nbytes];
			status = client->read((char*) read_buffer, &nbytes);
	        packer.initUnpack(read_buffer,nbytes);  
			for (int i = 0; i < 6; i++) 
				packer.unpackInt(&(remdate[i]));
	        delete read_buffer;                     

			if (status != QUANTAnet_parallelTcpClient_c::OK) 
				return QUANTAnet_ParallelServerClient32_c::FILE_ERROR;

			*remDay = remdate[0]; *remMon = remdate[1]; *remYear = remdate[2];
			*remHour = remdate[3]; *remMin = remdate[4]; *remSec = remdate[5];
			if ((remdate[0] == -1) && (remdate[1] == -1) && (remdate[2] == -1) &&
				(remdate[3] == -1) && (remdate[4] == -1) && (remdate[5] == -1)) {
				return QUANTAnet_ParallelServerClient32_c::FILE_ERROR;
			}
			return QUANTAnet_ParallelServerClient32_c::OK;
		}

		default:
			break;
	}
	return QUANTAnet_ParallelServerClient32_c::FAILED;
}

int QUANTAnet_ParallelServerClient32_c::initDownload(char *ipAddress, char* localFileName, 
                                               char* remoteFileName, int numOfSockets)
{
	QUANTAnet_datapack_c packer;
	
	client = new QUANTAnet_extendedParallelTcpClient_c;
	if (client == NULL || server == NULL) return QUANTAnet_ParallelServerClient32_c::MEM_ALLOC_ERR;

	int status = client->connectToServer(ipAddress, neighborPortNumber, numOfSockets);
	if (status == QUANTAnet_parallelTcpClient_c::FAILED) {
		return QUANTAnet_ParallelServerClient32_c::CANNOT_CONNECT;
	}

	char transferMode = 'o';
	status = client->writeChar(transferMode);
	if (status != QUANTAnet_parallelTcpClient_c::OK) 
		return QUANTAnet_ParallelServerClient32_c::FAILED;

	// Send filename length
	int filenamelen = strlen(remoteFileName) + 1;

	status = client->writeInt(filenamelen);
	if (status != QUANTAnet_parallelTcpClient_c::OK) 
		return QUANTAnet_ParallelServerClient32_c::FAILED;

	// Send filename
	int nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {
		delete client; return QUANTAnet_ParallelServerClient32_c::FAILED;
	}

	status = client->writeInt((serverIPNumber) ? serverIPNumber : QUANTAnet_socketbase_c::getSelfIP());
	if (status != QUANTAnet_parallelTcpClient_c::OK) 
		return QUANTAnet_ParallelServerClient32_c::FAILED;

	// now sends out server's port number
	status = client->writeInt((serverPortNumber) ? serverPortNumber : client->getSelfPort());
	if (status != QUANTAnet_parallelTcpClient_c::OK) 
		return QUANTAnet_ParallelServerClient32_c::FAILED;

	// Now wait for response.
	transferMode =' ';
	status = client->readChar(transferMode);
	if (status != QUANTAnet_parallelTcpClient_c::OK) 
		return QUANTAnet_ParallelServerClient32_c::FAILED;
	
	switch (transferMode) {
		case 'i': // invalid command
		case 'n': // File not exist
			return QUANTAnet_ParallelServerClient32_c::FAILED;

		case 'o': // Download 
			return QUANTAnet_ParallelServerClient32_c::DOWNLOAD;

		default:
			return QUANTAnet_ParallelServerClient32_c::INVALID_COMMAND;
	}
}

int QUANTAnet_ParallelServerClient32_c::readBlock(char* dataBlock, int nBytes)
{
	if (client) {
		// Read from file and write to socket in chunks
		int status = client->read(dataBlock, &nBytes);
		if (status != QUANTAnet_parallelTcpClient_c::OK)
			return QUANTAnet_ParallelServerClient32_c::FAILED;
	
		return QUANTAnet_ParallelServerClient32_c::OK;
	}
	return QUANTAnet_ParallelServerClient32_c::FAILED;
}

const int QUANTAnet_remoteParallelFileIOServer32_c::DEFAULT_PORT = 7673;
const int QUANTAnet_remoteParallelFileIOServer32_c::THREADED = 1;
const int QUANTAnet_remoteParallelFileIOServer32_c::NON_THREADED = 0;
const int QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE = 30000;

int QUANTAnet_remoteParallelFileIOServer32_c::init(int inPort, char* neighborServer, 
                               int neighborPort, int storeLocally, int numOfSockets)
{
	if (neighborServer) strcpy(neighborServerName, neighborServer);
	else neighborServerName[0] = '\0';
	store = storeLocally;
	numberOfSockets = numOfSockets;
	neighborPortNumber = neighborPort;

	return server.init(inPort);
}

void QUANTAnet_remoteParallelFileIOServer32_c :: process(int threaded)
{
	QUANTAnet_extendedParallelTcpClient_c *newClient = server.checkForNewConnections();
	if (newClient) {
		ServerData* data = new ServerData;

		data->client = newClient;
		data->neighborServerName = &neighborServerName[0];
		data->store = store;
		data->port = neighborPortNumber;
		data->numberOfSockets = numberOfSockets;
		data->server = &server;

		if (threaded== QUANTAnet_remoteParallelFileIOServer32_c::THREADED) {
			QUANTAts_thread_c newThread; 
			newThread.create(threadHandler, (void*) data);
		} 
		else threadHandler((void*) data);
	}
}

void QUANTAnet_remoteParallelFileIOServer32_c::setNeighborServer(char* ServerName)
{
  if (ServerName)
    strcpy(neighborServerName, ServerName);
  else
    neighborServerName[0] = '\0';
}

#define CLOSE_REMOVE_AND_RETURN 	if (readFile) fclose(readFile); \
									if (writeFile) fclose(writeFile); \
									if (downloadClient) delete downloadClient; \
									delete data; \
									return NULL; 

void* QUANTAnet_remoteParallelFileIOServer32_c :: threadHandler(void* inData)
{
	ServerData* data = (ServerData*) inData;
	if (data == NULL) return NULL;

	QUANTAnet_datapack_c packer;
	QUANTAnet_extendedParallelTcpClient_c *newClient = ((ServerData *) data)->client;
	char* neighborServerName = ((ServerData*) data)->neighborServerName;
	int store = ((ServerData*) data)->store;
	int neighborPort = ((ServerData*) data)->port;
	int numberOfSockets = ((ServerData*) data)->numberOfSockets;
	QUANTAnet_extendedParallelTcpServer_c* server = ((ServerData*) data)->server;
	int nbytes;

	// Check the command
	char transferMode;
	int status = newClient->readChar(transferMode);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {
	    delete data; return NULL;
	}
	int filenamelen;
    char* filename;
   
    if (transferMode != 'f')    //reading pathname len and pathname is taken care of at a later stage for 'f'
    {
        // Check filename size
	    status = newClient->readInt(filenamelen);
	    if (status != QUANTAnet_parallelTcpClient_c::OK) {
            delete data; return NULL;
	    }
	    // Check filename
	    filename = new char[filenamelen];
	    if (filename == NULL)
        { delete data; return NULL;}
	    status = newClient->read(filename, &filenamelen);
	    if (status != QUANTAnet_parallelTcpClient_c::OK) { delete filename; delete data; return NULL; }
	    filename[filenamelen - 1]= '\0';
	}    
    char dataBlock[QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE];
	// If handling check for return date/time stamp
	if (transferMode == 'c') {
	    int thedate[6];

	    // If file does not exist time stamp will return -1s.
	    status = QUANTAnet_remoteParallelFileIOClient32_c::getLocalDateTime(filename,
									  &thedate[0],
									  &thedate[1],
									  &thedate[2],
									  &thedate[3],
									  &thedate[4],
									  &thedate[5]);
		// If file does not exist locally, then check neighbor server if neighbor exists.
		if (status != QUANTAnet_remoteParallelFileIOClient32_c::OK && neighborServerName[0]) {
			if (QUANTAnet_socketbase_c::getSelfIP() == getHostIPNumber(neighborServerName) 
							&& server->getSelfPort() == neighborPort) { // trying to connect to itself
#ifdef DBG_MSG
				printf("trying to connecting to itself. server processing 'c'\n");
#endif
				status = newClient->writeChar('n');
				delete filename; delete data; return NULL;
			}

		    QUANTAnet_ParallelServerClient32_c client(server);

			client.setRemotePort(neighborPort);
			status = client.getRemoteDateTime(neighborServerName, filename, 
			                          &thedate[0], &thedate[1],
									  &thedate[2], &thedate[3],
									  &thedate[4], &thedate[5]);
			if (status != QUANTAnet_ParallelServerClient32_c::OK) {
				// Send back command 'n'
				status = newClient->writeChar('n');
				delete filename; delete data; return NULL;
			}
		}

		// Send back command 'c'
		status = newClient->writeChar('c');
		if (status != QUANTAnet_parallelTcpClient_c::OK) {
			delete filename; delete data; return NULL;
		}

		// Send back date/time stamp of file
		nbytes = 6 * packer.sizeof_int();
		char* send_buffer = new char[nbytes];
        packer.initPack(send_buffer, nbytes);  
		for (int i = 0; i < 6; i++)	packer.packInt(thedate[i]);
		status = newClient->write((char*) send_buffer, &nbytes);
		delete send_buffer;

		delete filename;
		delete data;
		return NULL;
	}

	// If handling check for return date/time stamp for the request from a server
	if (transferMode == 't') {
		// Get server IP
		int serverIP;
		status = newClient->readInt(serverIP);
		if (status != QUANTAnet_parallelTcpClient_c::OK) { delete filename; delete data; return NULL; }

		// get port number 
		int serverPort;
		newClient->readInt(serverPort);
		if (status != QUANTAnet_parallelTcpClient_c::OK) { delete filename; delete data; return NULL; }

/*		// Check whether we are getting back the request this server sent before.
		if (serverIP == QUANTAnet_socketbase_c::getSelfIP() && serverPort == server->getSelfPort()) { 
#ifdef DBG_MSG
			printf("cycle detected. server processing 't'\n");
#endif
			newClient->writeChar('n');
			delete filename; delete data; return NULL;
		}*/
		
		int thedate[6];

		// If file does not exist time stamp will return -1s.
		status = QUANTAnet_remoteParallelFileIOClient32_c::getLocalDateTime(filename,
									  &thedate[0],
									  &thedate[1],
									  &thedate[2],
									  &thedate[3],
									  &thedate[4],
									  &thedate[5]);

		// If file does not exist locally, then check neighbor server if neighbor exists.
		if (status != QUANTAnet_remoteParallelFileIOClient32_c::OK && neighborServerName[0]) {
			if (serverIP == getHostIPNumber(neighborServerName) 
							&& serverPort == neighborPort) { // detected cycle
#ifdef DBG_MSG
				printf("Cycle detected. Server processing 't'\n");
#endif
				status = newClient->writeChar('n');
				delete filename; delete data; return NULL;
			}

		    QUANTAnet_ParallelServerClient32_c client(server, serverIP, serverPort);

			client.setRemotePort(neighborPort);
			status = client.getRemoteDateTime(neighborServerName, filename, 
			                          &thedate[0], &thedate[1],
									  &thedate[2], &thedate[3],
									  &thedate[4], &thedate[5]);
			if (status != QUANTAnet_ParallelServerClient32_c::OK) {
				newClient->writeChar('n');
				delete filename; delete data; return NULL;
			}
		}

		// Send back command 't'
		status = newClient->writeChar('t');
		if (status != QUANTAnet_parallelTcpClient_c::OK) {
			delete filename; delete data; return NULL;
		}

		// Send back date/time stamp of file
		nbytes = 6 * packer.sizeof_int();
		char* send_buffer = new char[nbytes];
        packer.initPack(send_buffer,nbytes);
		for (int i = 0; i < 6; i++) packer.packInt(thedate[i]);
		status = newClient->write((char*) send_buffer, &nbytes);
		delete send_buffer;

		delete filename;
		delete data;
		return NULL;
	}

	// If handling upload
	if (transferMode == 'u') {
		// Get datasize
		int dataSize;
		
		status = newClient->readInt(dataSize);
		if (status != QUANTAnet_parallelTcpClient_c::OK) {delete filename; delete data; return NULL;}

		FILE *writeFile = fopen(filename,"wb");
		if (writeFile == NULL) {
			delete filename;
			delete data;
			return NULL;
		}

		// Read from file and write to socket in 8K chunks
		int numChunks = dataSize / QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE;
		for (int i = 0; i < numChunks; i++) {
			nbytes = QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE;
			status = newClient->read(dataBlock, &nbytes);

			if (status != QUANTAnet_parallelTcpClient_c::OK) {fclose(writeFile); delete filename; delete data; return NULL;}

			long n = fwrite(dataBlock, 1, nbytes, writeFile);
			if (n != nbytes) {
				printf("QUANTAnet_remoteParallelFileIO32::threadHandler: fwrite error.\n");	
				printf("nbytes = %d result = %ld\n", nbytes, n);
				delete filename; delete data; return NULL;	
			}
		}
		int remain = dataSize - (numChunks * QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE);
		if (remain) {
			nbytes = remain;
			status = newClient->read(dataBlock, &nbytes);

			if (status != QUANTAnet_parallelTcpClient_c::OK) {fclose(writeFile); delete filename; delete data; return NULL;}

			nbytes = remain;
			long n = fwrite(dataBlock, 1, nbytes, writeFile);
			if (n != nbytes) {
				printf("QUANTAnet_remoteParallelFileIO32::threadHandler: fwrite error.\n");	
				delete filename; delete data; return NULL;	
			}
		}
		fclose(writeFile);
		delete filename;
		delete data;

		return NULL;
	}

#if (defined(__sgi)|| defined(linux) || defined(sun))
    if (transferMode == 'f')
    {
        //protocol from client for a file list-> 'f' + pathnamelen + pathname + recursion flag
        int pathnamelen;
        status = newClient->readInt(pathnamelen);
        if (status != QUANTAnet_parallelTcpClient_c::OK) {delete data; return NULL; }
        char* pathname = new char[pathnamelen]; 
        
        nbytes = pathnamelen;
        status = newClient->read((char*)pathname, &nbytes); 
        if (status != QUANTAnet_parallelTcpClient_c::OK) { delete pathname; delete data; return NULL;}
        
        int doRecursion;
        status = newClient->readInt(doRecursion);
        if (status != QUANTAnet_parallelTcpClient_c::OK) { delete data; return NULL; }
       
        int linkTraversal;
        status = newClient->readInt(linkTraversal);
        if (status != QUANTAnet_parallelTcpClient_c::OK) { delete data; return NULL; }
        
        //send reply back to the client - 'f' for command identity
        status = newClient->writeChar('f');
        if (status != QUANTAnet_parallelTcpClient_c::OK) {delete data; return NULL;}
        
        //build the file list under the given path
        QUANTAmisc_fileList_c list;
        int result = list.buildFileList(pathname, doRecursion, linkTraversal);
        delete [] pathname; pathname = NULL;
        int noObjects;
        if (result == QUANTAmisc_fileList_c::FAILED)    //A file error occured - write error message to client
        {
            noObjects = -1;   //to indicate failure
            status = newClient->writeInt(noObjects);
            delete data; 
            return NULL;    
        }
        noObjects = list.getNoFiles();
        
        status = newClient->writeInt(noObjects);
        if (status != QUANTAnet_parallelTcpClient_c::OK) {delete data; return NULL;}
        
        //get the datapacking size
        int packingSize = list.getDataPackingSize();
        
        status = newClient->writeInt(packingSize);
        if (status != QUANTAnet_parallelTcpClient_c::OK) {delete data; return NULL;}
        
        //pack the file objects and send them across
        char* listBuffer = new char[packingSize];
        QUANTAnet_datapack_c* listPacker = new QUANTAnet_datapack_c;
        listPacker->initPack(listBuffer, packingSize);
        
        list.packFileList(listPacker);
        nbytes = packingSize;
        status = newClient->write((char*)listBuffer, &nbytes);        
        delete [] listBuffer; listBuffer = NULL;
        delete listPacker;
        delete data;
        return NULL;
    }
#endif
	/* If handling download... */
	if (transferMode == 'd') {
		int fileSize;

		FILE *readFile = NULL, *writeFile = NULL;
		QUANTAnet_ParallelServerClient32_c* downloadClient = NULL;
	
		readFile = fopen(filename, "rb");
		if (readFile == NULL) {
			if (neighborServerName[0]) {
				if (QUANTAnet_socketbase_c::getSelfIP() == getHostIPNumber(neighborServerName) 
								&& server->getSelfPort() == neighborPort) { // trying to connect to itself
#ifdef DBG_MSG
					printf("Trying to connecting to itself. server processing 'd'\n");
#endif
					status = newClient->writeChar('n');
					delete filename; delete data; return NULL;
				}

				downloadClient = new QUANTAnet_ParallelServerClient32_c(server);
				downloadClient->setRemotePort(neighborPort);

				if (downloadClient == NULL) {
					delete filename; delete data; return NULL;
				}
				if (downloadClient->initDownload(neighborServerName, filename, filename, numberOfSockets) !=
					                                  QUANTAnet_ParallelServerClient32_c::DOWNLOAD) {
					newClient->writeChar('i');
					delete filename; delete downloadClient; 
					delete data; return NULL;
				}

				if (store) writeFile = fopen(filename, "wb");
			}
			else { 
				status = newClient->writeChar('n');
				delete filename; delete data; return NULL; 
			}
		}

		delete filename;

		status = newClient->writeChar('d');
		if (status != QUANTAnet_parallelTcpClient_c::OK) {
			CLOSE_REMOVE_AND_RETURN
		}

		// Find length of file.
		if (readFile) { // if file exists locally
			fseek(readFile, 0, SEEK_END);
			fileSize = ftell(readFile);
			fseek(readFile , 0L, SEEK_SET);
		}
		else { // get the length of file from the neighbor server
			status = downloadClient->getFileSize(fileSize);
			if (status != QUANTAnet_ParallelServerClient32_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}

		// Send length of file
		status = newClient->writeInt(fileSize);
		if (status != QUANTAnet_parallelTcpClient_c::OK) {
			CLOSE_REMOVE_AND_RETURN
		}

		// Send file
		int numChunks = fileSize / (int) QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE;
		for(int i = 0; i < numChunks; i++) {
			nbytes = QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE;
			if (readFile) fread(dataBlock, 1, nbytes, readFile);
			else {
				if (downloadClient->readBlock(dataBlock, nbytes) != QUANTAnet_ParallelServerClient32_c::OK) {
					CLOSE_REMOVE_AND_RETURN
				}
				if (writeFile) fwrite(dataBlock, 1, nbytes, writeFile);
			}
			status = newClient->write(dataBlock, &nbytes);
			if (status != QUANTAnet_parallelTcpClient_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}

		int remain = fileSize - (numChunks * (int) QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE);
		if (remain) {
			nbytes = remain;
			if (readFile) fread(dataBlock, 1, nbytes, readFile);
			else {
				if (downloadClient->readBlock(dataBlock, nbytes) != QUANTAnet_ParallelServerClient32_c::OK) {
					CLOSE_REMOVE_AND_RETURN
				}
				if (writeFile) fwrite(dataBlock, 1, nbytes, writeFile);
			}
			status = newClient->write(dataBlock, &nbytes);
			if (status != QUANTAnet_parallelTcpClient_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}
		
		CLOSE_REMOVE_AND_RETURN
	}

	// If handling download for the request from a server... 
	if (transferMode == 'o') {
		int fileSize;
		FILE *readFile = NULL, *writeFile = NULL;
		QUANTAnet_ParallelServerClient32_c* downloadClient = NULL;

		// Get server name's length
		int serverIP;
		if (newClient->readInt(serverIP) != QUANTAnet_parallelTcpClient_c::OK) { 
			delete filename; delete data; return NULL; 
		}

		// get port number 
		int serverPort;
		if (newClient->readInt(serverPort) != QUANTAnet_parallelTcpClient_c::OK) { 
			delete filename; delete data; return NULL; 
		}
		
		// Check whether we are getting back the request this server sent before.
/*		if (serverIP == QUANTAnet_socketbase_c::getSelfIP() && serverPort == server->getSelfPort()) { 
			printf("cycle detected\n");
			newClient->writeChar('i');
			delete data;
			return NULL;
		}*/

		readFile = fopen(filename, "rb");
		if (readFile == NULL) {
			if (neighborServerName[0]) {
				if (serverIP == getHostIPNumber(neighborServerName) 
								&& serverPort == neighborPort) { // detected cycle
#ifdef DBG_MSG
					printf("Cycle detected in server processing 'o'\n");
#endif
					status = newClient->writeChar('n');
					delete filename; delete data; return NULL;
				}

				downloadClient = new QUANTAnet_ParallelServerClient32_c(server, serverIP, serverPort);
				downloadClient->setRemotePort(neighborPort);

				if (downloadClient == NULL) {
					newClient->writeChar('n');
					delete filename; delete data; return NULL;
				}
				if (downloadClient->initDownload(neighborServerName, filename, filename, numberOfSockets) !=
                 					QUANTAnet_ParallelServerClient32_c::DOWNLOAD) {
					newClient->writeChar('n');
					delete filename; delete downloadClient; 
					delete data; return NULL;
				}

				if (store) writeFile = fopen(filename, "wb");
			}
			else { 
				newClient->writeChar('n');
				delete filename; delete data; return NULL; 
			}
		}

		delete filename;

		if (newClient->writeChar('o') != QUANTAnet_parallelTcpClient_c::OK) {
			CLOSE_REMOVE_AND_RETURN
		}

		// Find length of file.
		if (readFile) { // if file exists locally
			fseek(readFile, 0L, SEEK_END);
			fileSize = ftell(readFile);
			fseek(readFile , 0L, SEEK_SET);
		}
		else { // get the length of file from the neighbor server
			if (downloadClient->getFileSize(fileSize) != QUANTAnet_ParallelServerClient32_c::OK) {
				newClient->writeChar('n');
				CLOSE_REMOVE_AND_RETURN
			}
		}

		// Send length of file
		if (newClient->writeInt(fileSize) != QUANTAnet_parallelTcpClient_c::OK) {
			CLOSE_REMOVE_AND_RETURN
		}

		// Send file
		int numChunks = fileSize / (int) QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE;
		for(int i = 0; i < numChunks; i++) {
			nbytes = QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE;
			if (readFile) fread(dataBlock, 1, nbytes, readFile);
			else {
				if (downloadClient->readBlock(dataBlock, nbytes) != QUANTAnet_ParallelServerClient32_c::OK) {
					CLOSE_REMOVE_AND_RETURN
				}
				if (writeFile) fwrite(dataBlock, 1, nbytes, writeFile);
			}
			status = newClient->write(dataBlock, &nbytes);
			if (status != QUANTAnet_parallelTcpClient_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}

		int remain = fileSize - (numChunks * (int) QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE);
		if (remain) {
			nbytes = remain;
			if (readFile) fread(dataBlock, 1, nbytes, readFile);
			else {
				if (downloadClient->readBlock(dataBlock, nbytes) != QUANTAnet_ParallelServerClient32_c::OK) {
					CLOSE_REMOVE_AND_RETURN
				}
				if (writeFile) fwrite(dataBlock, 1, nbytes, writeFile);
			}
			status = newClient->write(dataBlock, &nbytes);
			if (status != QUANTAnet_parallelTcpClient_c::OK) {
				CLOSE_REMOVE_AND_RETURN
			}
		}
		
		CLOSE_REMOVE_AND_RETURN
	}

	return NULL;
}
#undef CLOSE_REMOVE_AND_RETURN 	

int QUANTAnet_remoteParallelFileIOServer32_c::getHostIPNumber(char* hostNameOrIP)
{
	struct hostent* hostEnt;

	unsigned ip = inet_addr(hostNameOrIP);
	if ((ip != INADDR_NONE) || (strcmp(hostNameOrIP, "255.255.255.255") == 0))
		return (int) ip;
	else {
		hostEnt = gethostbyname(hostNameOrIP);
		if (hostEnt) 
			return *((int*) hostEnt->h_addr);
		else {
		    printf("Invalid address is given: %s\n", hostNameOrIP);
			return 0;
		}
	}
}

/*******************************************************************/

const int QUANTAnet_remoteParallelFileIOClient32_c::OK = 1;
const int QUANTAnet_remoteParallelFileIOClient32_c::FAILED = 0;
const int QUANTAnet_remoteParallelFileIOClient32_c::CANNOT_CONNECT = 2;
const int QUANTAnet_remoteParallelFileIOClient32_c::CANNOT_READ_LOCAL_FILE = 3;
const int QUANTAnet_remoteParallelFileIOClient32_c::CANNOT_WRITE_LOCAL_FILE = 3;
const int QUANTAnet_remoteParallelFileIOClient32_c::INVALID_COMMAND = 4;
const int QUANTAnet_remoteParallelFileIOClient32_c::MEM_ALLOC_ERR = 5;
const int QUANTAnet_remoteParallelFileIOClient32_c::FILE_ERROR = 6;
const int QUANTAnet_remoteParallelFileIOClient32_c::EARLIER = 7;
const int QUANTAnet_remoteParallelFileIOClient32_c::LATER = 8;
const int QUANTAnet_remoteParallelFileIOClient32_c::SAME = 9;

QUANTAnet_remoteParallelFileIOClient32_c :: QUANTAnet_remoteParallelFileIOClient32_c()
{
	remotePort = QUANTAnet_remoteParallelFileIOServer32_c::DEFAULT_PORT;
}

void QUANTAnet_remoteParallelFileIOClient32_c :: setRemotePort(int port) {remotePort = port;}

int QUANTAnet_remoteParallelFileIOClient32_c :: getRemoteDateTime(char *ipAddress, char* remoteFileName, int *remDay, int *remMon, int *remYear, int* remHour, int* remMin, int* remSec, int numberOfSockets)
{
	QUANTAnet_extendedParallelTcpClient_c* client = new QUANTAnet_extendedParallelTcpClient_c;
	QUANTAnet_datapack_c packer;
	char *read_buffer;
	int nbytes;
	int status;

	if (client == NULL) return QUANTAnet_remoteParallelFileIOClient32_c :: MEM_ALLOC_ERR;

	status = client->connectToServer(ipAddress, remotePort, numberOfSockets);
	if (status == QUANTAnet_parallelTcpClient_c::FAILED) {
		return QUANTAnet_remoteParallelFileIOClient32_c :: CANNOT_CONNECT;
	}

	status = client->writeChar('c');
	if (status != QUANTAnet_parallelTcpClient_c::OK) {
	    delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;
	}

    //Increment the data sent
    pmonitor.incrementDataSent(packer.sizeof_char());
	
    // Send filename length
	int filenamelen = strlen(remoteFileName) + 1;
	status = client->writeInt(filenamelen);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

    nbytes = packer.sizeof_int();
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);
        
	// Send filename
	nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

    pmonitor.incrementDataSent(nbytes);
	
    // Now wait for response.
	char transferMode = ' ';
	status = client->readChar(transferMode);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

    //Increment the data read
    nbytes = sizeof(char);
    pmonitor.incrementDataRead(nbytes);
            
	if (transferMode == 'c') {
		int remdate[6];

		nbytes = packer.sizeof_int() * 6;
		read_buffer = new char[nbytes];
		status = client->read(read_buffer, &nbytes);
        packer.initUnpack(read_buffer, nbytes);  
		for (int i = 0; i < 6; i++)
			packer.unpackInt(&(remdate[i]));
        delete read_buffer;                     

		if (status != QUANTAnet_parallelTcpClient_c::OK) {
			delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FILE_ERROR;
		}
        //Increment the data read
        pmonitor.incrementDataRead(nbytes);
        
        //set IP and port info for the performance monitoring statistics        
        char selfIp[256], remoteIp[256];
    
        QUANTAnet_socketbase_c::getSelfIP(selfIp);
        client->getRemoteIP(remoteIp);
        pmonitor.setIPs(selfIp, remoteIp);
        pmonitor.setPorts(client->getSelfPort(), client->getRemotePort());
          
		*remDay = remdate[0]; *remMon = remdate[1]; *remYear = remdate[2];
		*remHour = remdate[3]; *remMin = remdate[4]; *remSec = remdate[5];
		if ((remdate[0] == -1) && (remdate[1] == -1) && (remdate[2] == -1) &&
		    (remdate[3] == -1) && (remdate[4] == -1) && (remdate[5] == -1)) {
			delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FILE_ERROR;
		}
		delete client;
		return QUANTAnet_remoteParallelFileIOClient32_c::OK;
	}
	else {
		int result = QUANTAnet_remoteParallelFileIOClient32_c::FAILED;

		if (transferMode == 'i')
			result = QUANTAnet_remoteParallelFileIOClient32_c::INVALID_COMMAND;
		else if (transferMode == 'n')
			result = QUANTAnet_remoteParallelFileIOClient32_c::FILE_ERROR;

		*remDay = -1; *remMon = -1; *remYear = -1;
		*remHour = -1; *remMin = -1; *remSec = -1;
		return result;
	}
}

#if (defined(__sgi)|| defined(linux) || defined(sun))
int QUANTAnet_remoteParallelFileIOClient32_c::getRemoteFileList(char* ipAddress, char* remotePath, QUANTAmisc_fileInfo_c*& fileList, int* noObjects, int doRecursion, int linkTraversal, int numberOfSockets)
{
    QUANTAnet_extendedParallelTcpClient_c *client = new QUANTAnet_extendedParallelTcpClient_c;
    QUANTAnet_datapack_c packer;
    if (client == NULL) return QUANTAnet_remoteParallelFileIOClient32_c::MEM_ALLOC_ERR;

	int status = client->connectToServer(ipAddress, remotePort, numberOfSockets);
    if (status != QUANTAnet_parallelTcpClient_c::OK) {
        client->close(); delete client;
        return QUANTAnet_remoteParallelFileIOClient32_c::CANNOT_CONNECT;
        }
    int nbytes;
    char transferMode = 'f';                
    status = client->writeChar(transferMode);
    if (status != QUANTAnet_parallelTcpClient_c::OK) 
    {
        client->close(); delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;
    }
    //Increment the data sent
    nbytes = packer.sizeof_char();
    pmonitor.incrementDataSent(nbytes);

	// Send remotePathname length
	int pathnamelen = strlen(remotePath) + 1;
	status = client->writeInt(pathnamelen);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}
    
    //Increment the data sent
	nbytes = packer.sizeof_int();
    pmonitor.incrementDataSent(nbytes);

	// Send pathname
	nbytes = pathnamelen;
	status = client->write(remotePath, &nbytes);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

     //Increment the data sent
    pmonitor.incrementDataSent(nbytes);	
	
    // Send the recursion flag
    nbytes = packer.sizeof_int();
    status = client->writeInt(doRecursion);
    if (status != QUANTAnet_parallelTcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}
    
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);
    
    //Send the linkTraversal flag
    nbytes = packer.sizeof_int();
    status = client->writeInt(linkTraversal);
    if (status != QUANTAnet_parallelTcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);
     
    transferMode =' ';
	status = client->readChar(transferMode);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

    //Increment the data read
    nbytes = packer.sizeof_char();
    pmonitor.incrementDataRead(nbytes);

	if (transferMode == 'f') {    
        //unpack no_objects(int), datasize(int) + time stamp(double) and unpack the file list
    int noFiles;
    status = client->readInt(noFiles);
    if (status != QUANTAnet_parallelTcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}
    
    //Increment the data read
    nbytes = packer.sizeof_int();
    pmonitor.incrementDataRead(nbytes);
    if (noFiles == -1)    
        {
            printf("\nFile error on the remote server - please check the path nae");
            return FILE_ERROR;
        }
    fileList = new QUANTAmisc_fileInfo_c[noFiles];
    *noObjects = noFiles;
    int dataSize;
    nbytes = packer.sizeof_int();
   
    status = client->readInt(dataSize);
    if (status != QUANTAnet_parallelTcpClient_c::OK) {client->close(); delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}
    
    //Increment the data read
    pmonitor.incrementDataRead(nbytes);

    nbytes = dataSize;
    char* readBuffer = new char[nbytes];
    if (readBuffer == NULL)
        {
            printf("\nMemory allocation error\n");
            return MEM_ALLOC_ERR;
        }       
    status = client->read((char*) readBuffer, &nbytes);
    if (status != QUANTAnet_parallelTcpClient_c::OK) {client->close(); delete client; delete readBuffer; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}
    
    //Increment the data read
    pmonitor.incrementDataRead(nbytes);
    //set IP and port info for the performance monitoring statistics
    char selfIp[256], remoteIp[256];        
    QUANTAnet_socketbase_c::getSelfIP(selfIp);
    client->getRemoteIP(remoteIp);
    pmonitor.setIPs(selfIp, remoteIp);
    pmonitor.setPorts(client->getSelfPort(), client->getRemotePort()); 
    
    QUANTAmisc_fileList_c list;
    QUANTAnet_datapack_c* listPacker = new  QUANTAnet_datapack_c;
    listPacker->initUnpack(readBuffer, dataSize);
    if (list.unpackfileList(listPacker, remotePath, fileList, noFiles) == QUANTAmisc_fileList_c::FAILED)
        return FAILED;

    //list.displayFileList();
    delete [] readBuffer;   readBuffer =NULL;
    delete listPacker;

    } //endif (transferMode == 'f')
    
    client->close(); delete client;
    return OK;
}

#endif

int QUANTAnet_remoteParallelFileIOClient32_c :: upload(char *ipAddress, char* localFileName, char* remoteFileName, int numberOfSockets)
{
	QUANTAnet_extendedParallelTcpClient_c* client = new QUANTAnet_extendedParallelTcpClient_c;
	QUANTAnet_datapack_c packer;

	if (client == NULL) return QUANTAnet_remoteParallelFileIOClient32_c :: MEM_ALLOC_ERR;

	int status = client->connectToServer(ipAddress, remotePort, numberOfSockets);
	if (status == QUANTAnet_parallelTcpClient_c::FAILED) {
		return QUANTAnet_remoteParallelFileIOClient32_c :: CANNOT_CONNECT;
	}

	char transferMode = 'u';

	status = client->writeChar(transferMode);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

    //Increment the data sent
    int nbytes = sizeof(char);
    pmonitor.incrementDataSent(nbytes);
             
    // Send filename length
	int filenamelen = strlen(remoteFileName) + 1;

	status = client->writeInt(filenamelen);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

    //increment data sent
    nbytes = packer.sizeof_int();
    pmonitor.incrementDataSent(nbytes);
        
	// Send filename
	nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

    pmonitor.incrementDataSent(nbytes);

	// Send data size
	// Find length of file.
	FILE *readFile = fopen(localFileName, "rb");

	if (readFile==NULL) {
		delete client;
		return QUANTAnet_remoteParallelFileIOClient32_c::CANNOT_READ_LOCAL_FILE;
	}

	fseek(readFile, 0, SEEK_END);
	int fileSize = ftell(readFile);
	fseek( readFile , 0L, SEEK_SET);

	status = client->writeInt(fileSize);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; fclose(readFile); return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

    //Increment the data sent
    nbytes = packer.sizeof_int();
    pmonitor.incrementDataSent(nbytes);
        
	// Send data
	char dataBlock[QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE];

	int numChunks = fileSize / QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE;

	for (int i = 0; i < numChunks; i++) {
		fread(dataBlock, 1, QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE, readFile);
		nbytes = QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE;
		status = client->write(dataBlock, &nbytes);
		if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; fclose(readFile); return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}
	    
        //Increment the data sent
        pmonitor.incrementDataSent(nbytes);
        
    }
	int remain = fileSize - (numChunks * QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE);
	if (remain) {
		fread(dataBlock, 1, remain, readFile);
		status = client->write(dataBlock, &remain);
		if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; fclose(readFile); return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

        //Increment the data sent
        pmonitor.incrementDataSent(nbytes);
	}

    //set IP and port info for the performance monitoring statistics        
    char selfIp[256], remoteIp[256];
    QUANTAnet_socketbase_c::getSelfIP(selfIp);
    client->getRemoteIP(remoteIp);
    pmonitor.setIPs(selfIp, remoteIp);
    pmonitor.setPorts(client->getSelfPort(), client->getRemotePort());
 
	fclose(readFile);	
	delete client;

	return QUANTAnet_remoteParallelFileIOClient32_c :: OK;
}


int QUANTAnet_remoteParallelFileIOClient32_c :: download(char *ipAddress, char* localFileName, char* remoteFileName, int numberOfSockets)
{
	QUANTAnet_extendedParallelTcpClient_c* client = new QUANTAnet_extendedParallelTcpClient_c;
	QUANTAnet_datapack_c packer;

	if (client == NULL) return QUANTAnet_remoteParallelFileIOClient32_c :: MEM_ALLOC_ERR;

	int status = client->connectToServer(ipAddress, remotePort, numberOfSockets);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {
        delete client;
        
		return QUANTAnet_remoteParallelFileIOClient32_c :: CANNOT_CONNECT;
	}

	int nbytes;
	char transferMode = 'd';
	status = client->writeChar(transferMode);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

    //Increment the data sent
    nbytes = packer.sizeof_char();
    pmonitor.incrementDataSent(nbytes);
        
	// Send filename length
	int filenamelen = strlen(remoteFileName) + 1;
	status = client->writeInt(filenamelen);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

    //Increment the data sent
    nbytes = packer.sizeof_int();
    pmonitor.incrementDataSent(nbytes);
            
	// Send filename
	nbytes = filenamelen;
	status = client->write(remoteFileName, &nbytes);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);
    
	transferMode =' ';
	status = client->readChar(transferMode);
	if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

    //Increment the data read
    nbytes = sizeof(char);
    pmonitor.incrementDataRead(nbytes);
	
    if (transferMode == 'd') {
		int dataSize;

		status = client->readInt(dataSize);                                                     
		if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

        //Increment the data read
        nbytes = packer.sizeof_int();
        pmonitor.incrementDataRead(nbytes);

		FILE *writeFile = fopen(localFileName,"wb");
		if (writeFile == NULL) {
			delete client;
			return QUANTAnet_remoteParallelFileIOClient32_c :: CANNOT_WRITE_LOCAL_FILE;
		}
		// Read from file and write to socket in chunks
		int numChunks = dataSize / (int) QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE;
		char dataBlock[QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE];

		for (int i = 0; i < numChunks; i++) {
			nbytes = QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE;
			status = client->read(dataBlock, &nbytes);
			if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; fclose(writeFile); return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

            //Increment the data read
            pmonitor.incrementDataRead(nbytes);
            
			fwrite(dataBlock, 1, nbytes, writeFile);
		}

		int remain = dataSize - (numChunks * (int) QUANTAnet_remoteParallelFileIOServer32_c::TRANSFERSIZE);
		if (remain) {
			nbytes = remain;
			status = client->read(dataBlock, &nbytes);
			if (status != QUANTAnet_parallelTcpClient_c::OK) {delete client; fclose(writeFile); return QUANTAnet_remoteParallelFileIOClient32_c::FAILED;}

            //Increment the data read
            pmonitor.incrementDataRead(nbytes);
            
			nbytes = remain;
			fwrite(dataBlock, 1, nbytes, writeFile);
		}

        //set IP and port info for the performance monitoring statistics        
        char selfIp[256], remoteIp[256];
    
        QUANTAnet_socketbase_c::getSelfIP(selfIp);
        client->getRemoteIP(remoteIp);
        pmonitor.setIPs(selfIp, remoteIp);
        pmonitor.setPorts(client->getSelfPort(), client->getRemotePort());
     
		fclose(writeFile);
        delete writeFile;
		delete client;
		return QUANTAnet_remoteParallelFileIOClient32_c :: OK;
	} 
	else {
		int result;

		if (transferMode == 'i') 
			result = QUANTAnet_remoteParallelFileIOClient32_c::INVALID_COMMAND;
		else if (transferMode == 'n')
			result = QUANTAnet_remoteParallelFileIOClient32_c::FILE_ERROR;

		delete client;	
		return result;
	}
}

// Given a filename return its modification time
int QUANTAnet_remoteParallelFileIOClient32_c :: getLocalDateTime(char* filename,
		 int *day,
		 int *month,
		 int *year,
		 int *hour,
		 int *minute,
		 int *second)
{

	struct stat buf;

	*day = *month = *year = *hour = *minute = *second = -1;

	if (stat(filename, &buf) != 0) return QUANTAnet_remoteParallelFileIOClient32_c::FILE_ERROR;

	time_t tim;

#ifdef __sgi
    tim = buf.st_mtim.tv_sec; 
#else
    tim = buf.st_mtime;       
#endif                            
                                  
	struct tm *gmt;
	gmt = gmtime(&tim);

	*day = gmt->tm_mday;
	*month = gmt->tm_mon + 1;
	*year = gmt->tm_year + 1900;
	*hour = gmt->tm_hour;
	*minute = gmt->tm_min;
	*second = gmt->tm_sec;

	return QUANTAnet_remoteParallelFileIOClient32_c::OK;
}

int QUANTAnet_remoteParallelFileIOClient32_c :: compareDateTime(int day1, int month1, int year1,
						      int hour1, int minute1, int second1,
						      int day2, int month2, int year2,
						      int hour2, int minute2, int second2)
{
	if (year1 < year2) return QUANTAnet_remoteParallelFileIOClient32_c::EARLIER;
	if (year1 > year2) return QUANTAnet_remoteParallelFileIOClient32_c::LATER;

	if (month1 < month2) return QUANTAnet_remoteParallelFileIOClient32_c::EARLIER;
	if (month1 > month2) return QUANTAnet_remoteParallelFileIOClient32_c::LATER;

	if (day1 < day2) return QUANTAnet_remoteParallelFileIOClient32_c::EARLIER;
	if (day1 > day2) return QUANTAnet_remoteParallelFileIOClient32_c::LATER;

	if (hour1 < hour2) return QUANTAnet_remoteParallelFileIOClient32_c::EARLIER;
	if (hour1 > hour2) return QUANTAnet_remoteParallelFileIOClient32_c::LATER;

	if (minute1 < minute2) return QUANTAnet_remoteParallelFileIOClient32_c::EARLIER;
	if (minute1 > minute2) return QUANTAnet_remoteParallelFileIOClient32_c::LATER;

	if (second1 < second2) return QUANTAnet_remoteParallelFileIOClient32_c::EARLIER;
	if (second1 > second2) return QUANTAnet_remoteParallelFileIOClient32_c::LATER;

	return QUANTAnet_remoteParallelFileIOClient32_c::SAME;
}

void QUANTAnet_remoteParallelFileIOClient32_c::showStats(char* streamInfo, char* comment)        
{
    pmonitor.showStats(streamInfo, comment);    
}

int QUANTAnet_remoteParallelFileIOClient32_c::logStats(char* streamInfo, char* comment, FILE* filePtr)
{
    return pmonitor.logStats(streamInfo, comment, filePtr);
}


int QUANTAnet_remoteParallelFileIOClient32_c::sendStats(char* streamInfo, char* comment)
{
    return pmonitor.sendStats(streamInfo, comment);    
}

int QUANTAnet_remoteParallelFileIOClient32_c::initSendStats(char* monitorClientIP, int port)
{
    return pmonitor.initSendStats(monitorClientIP, port);
}

void QUANTAnet_remoteParallelFileIOClient32_c::exitSendStats()
{
    pmonitor.exitSendStats();
}

