/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <QUANTA/QUANTAnet_rpc_c.hxx>

#include <math.h>
#include <string.h>

const int QUANTAnet_rpcManager_c::DEFAULT_RPCS = 64;
const int QUANTAnet_rpcManager_c::DEFAULT_PORT = 9000;
const int QUANTAnet_rpcManager_c::OK = 1;
const int QUANTAnet_rpcManager_c::CANNOT_MANAGE_ANY_MORE = 4;
const int QUANTAnet_rpcManager_c::RPC_NOT_FOUND = 3;
const int QUANTAnet_rpcManager_c::MEM_ALLOC_ERR = 2;
const int QUANTAnet_rpcManager_c::FAILED = 0;

const int QUANTAnet_rpcCaller_c::NEVER_TIME_OUT = QUANTAnet_tcpClient_c::NO_TIME_OUT;
const int QUANTAnet_rpcCaller_c::FAILED = 0;
const int QUANTAnet_rpcCaller_c::OK = 1;
const int QUANTAnet_rpcCaller_c::MEM_ALLOC_ERR = 2;

/*
DATA FORMAT for RPC messages.

r = call()
r+(rpcnamesize, timestamp)+rpcname+outdatasize+outdata

'r' = char
rpcnamesize = int
timestamp =double
rpcname = char*
outdatasize = int
outdata =char*

s= call1Way()
s+(rpcnamesize, timestamp)+rpcname+outdatasize+outdata

REPLY:
-----
call1Way() gets no reply

call() receives the following reply:

'a'+(indatasize, timestamp)+indata

*/

QUANTAnet_rpcManager_c :: QUANTAnet_rpcManager_c ()
{
}

int QUANTAnet_rpcManager_c :: init(int max, int inport)
{
	// Set up # of RPCs that can be handled by a server.
	maxNumRPCs = max;
	handleList  = new QUANTAnet_rpcHandle_c[max];
	if (handleList == NULL) return QUANTAnet_rpcManager_c :: MEM_ALLOC_ERR;
	if (server.init(inport) == 0) {
		delete handleList;
		return QUANTAnet_rpcManager_c :: FAILED;
	}
	return QUANTAnet_rpcManager_c :: OK;
}

QUANTAnet_rpcHandle_c :: QUANTAnet_rpcHandle_c()
{
	rpcName = NULL;
	callback=NULL;
}

int QUANTAnet_rpcManager_c :: manage(char *procName, int (*callback) (char *indata, int inlen,
							 char **outdata, int *outlen))
{
	// RPC manager manages a handler. Basically register the procdure name
	// in an array along with its callback.
	for (int i = 0; i < maxNumRPCs; i++) {
		if (handleList[i].rpcName == NULL) {
			char *newname = new char[strlen(procName)+1];
			if (newname == NULL) return QUANTAnet_rpcManager_c :: MEM_ALLOC_ERR;
			strcpy(newname, procName);
			handleList[i].rpcName = newname;
			handleList[i].callback = callback;
			return QUANTAnet_rpcManager_c :: OK;
		}
	}
	return QUANTAnet_rpcManager_c :: CANNOT_MANAGE_ANY_MORE;
}

QUANTAnet_rpcHandle_c *QUANTAnet_rpcManager_c :: findRPC(char *procName)
{
	for (int i = 0; i < maxNumRPCs; i++) {
		if (handleList[i].rpcName) {
			if (strcmp(handleList[i].rpcName, procName) == 0) {
				return &handleList[i];
			}
		}
	}
	return NULL;
}

int QUANTAnet_rpcManager_c :: unmanage(char *procName)
{
	for (int i = 0; i < maxNumRPCs; i++) {
		if (handleList[i].rpcName) {
			if (strcmp(handleList[i].rpcName, procName) == 0) {
				delete[] handleList[i].rpcName;
				handleList[i].rpcName = NULL;
				handleList[i].callback = NULL;
				return QUANTAnet_rpcManager_c :: OK;
			}
		}
	}
	return QUANTAnet_rpcManager_c :: RPC_NOT_FOUND;
}

void QUANTAnet_rpcManager_c :: process()
{
	QUANTAnet_tcpClient_c *newClient = server.checkForNewConnections();

	if (newClient) {
		handleRPC(newClient);
		newClient->close();
		delete newClient;
	}	
}
int QUANTAnet_rpcManager_c :: handleRPC(QUANTAnet_tcpClient_c *client)
{
	QUANTAnet_datapack_c packer;
	char *read_buffer;
	
	// Retrieve first byte of RPC message
	int nbytes;
	char transferMode;
	nbytes = sizeof(char);
	int status = client->read((char*) &transferMode, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);

	if (status != QUANTAnet_tcpClient_c :: OK) return QUANTAnet_rpcManager_c::FAILED;
    //Increment the data read
    pmonitor.incrementDataRead(nbytes);

	// Retrieve RPC name size and the time stamp
	int rpcNameSize;
    double remoteTimeStamp;
    double latency;
    
	nbytes = packer.sizeof_int() + packer.sizeof_double();
	read_buffer = new char[nbytes];
	status = client->read((char*) read_buffer, &nbytes, QUANTAnet_tcpClient_c :: BLOCKING);
                                                      
    packer.initUnpack((char *)read_buffer,nbytes);
    packer.unpackInt(&rpcNameSize);
    packer.unpackDouble(&remoteTimeStamp);

    double currentTime = client->getTimeInSecs();
    latency = fabs(currentTime - remoteTimeStamp);
    pmonitor.setInstantLatency(latency);
    
    delete[] read_buffer;  
	if (status != QUANTAnet_tcpClient_c :: OK) return QUANTAnet_rpcManager_c::FAILED;
    //Increment the data read
    pmonitor.incrementDataRead(nbytes);

	// Retrieve RPC name
	char *rpcName = new char[rpcNameSize];
	nbytes = rpcNameSize;
	status = client->read(rpcName, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);

	if (status != QUANTAnet_tcpClient_c :: OK) {
		delete[] rpcName;
		return QUANTAnet_rpcManager_c::FAILED;
	}

    //Increment the data read
    pmonitor.incrementDataRead(nbytes);

	// Retrieve incoming data size.
	int dataSize;
	nbytes = packer.sizeof_int();
	read_buffer = new char[nbytes];
	status = client->read((char*) read_buffer, &nbytes, QUANTAnet_tcpClient_c :: BLOCKING);
                                                      
    packer.initUnpack((char *)read_buffer,nbytes);
    packer.unpackInt(&dataSize);
    delete[] read_buffer;  
	if (status != QUANTAnet_tcpClient_c :: OK) {
		delete[] rpcName;
		return QUANTAnet_rpcManager_c::FAILED;
	}
    //Increment the data read
    pmonitor.incrementDataRead(nbytes);

	// Retrieve incoming data.
	char * buffer = new char[dataSize];
	nbytes = dataSize;
	status = client->read(buffer, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c :: OK) {
		delete[] rpcName;
		delete[] buffer;
		return QUANTAnet_rpcManager_c::FAILED;
	}

    //Increment the data read
    pmonitor.incrementDataRead(nbytes);
    
    char selfIp[256], remoteIp[256];
    //set IP and port info for the performance monitoring statistics
    QUANTAnet_socketbase_c::getSelfIP(selfIp);
    client->getRemoteIP(remoteIp);
    pmonitor.setIPs(selfIp, remoteIp);
    pmonitor.setPorts(client->getSelfPort(), client->getRemotePort());
 
	char *returnValue;
	int returnSize;

	// Find the handler to call give the rpc name.
	QUANTAnet_rpcHandle_c *handler = findRPC(rpcName);
	if (handler == NULL) {
		delete[] rpcName;
		delete[] buffer;
		return QUANTAnet_rpcManager_c::FAILED;
	}

	// call the callback for the handler.
	handler->callback(buffer, dataSize, &returnValue, &returnSize);

	// If the header of this message was a 'r' then need to reply to RPC.
	if (transferMode == 'r') {
		replyRPC(client, returnValue, returnSize);
		if (returnValue)
			delete[] returnValue;
	}

	// Otherwise this was a 1 way call.
	delete[] buffer;
	delete[] rpcName;

	return QUANTAnet_rpcManager_c::OK;
}

int QUANTAnet_rpcManager_c :: replyRPC(QUANTAnet_tcpClient_c *client, char* returnValue, int returnSize)
{
	QUANTAnet_datapack_c packer;
	char *send_buffer=NULL;

	// to reply to an RPC send the 'a' character as header.
	char transferMode = 'a';
	int nbytes = sizeof(char);
	int status = client->write((char*) &transferMode, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);

	if (status != QUANTAnet_tcpClient_c :: OK) {
		return QUANTAnet_rpcManager_c::FAILED;
	}
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);

	// Then send # of bytes to return and the timestamp for latency calculations
	nbytes = packer.sizeof_int() + packer.sizeof_double();
	send_buffer = new char[nbytes];
    double timeStamp = client->getTimeInSecs();
    
    packer.initPack(send_buffer,nbytes);
	packer.packInt(returnSize);
    packer.packDouble(timeStamp);
    
	status = client->write((char*) send_buffer, &nbytes, QUANTAnet_tcpClient_c :: BLOCKING);
	delete[] send_buffer;
	if (status != QUANTAnet_tcpClient_c :: OK) {
		return QUANTAnet_rpcManager_c::FAILED;
	}
    
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);
    
	// then send the data to return
	nbytes = returnSize;
	if (nbytes) {
		status = client->write((char*) returnValue, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
		if (status != QUANTAnet_tcpClient_c :: OK) {
			return QUANTAnet_rpcManager_c::FAILED;
		}
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);        
	}
	return QUANTAnet_rpcManager_c::OK;
}

void QUANTAnet_rpcManager_c::showStats(char* streamInfo, char* comment)
{
    pmonitor.showStats(streamInfo, comment);    
}
 
int QUANTAnet_rpcManager_c::logStats(char* streamInfo, char* comment, FILE* filePtr)
{
    return pmonitor.logStats(streamInfo, comment, filePtr);
}

int QUANTAnet_rpcManager_c::sendStats(char* streamInfo, char* comment)
{
     return pmonitor.sendStats(streamInfo, comment);
}

int QUANTAnet_rpcManager_c::initSendStats(char* monitorClientIP, int port)
{
    return pmonitor.initSendStats(monitorClientIP, port);
}

void QUANTAnet_rpcManager_c::exitSendStats()
{
    pmonitor.exitSendStats();
}


/****************************************************************************/

QUANTAnet_rpcCaller_c :: QUANTAnet_rpcCaller_c()
{
	rpcName = NULL;
	destinationPort = QUANTAnet_rpcManager_c::DEFAULT_PORT;
	itsTimeOut = QUANTAnet_rpcCaller_c::NEVER_TIME_OUT;
}

int QUANTAnet_rpcCaller_c :: init(char *name, char* ip, int destPort)
{
	if (rpcName) delete rpcName;
	rpcName = new  char[strlen(name)+1];
	if (rpcName == NULL) return QUANTAnet_rpcCaller_c :: MEM_ALLOC_ERR;

	strcpy(rpcName, name);
	strcpy(ipAddress, ip);
	destinationPort = destPort;
	return QUANTAnet_rpcCaller_c::OK;
}

int QUANTAnet_rpcCaller_c :: call(char *outdata, int outlen, char** retdata, int *retlen)
{
	QUANTAnet_datapack_c packer;
	char *send_buffer=NULL;
	char *read_buffer=NULL;

	// Create new client
	QUANTAnet_tcpClient_c *client = new QUANTAnet_tcpClient_c;
	if (client == NULL) return QUANTAnet_rpcCaller_c :: MEM_ALLOC_ERR;
	client->setTimeOut(itsTimeOut);

	// Connect to remote server
	int  status = client->connectToServer(ipAddress, destinationPort);
	if (status < 0) return QUANTAnet_rpcCaller_c::FAILED;

	// Set transfer mode to 'r' for call and reply
	char transferMode = 'r';
	int nbytes = sizeof(char);

	status = client->write((char*) &transferMode, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {
		client->close(); delete client;
		return QUANTAnet_rpcCaller_c::FAILED;
	}
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);

	// Send rpc name size and the time stamp - for latency calculations
	int rpcNameSize = strlen(rpcName)+1;
	nbytes = packer.sizeof_int() +packer.sizeof_double();
	send_buffer = new char[nbytes];
    double timeStamp = client->getTimeInSecs();
    
    packer.initPack(send_buffer,nbytes);
	packer.packInt(rpcNameSize);
	packer.packDouble(timeStamp);
    status = client->write((char*) send_buffer, &nbytes, QUANTAnet_tcpClient_c :: BLOCKING);
	delete[] send_buffer;
	if (status != QUANTAnet_tcpClient_c::OK) {
		client->close(); delete client;
		return QUANTAnet_rpcCaller_c::FAILED;
	}

    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);

	// Send RPC name
	nbytes = rpcNameSize;
	status = client->write((char *) rpcName, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {
		client->close(); delete client;
		return QUANTAnet_rpcCaller_c::FAILED;
	}
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);

	// Send outgoing data size
	nbytes = packer.sizeof_int();
	send_buffer = new char[nbytes];
    packer.initPack(send_buffer,nbytes);
	packer.packInt(outlen);
	status = client->write((char*) send_buffer, &nbytes, QUANTAnet_tcpClient_c :: BLOCKING);
	delete[] send_buffer;
	if (status != QUANTAnet_tcpClient_c::OK) {
		client->close(); delete client;
		return QUANTAnet_rpcCaller_c::FAILED;
	}
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);

	// Send outgoing data
	nbytes = outlen;
	status = client->write(outdata, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {
		client->close(); delete client;
		return QUANTAnet_rpcCaller_c::FAILED;
	}
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);

	// Read header of reply
	nbytes = sizeof(char);
	status = client->read((char*) &transferMode, &nbytes,  QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {
		client->close(); delete client;
		return QUANTAnet_rpcCaller_c::FAILED;
	}

    //Increment the data read
    pmonitor.incrementDataRead(nbytes);

	// If reply is 'a' then acknowledged
	if (transferMode == 'a') {
		// Read incoming datasize and the timestamp
		int indatasize;
        double remoteTimeStamp;
        
		nbytes = packer.sizeof_int()+packer.sizeof_double();
		read_buffer = new char[nbytes];
		status = client->read((char*) read_buffer, &nbytes, QUANTAnet_tcpClient_c :: BLOCKING);                                                    
	    packer.initUnpack((char *)read_buffer,nbytes);
		packer.unpackInt(&indatasize);
        packer.unpackDouble(&remoteTimeStamp);
        
	    delete[] read_buffer;  
		if (status != QUANTAnet_tcpClient_c::OK) {
			client->close(); delete client;
			return QUANTAnet_rpcCaller_c::FAILED;
		}
        //Increment the data read
        pmonitor.incrementDataRead(nbytes);

        //Calculations for performance monitoring
        double localTime = client->getTimeInSecs();
        double latency = fabs(localTime - remoteTimeStamp);
        pmonitor.setInstantLatency(latency);
        
		char *indata = new char[indatasize];
		nbytes = indatasize;
		status = client->read(indata, &nbytes,  QUANTAnet_tcpClient_c::BLOCKING);
		if (status != QUANTAnet_tcpClient_c::OK) {
			delete[] indata;
			client->close(); delete client;
			return QUANTAnet_rpcCaller_c::FAILED;
		}
        //Increment the data read
        pmonitor.incrementDataRead(nbytes);

        //set IP and port info for the performance monitoring statistics        
        char selfIp[256], remoteIp[256];
        
        QUANTAnet_socketbase_c::getSelfIP(selfIp);
        client->getRemoteIP(remoteIp);
        pmonitor.setIPs(selfIp, remoteIp);
        pmonitor.setPorts(client->getSelfPort(), client->getRemotePort());
        
        *retdata = indata;
        *retlen = indatasize;
        client->close();
        delete client;
        
        return QUANTAnet_rpcCaller_c::OK;
		
	} 

	client->close();
	delete client;
	return QUANTAnet_rpcCaller_c::FAILED; // failure
}

int QUANTAnet_rpcCaller_c :: call1way(char *outdata, int outlen)
{
	QUANTAnet_datapack_c packer;
	char *send_buffer=NULL;
	
	// Create new client
	QUANTAnet_tcpClient_c *client = new QUANTAnet_tcpClient_c;
	client->setTimeOut(itsTimeOut);

	// Connect to remote server
	int  status = client->connectToServer(ipAddress, destinationPort);
	if (status < 0) {
		delete client;
		return QUANTAnet_rpcCaller_c :: FAILED;
	}

	// Send transfer mode to 's' for call1way
	char transferMode = 's';
	int nbytes = sizeof(char);

	status = client->write((char*) &transferMode, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {
		client->close(); delete client;
		return QUANTAnet_rpcCaller_c :: FAILED;
	}
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);

	// Send rpcname size and the time stamp
	int rpcNameSize = strlen(rpcName)+1;
    double timeStamp = client->getTimeInSecs();
    
	nbytes = packer.sizeof_int() + packer.sizeof_double();
	send_buffer = new char[nbytes];
    packer.initPack(send_buffer,nbytes);
	packer.packInt(rpcNameSize);
    packer.packDouble(timeStamp);
    
	status = client->write((char*) send_buffer, &nbytes, QUANTAnet_tcpClient_c :: BLOCKING);
	delete[] send_buffer;
	if (status != QUANTAnet_tcpClient_c::OK) {
		client->close(); delete client;
		return QUANTAnet_rpcCaller_c :: FAILED;
	}
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);

	// Send rpcname
	nbytes = rpcNameSize;
	status = client->write((char *) rpcName, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {
		client->close(); delete client;
		return QUANTAnet_rpcCaller_c :: FAILED;
	}
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);

	// Send data length
	nbytes = packer.sizeof_int();
	send_buffer = new char[nbytes];
    packer.initPack(send_buffer,nbytes);
	packer.packInt(outlen);
	status = client->write((char*) send_buffer, &nbytes, QUANTAnet_tcpClient_c :: BLOCKING);
	delete[] send_buffer;
	if (status != QUANTAnet_tcpClient_c::OK) {
		client->close(); delete client;
		return QUANTAnet_rpcCaller_c :: FAILED;
	}
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);

	// Send data
	nbytes = outlen;
	status = client->write(outdata, &nbytes, QUANTAnet_tcpClient_c::BLOCKING);
	if (status != QUANTAnet_tcpClient_c::OK) {
		client->close(); delete client;
		return QUANTAnet_rpcCaller_c :: FAILED;
	}
    //Increment the data sent
    pmonitor.incrementDataSent(nbytes);
    
    //set IP and port info for the performance monitoring statistics
    char selfIp[256], remoteIp[256];
    QUANTAnet_socketbase_c::getSelfIP(selfIp);
    client->getRemoteIP(remoteIp);
    pmonitor.setIPs(selfIp, remoteIp);
    pmonitor.setPorts(client->getSelfPort(), client->getRemotePort());

	// That's it. Hope the remote end called it's RPC.
	client->close();
	delete client;
	return QUANTAnet_rpcCaller_c::OK;
}

void  QUANTAnet_rpcCaller_c::showStats(char* streamInfo, char* comment)
{
    pmonitor.showStats(streamInfo, comment);    
}

int  QUANTAnet_rpcCaller_c::logStats(char* streamInfo, char* comment, FILE* filePtr)
{
    return pmonitor.logStats(streamInfo, comment, filePtr);
}

int QUANTAnet_rpcCaller_c::sendStats(char* streamInfo, char* comment)
{
     return pmonitor.sendStats(streamInfo, comment);
}

int QUANTAnet_rpcCaller_c::initSendStats(char* monitorClientIP, int port)
{
    return pmonitor.initSendStats(monitorClientIP, port);
}

void QUANTAnet_rpcCaller_c::exitSendStats()
{
    pmonitor.exitSendStats();
}



