/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <QUANTA/QUANTAnet_udpReflector_c.hxx>

#include <string.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

const int QUANTAnet_udpReflector_c::FAILED = 0;
const int QUANTAnet_udpReflector_c::OK = 1;
const int QUANTAnet_udpReflector_c::SKIP_DISTRIBUTION = 2;

QUANTAnet_udpReflector_c :: QUANTAnet_udpReflector_c()
{
	itsInit = 0;
	itsTimeOutTime = QUANTANET_UDP_REFLECTOR_DEFAULT_TIMEOUT;
	itsNumberOfClients = 0;
	itsMainUDPSocket = NULL;
	doMcast = 0;
	doForcedUDP = 0;
	itsLastTime = 0;
	itsUDPPort = 0;
	itsMcastTTL = QUANTANET_UDP_REFLECTOR_MCAST_TTL;
    	inBufferUDP = new char[QUANTANET_UDP_REFLECTOR_MAX_MSG_SIZE];
    	inBufferMCAST = new char[QUANTANET_UDP_REFLECTOR_MAX_MSG_SIZE];
	interceptCallback = NULL;
	interceptUserData = NULL;
	perfDaemonIP = NULL;
    	perfDaemonPort = PERF_DAEMON_DEFAULT_PORT;
    	perfDaemonStatus = PERFDAEMON_NOT_ALIVE ;
}

// Destructor
QUANTAnet_udpReflector_c :: ~QUANTAnet_udpReflector_c()
{
  // css: The destructor needs to do some non-trivial cleanup. Its
  // processing is handled in separate threads. Those threads may be
  // blocking if data has not been received. Calling the destructor
  // indicates that the connection should be closed. Therefore, first close
  // the socket.
  if (itsMainUDPSocket) {
    itsMainUDPSocket->close();
  }
  // css: Now that the socket is closed, all blocking functions should
  // unblock. Next, stop the thread objects. QUANTAts_thread_c::stop() does
  // nothing more than set the object's isRunning property to
  // false. Finally, join the thread to make sure that it finishes properly.
  udpThread->stop();
  udpThread->join();
  delete udpThread;
  if (doMcast) {
    mcastThread->stop();
    mcastThread->join();
    delete mcastThread;
  }
  if (perfDaemonIP) {
    exitSendStats();
    delete[] perfDaemonIP;
  }

#if 0
  QUANTAnet_udp_c *aClient;
  for (int i = 0; i < itsNumberOfClients; i++) {
    aClient = itsBunchOfClients[i];
    if (aClient) {
      delete aClient;
    }
  }
#endif
  // css: This list of clients is initialized to contain the
  // maximum number of clients available. It's also initialized
  // using the array new operator. Therefore, the array delete
  // operator must be used in order to properly free memory.
  delete[] itsBunchOfClients;

  if (itsMainUDPSocket) {
    delete itsMainUDPSocket;
  }
  if (itsMcastSocket)
    delete itsMcastSocket;

  delete iomutex;

  //JASON
  if (inBufferUDP)
    delete[] inBufferUDP;
  if (inBufferMCAST)
    delete[] inBufferMCAST;
}

void QUANTAnet_udpReflector_c :: setMulticastTTL(int ttl)
{
	itsMcastTTL = ttl;
}

void QUANTAnet_udpReflector_c :: setIncomingPort(unsigned short port)
{
	itsUDPPort = port;
}

void QUANTAnet_udpReflector_c :: setTimeOutTime(int secs)
{
	itsTimeOutTime = secs;
}

void QUANTAnet_udpReflector_c::setBlockingTimeout(const int& seconds)
{
  if (itsMainUDPSocket != NULL) {
    itsMainUDPSocket->setTimeOut(seconds);
  }
}

void QUANTAnet_udpReflector_c :: setForcedDestination(char *ip, unsigned short port)
{
	itsForcedUDPPort = port;
	strcpy(itsForcedUDPIP,ip);
	doForcedUDP = 1;
}

void QUANTAnet_udpReflector_c:: setMulticastAddress(char *ip, unsigned short port)
{
	doMcast  = 1;
	itsMcastPort = port;
	strcpy(itsMcastIP, ip);
}

int QUANTAnet_udpReflector_c :: init()
{

	if (itsInit) {
		printf("QUANTAnet_udpReflector_c :: init: Initialization already done before\n");
		return QUANTAnet_udpReflector_c::FAILED;
	}

	itsInit = 1;

	itsMainUDPSocket = new QUANTAnet_udp_c;

	if (!itsMainUDPSocket) {
		printf("QUANTAnet_udpReflector_c :: init: UDP cannot allocate memory for socket.\n");
		return QUANTAnet_udpReflector_c::FAILED;
	}

	if (itsMainUDPSocket->init(itsUDPPort) == 0) {
		printf("QUANTAnet_udpReflector_c :: init: UDP Init failed.\n");
		return QUANTAnet_udpReflector_c::FAILED;
	}

	// JASON
	//	itsMainUDPSocket->makeNonBlocking();

	if (initClientList() == 0) {
		printf("QUANTAnet_udpReflector_c :: init: unable to allocate memory.\n");
		return QUANTAnet_udpReflector_c::FAILED;
	}


	if (doForcedUDP) {
		itsForcedUDP = itsMainUDPSocket->clone();
		itsForcedUDP->setSendAddress(itsForcedUDPIP, itsForcedUDPPort);
		addClient(itsForcedUDP);
	}

	itsMcastSocket = new QUANTAnet_mcast_c;

	if (!itsMcastSocket) {
		printf("QUANTAnet_udpReflector_c :: init: unable to allocate memory for multicast.\n");
		return QUANTAnet_udpReflector_c::FAILED;
	}

	if (doMcast) {
		itsMcastSocket->init(itsMcastIP, itsMcastPort, itsMcastTTL);

		// JASON
		// Make all sockets blocking
		//		itsMcastSocket->makeNonBlocking();
	}

	//JASON /////////////////////////////////////////////////////
	// Setup mutex so that udp and multicast reflectors don't try to use
	// the same sockets at the same time during reflection.

	iomutex = new QUANTAts_mutex_c;
	if (!iomutex) {
		printf("QUANTAnet_udpReflector_c :: init: unable to allocate memory for mutex.\n");
		return QUANTAnet_udpReflector_c::FAILED;
	}

	udpThread = new QUANTAts_thread_c;
	if (!udpThread) {
		printf("QUANTAnet_udpReflector_c :: init: unable to allocate memory for udp thread.\n");
		return QUANTAnet_udpReflector_c::FAILED;
	}
  udpThread->start();
	udpThread->create(threadUDP, (void*) this);

	// css: The reflector should spawn a multicasting thread if
	// and only if it's supposed to actually do
	// multicasting. Otherwise, this thread will churn a lot of
	// cpu time.
	if (doMcast) {
	  mcastThread = new QUANTAts_thread_c;
	  if (!mcastThread) {
	    printf("QUANTAnet_udpReflector_c :: init: " \
		   "unable to allocate memory for mcast thread.\n");
	    return QUANTAnet_udpReflector_c::FAILED;
	  }
    mcastThread->start();
	  mcastThread->create(threadMCAST, (void*) this);
	}
	/////////////////////////////////////////////////////////////


	return QUANTAnet_udpReflector_c::OK;
}

int QUANTAnet_udpReflector_c :: addClient(QUANTAnet_udp_c *udp)
{
	if (itsNumberOfClients >= QUANTANET_UDP_REFLECTOR_MAX_NUM_CLIENTS)
		return 0;

	itsBunchOfClients[itsNumberOfClients] = udp;
	/*
    If perfdaemon is alive and initSendStats has been done already by the other clients, then this is a new client which has connected to the reflector and a initSendStats has to be done for this client too
     */
    if (perfDaemonIP)
    {
        if (itsBunchOfClients[itsNumberOfClients]->initSendStats(perfDaemonIP, perfDaemonPort) != QUANTAnet_perfMonitor_c::OK)
            perfDaemonStatus = PERFDAEMON_NOT_ALIVE;               
    }
    itsAliveClients[itsNumberOfClients] = getTimeInSecs();
	itsNumberOfClients++;
	return 1;
}

int QUANTAnet_udpReflector_c :: initClientList()
{
	itsNumberOfClients = 0;

	itsAliveClients = new double[QUANTANET_UDP_REFLECTOR_MAX_NUM_CLIENTS];
	if (!itsAliveClients) return 0;

	itsBunchOfClients = new QUANTAnet_udp_c*[QUANTANET_UDP_REFLECTOR_MAX_NUM_CLIENTS];
	if (!itsBunchOfClients) return 0;

	for (int i = 0; i < QUANTANET_UDP_REFLECTOR_MAX_NUM_CLIENTS; i++){
		itsAliveClients[i] = 0;
		itsBunchOfClients[i] = NULL;
	}
	return 1;
}

void QUANTAnet_udpReflector_c:: intercept(int (*callback)(QUANTAnet_udp_c*, char** buffer, int* datasize, void* userData), void* userData)
{
    interceptCallback = callback;
    interceptUserData = userData;
}

//JASON
//This is the multicast thread.
void * QUANTAnet_udpReflector_c :: threadMCAST(void * inData)
{
  QUANTAnet_udpReflector_c *theReflector = (QUANTAnet_udpReflector_c *) inData;
  while (theReflector->mcastThread->isRunning())
    theReflector->processMCAST();
  return NULL;
}

//JASON
//Process multicast packets by reflecting them out to UDP
int QUANTAnet_udpReflector_c::processMCAST()
{
  if (doMcast) {
    int realSize = itsMcastSocket->receive(inBufferMCAST, QUANTANET_UDP_REFLECTOR_MAX_MSG_SIZE);
    if (realSize > 0) {
      this->iomutex->lock();
      this->incrementDataRead(realSize);
      this->replyToClients(inBufferMCAST, realSize, QUANTANET_UDP_REFLECTOR_SEND_TO_ALL_UDP_CLIENTS);
      this->iomutex->unlock();
    }
  }
  return OK;
}

//JASON
//This will eventually get phased out since there is no need to do a continuous process call when the udp and multicast connections are handled by the threads.

int QUANTAnet_udpReflector_c::process()
{
  QUANTAsleep(1);
  return OK;
}

//JASON
// This is the UDP handler thread.
void * QUANTAnet_udpReflector_c :: threadUDP(void *inData)
{
  QUANTAnet_udpReflector_c *theReflector = (QUANTAnet_udpReflector_c *) inData;

  while (theReflector->udpThread->isRunning())
    theReflector->processUDP();
  return NULL;
}

//JASON
int QUANTAnet_udpReflector_c :: processUDP()
{

	int currentConnectionIndex;
	int anError = 1;
	int distribution = QUANTAnet_udpReflector_c::OK;

	int realSize = itsMainUDPSocket->receive(inBufferUDP, QUANTANET_UDP_REFLECTOR_MAX_MSG_SIZE, QUANTAnet_udp_c::BLOCKING);
	  if (realSize > 0) {
	    iomutex->lock();
	    incrementDataRead(realSize);
	    
	    currentConnectionIndex = checkClient(itsMainUDPSocket);
	    if (currentConnectionIndex >= 0) {
	      if (doMcast) {
		itsMcastSocket->send(inBufferUDP, realSize);
		incrementDataSent(realSize);
	      }
	      if (interceptCallback)
                distribution = interceptCallback(itsBunchOfClients[currentConnectionIndex], &inBufferUDP,&realSize, interceptUserData);
	      if (distribution != QUANTAnet_udpReflector_c::SKIP_DISTRIBUTION)
		replyToClients(inBufferUDP, realSize, currentConnectionIndex);
	    } else {
	      printf("QUANTAnet_udpReflector_c :: process: Can't handle any more clients\n");
	      anError = 0;
	    }
	    iomutex->unlock();	    
	  } else {
#ifdef WIN32
	    if (WSAGetLastError() != WSAEWOULDBLOCK && WSAGetLastError() != 0) {
	      printf("QUANTAnet_udpReflector_c :: process: UDP RECEIVE ERROR %d\n", WSAGetLastError());
	      anError = 0;
	    }
#else
	    if (errno != EWOULDBLOCK && errno!= 0) {
	      printf("QUANTAnet_udpReflector_c :: process: UDP RECEIVE ERROR %d\n", errno);
	      anError = 0;
	    }
#endif
	  }
	  iomutex->lock();
	  purgeOldClients();
	  iomutex->unlock();
	  return anError;
}

int QUANTAnet_udpReflector_c :: checkClient(QUANTAnet_udp_c *mainUDPSocket)
{

	QUANTAnet_udp_c *eachClient;

	for (int i = 0; i < itsNumberOfClients; i++) {
		eachClient = itsBunchOfClients[i];
		if ((eachClient->getSendIP() == mainUDPSocket->getReceiveIP()) &&
		    (eachClient->getSendPort() == mainUDPSocket->getReceivePort())) {
			itsAliveClients[i] = getTimeInSecs();
			return i;
		}
	}
	eachClient = mainUDPSocket->clone();
	eachClient->copyReceiveAddressToSendAddress();
	if (addClient(eachClient) == 0) {
		return -1;
	}

	printf("QUANTAnet_udpReflector_c: New Client: %d\n",itsNumberOfClients-1);
	eachClient->printInfo();
	return itsNumberOfClients-1;
}

void QUANTAnet_udpReflector_c :: purgeOldClients()
{
	double clock = getTimeInSecs();

	if ((clock - itsLastTime) < double(itsTimeOutTime)) return;

	// printf("Purging...\n"); fflush(stdout);
	itsLastTime = clock;

	int start = 0;

	// If forced udp skip past the 0th client.
	if (doForcedUDP) start = 1;

	for (int i = start ; i< itsNumberOfClients; i++) {
		if ((clock - itsAliveClients[i]) > double(itsTimeOutTime)) {
			printf("Purging client: %i of %i\n", i, itsNumberOfClients-1);
			itsBunchOfClients[i]->printInfo();

			delete itsBunchOfClients[i];
			
			itsBunchOfClients[i] = itsBunchOfClients[itsNumberOfClients-1];
			itsAliveClients[i] = itsAliveClients[itsNumberOfClients-1];
			itsAliveClients[itsNumberOfClients-1] = 0;
			itsNumberOfClients--;
		}
	}
}

/* Reply to all clients that are conencted to me. */
void QUANTAnet_udpReflector_c :: replyToClients(char *mesg, int size, int exceptClientIndex)
{
	for(int i = 0 ; i<itsNumberOfClients; i++) {
		if (i != exceptClientIndex) {
			itsBunchOfClients[i]->send(mesg, size);
			incrementDataSent(size);
		}
	}
}

void QUANTAnet_udpReflector_c ::enableInstrumentation()
{
    if (itsMainUDPSocket)
        itsMainUDPSocket->enableInstrumentation();
    for(int i=0; i<itsNumberOfClients; i++)
        {
            if(itsBunchOfClients[i])
               itsBunchOfClients[i]->enableInstrumentation();        
        }
}

void QUANTAnet_udpReflector_c ::disableInstrumentation()
{
    if(itsMainUDPSocket)
        itsMainUDPSocket->disableInstrumentation();
    for(int i=0; i<itsNumberOfClients; i++)
        {
            if(itsBunchOfClients[i])
                itsBunchOfClients[i]->disableInstrumentation();
        }
}


void QUANTAnet_udpReflector_c ::showStats(char* streamInfo, char* comment)
{
    this->iomutex->lock();
    if(itsMainUDPSocket)
        itsMainUDPSocket->showStats(streamInfo, comment);
    for(int i=0; i<itsNumberOfClients; i++)
        {
            if(itsBunchOfClients[i])
                itsBunchOfClients[i]->showStats(streamInfo, comment);
        }
    this->iomutex->unlock();
}

int QUANTAnet_udpReflector_c ::logStats(char* streamInfo, char* comment, FILE* filePtr)
{
    int status=1;
    if(itsMainUDPSocket)    
        status = itsMainUDPSocket->logStats(streamInfo, comment, filePtr);
    this->iomutex->lock();
    for(int i=0; i<itsNumberOfClients; i++)
        {
            if(itsBunchOfClients[i])
                status = ((itsBunchOfClients[i]->logStats(streamInfo, comment, filePtr))&& status);
        }
    this->iomutex->unlock();
     if (status)    
        return QUANTAnet_perfMonitor_c::OK;
     else return QUANTAnet_perfMonitor_c::FAILED;                              
}

int QUANTAnet_udpReflector_c ::sendStats(char* streamInfo, char* comment)
{
    int status=1;
    if(itsMainUDPSocket)
        status = itsMainUDPSocket->sendStats(streamInfo, comment);
    for(int i=0; i<itsNumberOfClients; i++)
       {
           if(itsBunchOfClients[i])               
           status = ((itsBunchOfClients[i]->sendStats(streamInfo, comment))&& status);
       }
    if (status)          
      return QUANTAnet_perfMonitor_c::OK;
    else return QUANTAnet_perfMonitor_c::FAILED;                            
}

int QUANTAnet_udpReflector_c ::initSendStats(char* monitorClientIP, int port)
{
    int status=1;
    if (monitorClientIP == NULL) return QUANTAnet_perfMonitor_c::FAILED;
    
    perfDaemonIP = new char[strlen(monitorClientIP) + 1];
    strcpy(perfDaemonIP, monitorClientIP);
    perfDaemonPort = port;
            
    if(itsMainUDPSocket)
        status = itsMainUDPSocket->initSendStats(monitorClientIP, port);
    for(int i=0; i<itsNumberOfClients; i++)
        {
            if(itsBunchOfClients[i])
                status = ((itsBunchOfClients[i]->initSendStats(monitorClientIP, port))&&status);
        }
    if (status)
    {
        perfDaemonStatus = PERFDAEMON_ALIVE;
        return QUANTAnet_perfMonitor_c::OK;
    }    
    else
    {
        perfDaemonStatus = PERFDAEMON_NOT_ALIVE;
        return QUANTAnet_perfMonitor_c::FAILED; 
    }    
}

void QUANTAnet_udpReflector_c ::exitSendStats()
{
    if (itsMainUDPSocket)
        itsMainUDPSocket->exitSendStats();
    for (int i=0; i<itsNumberOfClients; i++)
    {
        if(itsBunchOfClients[i])
            itsBunchOfClients[i]->exitSendStats();
    }
}





























