/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <QUANTA/QUANTAts_condition_c.hxx>

QUANTAts_condition_c :: QUANTAts_condition_c()
{

#ifdef QUANTA_USE_PTHREADS
	pthread_cond_init(&itsCV, NULL);
#endif

#ifdef QUANTA_USE_WINDOWSTHREADS
    itsCV = new _QUANTA_COND_T;
#endif
}

QUANTAts_condition_c :: ~QUANTAts_condition_c()
{
#ifdef QUANTA_USE_PTHREADS
	pthread_cond_destroy(&itsCV);
#endif
#ifdef QUANTA_USE_WINDOWSTHREADS
  if (itsCV) delete itsCV;
#endif
}

int QUANTAts_condition_c :: wait(QUANTAts_mutex_c *mutex)
{
#ifdef QUANTA_USE_PTHREADS
	return pthread_cond_wait(&itsCV, mutex->getMutex());
#endif
#ifdef QUANTA_USE_WINDOWSTHREADS
  // Prevent race conditions on the <waiters_> count.
  WaitForSingleObject(itsCV->waitersLock, INFINITE);
  itsCV->waiters++;
  ReleaseMutex(itsCV->waitersLock);

  int result = 0;

  // We keep the lock held just long enough to increment the count of
  // waiters by one.  Note that we can't keep it held across the call
  // to ACE_OS::sema_wait() since that will deadlock other calls to
  // ACE_OS::cond_signal().
  mutex->unlock();

  // Wait to be awakened by a ACE_OS::cond_signal() or
  // ACE_OS::cond_broadcast().
  result = WaitForSingleObject(itsCV->sema, INFINITE);
  if (result == WAIT_FAILED) return -1;
  
  // Reacquire lock to avoid race conditions on the <waiters_> count.
  WaitForSingleObject(itsCV->waitersLock, INFINITE);

  // We're ready to return, so there's one less waiter.
  itsCV->waiters--;

  int lastWaiter = itsCV->wasBroadcast && itsCV->waiters == 0;

  // Release the lock so that other collaborating threads can make
  // progress.
  ReleaseMutex(itsCV->waitersLock);

  if (result == -1)
    // Bad things happened, so let's just return below.
    /* NOOP */;
  // If we're the last waiter thread during this particular broadcast
  // then let all the other threads proceed.
  else if (lastWaiter)
    SetEvent(itsCV->waitersDone);

  // We must always regain the <external_mutex>, even when errors
  // occur because that's the guarantee that we give to our callers.
  mutex->lock();

  return result;
#endif
}


int QUANTAts_condition_c :: signal()
{
#ifdef QUANTA_USE_PTHREADS
	return pthread_cond_signal(&itsCV);
#endif
#ifdef QUANTA_USE_WINDOWSTHREADS
    // If there aren't any waiters, then this is a no-op.  Note that
    // this function *must* be called with the <external_mutex> held
    // since other wise there is a race condition that can lead to the
    // lost wakeup bug...  This is needed to ensure that the <waiters_>
    // value is not in an inconsistent internal state while being
    // updated by another thread.
    WaitForSingleObject(itsCV->waitersLock, INFINITE);
    int haveWaiters = itsCV->waiters > 0;
    ReleaseMutex(itsCV->waitersLock);

    if (haveWaiters != 0) return !ReleaseSemaphore(itsCV->sema, 1, NULL);
    else return 0; // No-op
#endif
}

int QUANTAts_condition_c :: broadcastSignal()
{
#ifdef QUANTA_USE_PTHREADS
	return pthread_cond_broadcast(&itsCV);
#endif
#ifdef QUANTA_USE_WINDOWSTHREADS
    // The <external_mutex> must be locked before this call is made.

    // This is needed to ensure that <waiters_> and <was_broadcast_> are
    // consistent relative to each other.
    WaitForSingleObject(itsCV->waitersLock, INFINITE);
    int haveWaiters = 0;

    if (itsCV->waiters > 0) {
    // We are broadcasting, even if there is just one waiter...
    // Record the fact that we are broadcasting.  This helps the
    // cond_wait() method know how to optimize itself.  Be sure to
    // set this with the <waiters_lock_> held.
        itsCV->wasBroadcast = 1;
        haveWaiters = 1;
    }
    ReleaseMutex(itsCV->waitersLock);
    int result = 0;
    if (haveWaiters) {
        // Wake up all the waiters.
        if (ReleaseSemaphore(itsCV->sema, 1, NULL) == 0) result = -1;
            // Wait for all the awakened threads to acquire their part of
            // the counting semaphore.
        else {
            switch (WaitForSingleObject(itsCV->waitersDone, INFINITE)) {
                case WAIT_OBJECT_0:
                    result = 0;

                default:
                    result = -1;
            }

            if (result == -1) result = -1;
        }
        
        // This is okay, even without the <waiters_lock_> held because
        // no other waiter threads can wake up to access it.
        itsCV->wasBroadcast = 0;
    }
    return result;
#endif
}

QUANTA_COND_T *QUANTAts_condition_c::getCondition()
{
	return &itsCV;
}
