/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef WIN32
#include <sys/timeb.h>
#include <math.h>
#endif

#if defined(WIN32) && defined(_MSC_VER)
#include <direct.h>
#include <QUANTA/gettimeofday.hxx>
#else
#include <sys/time.h>
#include <unistd.h>
#endif

#include <QUANTA/keytool.hxx>
#include <QUANTA/QUANTAinit.hxx>
#include <QUANTA/QUANTAglobals.hxx>

#ifdef WIN32
#define _MAX_MAKE_DIR_SIZE _MAX_PATH
#define S_ISDIR(mode)   ((mode&S_IFMT) == S_IFDIR) 
#else
#define _MAX_MAKE_DIR_SIZE 1024
#endif

// Create a directory given a path. This function unlike mkdir will
// create each directory along the path unless it hits a non-directory
// file.
// Returns -1 if mkdir failed somewhere.
// Returns -2 if a file rather than a dir already exists.
int makeDir(char *dir)
{
	struct stat statBuf;
	char *runningDir = dir;
	char filename[_MAX_MAKE_DIR_SIZE];

	char* firstSlash;
	char* last = dir + strlen(dir);
	char gather[_MAX_MAKE_DIR_SIZE];

	memset(gather, 0, _MAX_MAKE_DIR_SIZE);

	// Look for first /
#ifdef WIN32
  // Check for ":\" in the name.
	firstSlash = strstr(dir, ":\\");
	if (firstSlash != NULL) {
		// Skip over it and find the next slash.
		firstSlash += 2;
		firstSlash = strchr(firstSlash, '\\');
	} else {
		firstSlash  = strchr(dir,'\\');
	}
#else
	firstSlash  = strchr(dir,'/');
#endif

	// If exists then process everything before it.
	while (firstSlash) {

		// Copy the filename before the /
		memset(filename, 0, _MAX_MAKE_DIR_SIZE);

		if (firstSlash != runningDir) {
			memcpy(filename, runningDir, (firstSlash - runningDir));
			// Accumulate the current path
			strcat(gather,filename);
		} 
		else 
#ifdef WIN32
			strcat(gather,"\\");
#else
			strcat(gather,"/");
#endif

		// See if it exists.
		if (stat(gather, &statBuf)) {

			// If not then create it.
#if defined(WIN32) && defined(_MSC_VER)
			if (_mkdir(gather) == -1) return -1;
#else
			if (mkdir(gather,0777) == -1) return -1;
#endif

		} else {
			// Otherwise if it exists and it is not a directory quit
			if (!S_ISDIR(statBuf.st_mode)) return -2;
		}

#ifdef WIN32
		if (firstSlash != runningDir)
			strcat(gather,"\\");
#else
		if (firstSlash != runningDir)
			strcat(gather,"/");
#endif

		runningDir = firstSlash + 1;
#ifdef WIN32
		firstSlash  = strchr(runningDir,'\\');
#else
		firstSlash  = strchr(runningDir,'/');
#endif
	}

	// At the end create the directory for the last part of the path.
	if (last > runningDir) {
		memset(filename, 0, _MAX_MAKE_DIR_SIZE);
		memcpy(filename, runningDir, (last - runningDir));
		strcat(gather,filename);

		if (stat(gather, &statBuf)) {
			printf("QUANTA DB MAKE DIR: %s\n",gather);
#if defined(WIN32) && defined(_MSC_VER)
			if (_mkdir(gather) == -1) return -1;
#else
			if (mkdir(gather,0777) == -1) return -1;
#endif
		} else {
			if (!S_ISDIR(statBuf.st_mode)) return -2;
		}
	}
	return 0;
}

keyEntity_c::keyEntity_c()
{
  permissionLockStatus = keyEntity_c::UNLOCKED;
  permissionPassword[0] = '\0';

  memset(filename, 0, _kt_nameSize);
  memset(pathname, 0, _kt_nameSize);
  memset(permissionPassword, 0, _kt_nameSize);

  firstTimeKey = 0;
  timeStamp = 0;
  auxTimeStamp = 0;
  data = NULL;
  internalRealDataSize = userViewDataSize = 0;
  meta = NULL;
  metaSize = 0;

  //printf("NEW KEYENTITY\n");
#ifdef QUANTA_THREAD_SAFE
  mutex = new QUANTAts_mutex_c;
#endif
}


// Show profile of the manager. So far it only shows the hash table's fullness.
void keyToolManager_c :: showProfile()
{
	printf("Hash profile:\n");
	if (keyDB)
	keyDB->showHashProfile();
}

// Tool manager constructor must create a working root directory in which all
// the data hangs.
keyToolManager_c :: keyToolManager_c(char *dir)
{
	keyDB = new QUANTAmisc_hashDict<keyEntity_c*, md5Key_c>();
	strcpy(rootdirectory,dir);
#if defined(WIN32) && defined(_MSC_VER)
	_mkdir(rootdirectory);
#else
	mkdir(rootdirectory,0777);
#endif

#ifdef PTHREAD_SAFE
		pthread_mutex_init(&mutex,NULL);
#endif
#ifdef QUANTA_THREAD_SAFE
		mutex = new QUANTAts_mutex_c;
#endif

}

keyToolManager_c ::~keyToolManager_c()
{
#ifdef PTHREAD_SAFE
		pthread_mutex_destroy(&mutex);
#endif
#ifdef QUANTA_THREAD_SAFE
		delete mutex;
#endif
}

// Check to see if a key exists at all.
int keyToolManager_c :: existsKey(char *path, char *name)
{
	if (!name) return 0;
	char fullname[_kt_idSize];

#ifdef WIN32
	if (!path)
		sprintf(fullname, "%s\\%s",rootdirectory, name);
	else
		sprintf(fullname,"%s\\%s\\%s",rootdirectory,path,name);
#else
	if (!path)
		sprintf(fullname, "%s/%s",rootdirectory, name);
	else
		sprintf(fullname,"%s/%s/%s",rootdirectory,path,name);
#endif

	FILE *fp;
	fp = fopen(fullname,"r");
	if (fp) {
		fclose(fp);
		return 1;
	} else {
		fclose(fp);
		return 0;
	}

}

// Get key from db manager or possibly from disk if necessary.
keyToolKey_c* keyToolManager_c :: getKey(char *path, char *name, status_t *status)
{

	if (!name){
		*status = FAILED;
		return NULL;
	}

	// First see if we can create the subdirectories leading upto the path.
	int mkdirstatus;
	if (path) {
		char fullpath[_kt_nameSize];

#ifdef WIN32
		sprintf(fullpath, "%s\\%s",rootdirectory, path);
#else
		sprintf(fullpath, "%s/%s",rootdirectory, path);
#endif
		mkdirstatus = makeDir(fullpath);
		if (mkdirstatus == -1) {
			printf("Unable to create sub-directory %s for data store.\n", fullpath);
			*status = FAILED;
			return NULL;
		}

		if (mkdirstatus == -2) {
			printf("Subdirectory path %s name already held by a key.\n",fullpath);
			*status = FAILED;
			return NULL;
		}
	}

	md5Key_c keyId;
	char fullname[1024];
#ifdef WIN32
	if (!path) sprintf(fullname,"%s",name);
	else sprintf(fullname,"%s\\%s",path,name);
#else
	if (!path) sprintf(fullname,"%s",name);
	else sprintf(fullname,"%s/%s",path,name);
#endif
	keyId.hash(fullname, strlen(fullname));

	keyToolKey_c *aKey;

	// Look for the key in the database.
	keyEntity_c* aKeyEntity;
	keyDB->find(keyId, aKeyEntity);

	// If key already exists in the active DB then create a new wrapper for it.
	if (aKeyEntity) {
		aKey = new keyToolKey_c(this);
		if (!aKey) {
			fprintf(stderr,"keyToolManager: Cannot allocate memory for key.");
			return NULL;
		}
		aKeyEntity->keyId = keyId;
		aKey->itsKeyId = keyId;

		*status = KEY_ALREADY_IN_MANAGER;
		return aKey;
	} else {
		// Else try and load the key from disk.

		FILE *fp;
		char keyIdName[_kt_idSize];
		int realDataSize = 0;
		int uVDataSize = 0;

		char *metadata = NULL;
		int msize = 0;

		double timeStamp, auxTimeStamp;
		
#ifdef WIN32
		if (path)
			sprintf(keyIdName,"%s\\%s\\%s._meta",rootdirectory,path,name);
		else
			sprintf(keyIdName,"%s\\%s._meta",rootdirectory,name);
#else
		if (path)
			sprintf(keyIdName,"%s/%s/%s._meta",rootdirectory,path,name);
		else
			sprintf(keyIdName,"%s/%s._meta",rootdirectory,name);
#endif
				
		// Open the meta file. If it does exist then
		// read its contents.
		fp = fopen(keyIdName,"r");
		if (fp) {
			msize = 0;
			realDataSize = uVDataSize = 0;
			fscanf(fp,"%d %lf %lf %d",&realDataSize, &timeStamp, &auxTimeStamp, &msize);
			uVDataSize = realDataSize;

			// Extract password status & password
			int passStatus, passLen;
			char password[_kt_nameSize];
			fscanf(fp,"%d",&passStatus);

			// Add a space in scanf to remote the blank space
			fscanf(fp,"%d ",&passLen);
			fread(password,passLen,1,fp);

			// Extract metadata
			if (msize) {
				metadata = new char[msize];
				fread(metadata,msize,1,fp);
			}

			fclose(fp);

#ifdef WIN32
			if (path)
				sprintf(keyIdName,"%s\\%s\\%s",rootdirectory, path,name);
			else
				sprintf(keyIdName,"%s\\%s",rootdirectory, name);
#else
			if (path)
				sprintf(keyIdName,"%s/%s/%s",rootdirectory, path,name);
			else
				sprintf(keyIdName,"%s/%s",rootdirectory, name);
#endif

			char *buffer = NULL;
			if (realDataSize) {

				printf("QUANTA: Database Loading... %s\n", keyIdName);

				fp = fopen(keyIdName,"r");
				buffer = new char[realDataSize];

				if (!buffer) {
					fprintf(stderr,"keyToolManager: Cannot allocate buffer to hold key data.");
					if (fp)
					fclose(fp);
					*status = FAILED;
					return NULL;
				}
				if (fp)
				if (realDataSize)
					fread(buffer,realDataSize,1,fp);

				if (fp)
				fclose(fp);
				printf("QUANTA: Database Done Loading... %s\n", keyIdName);

			}

			// Create a new key entity and add it to the key database.
			aKeyEntity = new keyEntity_c;

			if (!aKeyEntity) {
				fprintf(stderr,"keyToolManager: Cannot allocate memory for key.");
				*status = FAILED;
				return NULL;
			}

			aKeyEntity->data = buffer;
			aKeyEntity->internalRealDataSize = realDataSize;
			aKeyEntity->userViewDataSize = uVDataSize;
			aKeyEntity->meta = metadata;
			aKeyEntity->metaSize = msize;
			aKeyEntity->timeStamp = timeStamp;
			aKeyEntity->auxTimeStamp = auxTimeStamp;
			aKeyEntity->permissionLockStatus = (keyEntity_c::permission_lock_status_t) passStatus;
			strcpy(aKeyEntity->permissionPassword,password);

			// Copy path and filename into entity. Needed for later when we commit.
			if (path)
				strcpy(aKeyEntity->pathname, path);
			else memset(aKeyEntity->pathname, 0, _kt_nameSize);

			strcpy(aKeyEntity->filename, name);

			// Place entity into database.
			keyDB->enter(keyId, aKeyEntity);

			// Create a new key wrapper to wrap the entity.
			aKey = new keyToolKey_c(this);
			if (!aKey) {
				fprintf(stderr,"keyToolManager: Cannot allocate memory for key.");
				if (buffer) delete buffer;
				delete aKeyEntity;
				*status = FAILED;
				return NULL;
			}

			//			aKey->itsKeyEntity = aKeyEntity;
			//			aKey->itsKeyEntity->keyId = keyId;
			aKey->itsKeyId = keyId;
			aKeyEntity->keyId = keyId;

			*status = KEY_FOUND_IN_FILE;
			return aKey;

		} else {

			// If it does not exist then assume a new key.
			// create a new key entity.
			aKeyEntity = new keyEntity_c;
			if (!aKeyEntity) {
				fprintf(stderr,"keyToolManager: Cannot allocate memory for key.");
				*status = FAILED;
				return NULL;
			}

			// Copy path and filename into entity. Needed for later when we commit.
			if (path)
				strcpy(aKeyEntity->pathname, path);
			else memset(aKeyEntity->pathname, 0, _kt_nameSize);

			strcpy(aKeyEntity->filename, name);


			aKey = new keyToolKey_c(this);
			if (!aKey) {
				fprintf(stderr,"keyToolManager: Cannot allocate memory for key.");
				delete aKeyEntity;
				*status = FAILED;
				return NULL;
			}

			keyDB->enter(keyId,aKeyEntity);

			aKeyEntity->firstTimeKey = 1;
			//			aKey->itsKeyEntity = aKeyEntity;
			//			aKey->itsKeyEntity->keyId = keyId;
			aKey->itsKeyId = keyId;
			aKeyEntity->keyId = keyId;

			*status = KEY_COMPLETELY_NEW;

			return aKey;
			
		}

	
	}

}


// Get key from key database manager. This assumes that you have already
// retrieved the key at least once.
keyToolKey_c* keyToolManager_c :: getKey(md5Key_c keyId, status_t *status)
{
	keyToolKey_c *aKey;

	// Look for the key in the database.
	keyEntity_c* aKeyEntity;
	keyDB->find(keyId, aKeyEntity);

	// If key already exists in the active DB then create a new wrapper for it.
	if (aKeyEntity) {
		aKey = new keyToolKey_c(this);
		if (!aKey) {
			fprintf(stderr,"keyToolManager: Cannot allocate memory for key.");
			return NULL;
		}
		//aKey->itsKeyEntity = aKeyEntity;
		//aKey->itsKeyEntity->keyId = keyId;
		aKey->itsKeyId = keyId;
		aKeyEntity->keyId = keyId;

		*status = KEY_ALREADY_IN_MANAGER;
		return aKey;
	} else {
		*status = FAILED;
		return NULL;
	}


}

int _commitEntity(md5Key_c key, keyEntity_c *obj, void *manager, void* dum2, void* dum3,
		  void *& retvalue)
{
	((keyToolManager_c*) manager)->commit(obj);

	// Must return 1 in order to keep applyToAll iterating through every key item.
	return 1;
}

// Commit the entire database.
keyToolManager_c::status_t keyToolManager_c :: commit()
{
	void *retdummy;
	keyDB->applyToAll(_commitEntity, (void *)this,NULL,NULL,retdummy);

	return OK;
}


const int _KEYTOOL_WRITE_CHUNK_SIZE = 256*1024;
void gentlerWrite(char *data, int size, FILE *fp)
{
	int numChunks = size / _KEYTOOL_WRITE_CHUNK_SIZE;
	for(int i = 0; i < numChunks; i++) {
		fwrite(data, _KEYTOOL_WRITE_CHUNK_SIZE,1,fp);
		data += _KEYTOOL_WRITE_CHUNK_SIZE;
		QUANTAusleep(1000);
	}
	int remain = size - (numChunks * _KEYTOOL_WRITE_CHUNK_SIZE);
	if (remain) fwrite(data, remain,1,fp);

}

/* Commit one entry of the database. The commit saves both the meta
data and data to temporary files suffixed with '_'. Once done
the new files are renamed to replace the old files.
*/

keyToolManager_c::status_t keyToolManager_c :: commit(keyEntity_c *aKeyEntity)
{
	FILE *fp;
	char keyIdName[_kt_nameSize], keyIdNameTemp[_kt_nameSize],
		keyIdNameSave[_kt_nameSize];

	char metakeyIdName[_kt_nameSize], metakeyIdNameTemp[_kt_nameSize],
		metakeyIdNameSave[_kt_nameSize];

#ifdef WIN32
	if (aKeyEntity->pathname[0]) {
		sprintf(metakeyIdName,"%s\\%s\\%s._meta",
			rootdirectory, 
			aKeyEntity->pathname,
			aKeyEntity->filename);

		sprintf(metakeyIdNameTemp,"%s\\%s\\%s._meta_",
			rootdirectory, 
			aKeyEntity->pathname,
			aKeyEntity->filename);

		sprintf(metakeyIdNameSave,"%s\\%s\\%s._savemeta",
			rootdirectory, 
			aKeyEntity->pathname,
			aKeyEntity->filename);

		sprintf(keyIdName,"%s\\%s\\%s",
			rootdirectory, 
			aKeyEntity->pathname,
			aKeyEntity->filename);

		sprintf(keyIdNameTemp,"%s\\%s\\%s_",
			rootdirectory, 
			aKeyEntity->pathname,
			aKeyEntity->filename);

		sprintf(keyIdNameSave,"%s\\%s\\%s.save",
			rootdirectory, 
			aKeyEntity->pathname,
			aKeyEntity->filename);
	} else {
		sprintf(metakeyIdName,"%s\\%s._meta",
			rootdirectory, 
			aKeyEntity->filename);

		sprintf(metakeyIdNameTemp,"%s\\%s._meta_",
			rootdirectory, 
			aKeyEntity->filename);

		sprintf(metakeyIdNameSave,"%s\\%s._savemeta",
			rootdirectory, 
			aKeyEntity->filename);

		sprintf(keyIdName,"%s\\%s",
			rootdirectory, 
			aKeyEntity->filename);

		sprintf(keyIdNameTemp,"%s\\%s_",
			rootdirectory, 
			aKeyEntity->filename);

		sprintf(keyIdNameSave,"%s\\%s._save",
			rootdirectory, 
			aKeyEntity->filename);
#else
	if (aKeyEntity->pathname[0]) {
		sprintf(metakeyIdName,"%s/%s/%s._meta",
			rootdirectory, 
			aKeyEntity->pathname,
			aKeyEntity->filename);

		sprintf(metakeyIdNameTemp,"%s/%s/%s._meta_",
			rootdirectory, 
			aKeyEntity->pathname,
			aKeyEntity->filename);

		sprintf(metakeyIdNameSave,"%s/%s/%s._savemeta",
			rootdirectory, 
			aKeyEntity->pathname,
			aKeyEntity->filename);

		sprintf(keyIdName,"%s/%s/%s",
			rootdirectory, 
			aKeyEntity->pathname,
			aKeyEntity->filename);

		sprintf(keyIdNameTemp,"%s/%s/%s_",
			rootdirectory, 
			aKeyEntity->pathname,
			aKeyEntity->filename);

		sprintf(keyIdNameSave,"%s/%s/%s.save",
			rootdirectory, 
			aKeyEntity->pathname,
			aKeyEntity->filename);
	} else {
		sprintf(metakeyIdName,"%s/%s._meta",
			rootdirectory, 
			aKeyEntity->filename);

		sprintf(metakeyIdNameTemp,"%s/%s._meta_",
			rootdirectory, 
			aKeyEntity->filename);

		sprintf(metakeyIdNameSave,"%s/%s._savemeta",
			rootdirectory, 
			aKeyEntity->filename);

		sprintf(keyIdName,"%s/%s",
			rootdirectory, 
			aKeyEntity->filename);

		sprintf(keyIdNameTemp,"%s/%s_",
			rootdirectory, 
			aKeyEntity->filename);

		sprintf(keyIdNameSave,"%s/%s._save",
			rootdirectory, 
			aKeyEntity->filename);
#endif
	}


	// Get the key entity from the database.
	//	keyDB->find(theId, aKeyEntity);


	// Save the meta data.
	// Meta data is simply, data size followed by some user given meta data.
	if (aKeyEntity->firstTimeKey)
	fp = fopen(metakeyIdName,"w");
	else
	fp = fopen(metakeyIdNameTemp,"w");

	if (!fp) {
		fprintf(stderr,"keyToolManager: Cannot write to %s\nCommit aborted.\n", metakeyIdNameTemp);
		return FAILED;
	}

	aKeyEntity->stampTime();

	// Write time stamps
	fprintf(fp,"%d %lf %lf %d",aKeyEntity->internalRealDataSize,
		aKeyEntity->timeStamp,
		aKeyEntity->auxTimeStamp,
		aKeyEntity->metaSize);

	// Write password status
	fprintf(fp," %d",aKeyEntity->permissionLockStatus);
	int passlen = strlen(aKeyEntity->permissionPassword) + 1;
	fprintf(fp," %d ",passlen);
	fwrite(aKeyEntity->permissionPassword, passlen,1,fp);

	if (aKeyEntity->metaSize)
		fwrite(aKeyEntity->meta,aKeyEntity->metaSize,1,fp);

	fclose(fp);

	// Save the actual data.
	if (aKeyEntity->firstTimeKey)
	fp = fopen(keyIdName,"w");
	else
	fp = fopen(keyIdNameTemp,"w");

	if (!fp) {
		fprintf(stderr,"keyToolManager: Cannot write to %s\nCommit aborted.\n", keyIdNameTemp);
		return FAILED;
	}

	fwrite(aKeyEntity->data,aKeyEntity->internalRealDataSize,1,fp);
	//	gentlerWrite(aKeyEntity->data,aKeyEntity->internalRealDataSize,fp);
	fclose(fp);

	if (!aKeyEntity->firstTimeKey) {
		if (rename(metakeyIdName,metakeyIdNameSave) == -1) return FAILED;
		if (rename(metakeyIdNameTemp,metakeyIdName) == -1) {
			rename(metakeyIdNameSave, metakeyIdName);
			return FAILED;
		}

		if (rename(keyIdName,keyIdNameSave) == -1) {
			rename(metakeyIdNameSave, metakeyIdName);
			return FAILED;
		}
		if (rename(keyIdNameTemp, keyIdName) == -1) {
			rename(metakeyIdNameSave, metakeyIdName);
			rename(keyIdNameSave, keyIdName);
			return FAILED;
		}

		unlink(keyIdNameSave);
		unlink(metakeyIdNameSave);
	}
	aKeyEntity->firstTimeKey = 0;
	return OK;
}

keyToolManager_c::status_t keyToolManager_c :: swapContents(keyToolKey_c *theKey1, char* password1, keyToolKey_c *theKey2, char* password2)
{
	char thePassword[_kt_nameSize];
	int pass=1;
	if (theKey1->getPermissionLockStatus(thePassword, _kt_nameSize)
	    == keyToolKey_c::LOCKED) {
		if (password1) {
			if (strcmp(password1,thePassword)) pass=0;
		} else pass=0;
	}

	if (theKey2->getPermissionLockStatus(thePassword, _kt_nameSize)
	    == keyToolKey_c::LOCKED) {
		if (password2) {
			if (strcmp(password1,thePassword)) pass=0;
		} else pass = 0;
	}
	if (pass) {
		return swapContents(theKey1, theKey2);
	}
	return LOCKED;
}

// Swap the contents of the keys. This assumes that the keys have already
// been loaded into the manager. Latest addition of (5/18/98) also swaps meta data.
keyToolManager_c::status_t keyToolManager_c :: swapContents(keyToolKey_c *thekey1, keyToolKey_c *thekey2)
{
	
	md5Key_c keyId1, keyId2;
	keyEntity_c *key1, *key2;

	
	keyDB->find(thekey1->itsKeyId, key1);
	if (key1 == NULL) return FAILED;
	keyDB->find(thekey2->itsKeyId, key2);
	if (key2 == NULL) return FAILED;

	// Swap data

	char *temp;
	temp = 	key1->data;
	key1->data = key2->data;
	key2->data = temp;

	int tempSize = key1->internalRealDataSize;
	key1->internalRealDataSize = key2->internalRealDataSize;
	key2->internalRealDataSize = tempSize;

	tempSize = key1->userViewDataSize;
	key1->userViewDataSize = key2->userViewDataSize;
	key2->userViewDataSize = tempSize;

	// Swap meta data
	temp = key1->meta;
	key1->meta = key2->meta;
	key2->meta = temp;

	tempSize = key1->metaSize;
	key1->metaSize = key2->metaSize;
	key2->metaSize = tempSize;

	return OK;
}

// Remote a key from the db manager. Currently this does not remove the
// persistent file.
keyToolManager_c::status_t keyToolManager_c :: removeKey(keyToolKey_c *thekey)
{

	keyEntity_c *key;

	keyDB->find(thekey->itsKeyId,key);

	if (!key) return FAILED;

	keyDB->remove(thekey->itsKeyId);


	/* 3/10/99 not needed since added to destructor
	if (key->data)
	delete key->data;
	*/
	//	printf("removeKey DELETING\n");
	delete key;


	return OK;
	// Should also delete affiliated file.
}

// Remove a key permanently from the db manager including the
// persistent file.
keyToolManager_c::status_t keyToolManager_c :: removeKeyPermanently(keyToolKey_c *thekey)
{

	keyEntity_c *key;

	keyDB->find(thekey->itsKeyId, key);

	if (!key) return FAILED;

	char keyName[_kt_nameSize];

#ifdef WIN32
	if (key->pathname[0]) {
		sprintf(keyName,"%s\\%s\\%s._meta",rootdirectory, key->pathname, key->filename);
		unlink(keyName);

		sprintf(keyName,"%s\\%s\\%s",rootdirectory,key->pathname, key->filename);
		unlink(keyName);
	}else {
		sprintf(keyName,"%s\\%s._meta",rootdirectory, key->filename);
		unlink(keyName);

		sprintf(keyName,"%s\\%s",rootdirectory, key->filename);
#else
	if (key->pathname[0]) {
		sprintf(keyName,"%s/%s/%s._meta",rootdirectory, key->pathname, key->filename);
		unlink(keyName);

		sprintf(keyName,"%s/%s/%s",rootdirectory,key->pathname, key->filename);
		unlink(keyName);
	}else {
		sprintf(keyName,"%s/%s._meta",rootdirectory, key->filename);
		unlink(keyName);

		sprintf(keyName,"%s/%s",rootdirectory, key->filename);
#endif
		unlink(keyName);
	}


	keyDB->remove(thekey->itsKeyId);

	/* 3/10/99 not needed since added to destructor
	if (key->data)
	delete key->data;
	*/
	//	printf("removeKeyPermanenetly DELETING\n");

	delete key;

	return OK;
	// Should also delete affiliated file.
}

// Get data field of the key. If user's view of the data is zero in size
// then return NULL
char *keyToolKey_c :: getData()
{
	keyEntity_c *keyEnt;
	itsManager->keyDB->find(itsKeyId, keyEnt);
	if ((keyEnt) && (keyEnt->userViewDataSize))
	return keyEnt->data;
	else return NULL;

}

// Resize key data buffer.
#if(0)
keyToolKey_c::status_t keyToolKey_c :: resize(int newsize)
{
	char *theData;
	char *newData;

	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	if (!itsKeyEntity)  return FAILED;

	theData = itsKeyEntity->data;

	// If there is data already there,
	if (theData) {

		// Allocate the memory, copy its contents.
		newData = new char[newsize];
		if (!newData) return FAILED;
		memcpy(newData,theData,max(newsize,itsKeyEntity->size));
		itsKeyEntity->data = newData;

		// Then delete the old data.
		delete theData;
		itsKeyEntity->size = newsize;
		return OK;
	} else {

		// If the data does not exist, simply
		// allocate memory for it.

		if (newsize) {
			newData = new char[newsize];
			if (!newData) return FAILED;
			itsKeyEntity->data = newData;
			itsKeyEntity->size = newsize;
			return OK;
		} else {
			// Case when resizing is set to 0 bytes.
			newData = NULL;
			itsKeyEntity->data = NULL;
			itsKeyEntity->size = 0;
			return OK;
		}
	}

}
#endif

keyToolKey_c :: keyToolKey_c (keyToolManager_c *ktm)
{
	itsManager = ktm;
}

md5Key_c keyToolKey_c :: getKeyId()
{
	return itsKeyId;
}

// This is a very expensive operation for a large buffer!
md5Key_c keyToolKey_c :: getDataCheckSum()
{
	keyEntity_c *keyEnt;
	md5Key_c checkSum;

	itsManager->keyDB->find(itsKeyId, keyEnt);
	if ((keyEnt) && (keyEnt->userViewDataSize)) {
		checkSum.hash(keyEnt->data, keyEnt->userViewDataSize);
	}
	return checkSum;
}

keyToolKey_c::status_t keyToolKey_c :: trash()
{
	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	if (!itsKeyEntity) return FAILED;

	if (itsKeyEntity->data) delete itsKeyEntity->data;
	itsKeyEntity->data = NULL;
	itsKeyEntity->internalRealDataSize = 0;
	itsKeyEntity->userViewDataSize = 0;

	return OK;
}

// This is the same as resize except it does not maintain
// the integrity of the original contents.
/* Allocation works as follows:
If the new size is bigger than the current internal buffer then
a new buffer is allocated.
If the new size is smaller than the user view of the data size
is reduced and the buffer remains unchanged.
*/
keyToolKey_c::status_t keyToolKey_c :: alloc(int newsize)
{
	char *theData, *newData;

	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	if (!itsKeyEntity) return FAILED;

	theData = itsKeyEntity->data;
	if (theData) {

		// If new size is bigger than internal size then allocate bigger pool.
		if (newsize > itsKeyEntity->internalRealDataSize) {
			if (newsize) {
				newData = new char[newsize];
				if (!newData) return FAILED;
				itsKeyEntity->data = newData;
				itsKeyEntity->internalRealDataSize = newsize;
				itsKeyEntity->userViewDataSize = newsize;
			} else {
				// If newsize is zero...
				newData = NULL;
				itsKeyEntity->data = newData;
				itsKeyEntity->internalRealDataSize = 0;
				itsKeyEntity->userViewDataSize = 0;
			}

			delete theData;
		} else {
			// Otherwise set the user's view of the data size to the newsize.
			itsKeyEntity->userViewDataSize = newsize;
		}
		return OK;
	} else {

		if (newsize) {
			newData = new char[newsize];
			if (!newData) return FAILED;
			itsKeyEntity->data = newData;
			itsKeyEntity->internalRealDataSize = newsize;
			itsKeyEntity->userViewDataSize = newsize;
			return OK;
		} else {

			// If newsize is 0...
			itsKeyEntity->data = NULL;
			itsKeyEntity->internalRealDataSize = 0;
			itsKeyEntity->userViewDataSize = 0;
			return OK;
		}

	}

}

// Return size of data in key.
int keyToolKey_c :: getDataSize(){
	keyEntity_c *keyEnt;
	itsManager->keyDB->find(itsKeyId, keyEnt);
	if (keyEnt)
	return keyEnt->userViewDataSize;
	else return 0;
}

int keyToolKey_c :: getRealSize() {
	keyEntity_c *keyEnt;
	itsManager->keyDB->find(itsKeyId, keyEnt);
	if (keyEnt)
	return keyEnt->internalRealDataSize;
	else return 0;
}

// Commit current key item to database.
keyToolKey_c::status_t keyToolKey_c :: commit(char *password)
{
	char thePassword[_kt_nameSize];
	int pass=1;
	if (getPermissionLockStatus(thePassword, _kt_nameSize) == keyToolKey_c::LOCKED) {
		if (password) {
			if (strcmp(password,thePassword)) pass=0;
		} else pass=0;
	}
	if (pass) return commit();
	else return keyToolKey_c::LOCKED;
}


// Commit current key item to database.
keyToolKey_c::status_t keyToolKey_c :: commit()
{
	//	return itsManager->commit(itsManager->keyDB->findNth(itsKeyEntity->keyId));
	keyEntity_c *keyEnt;
	itsManager->keyDB->find(itsKeyId, keyEnt);
	if (keyEnt) {
		if (itsManager->commit(keyEnt) == keyToolManager_c::OK) return OK;
	}
	return FAILED;
}

// Get meta data from key.
char* keyToolKey_c :: getMeta(){
	keyEntity_c *keyEnt;
	itsManager->keyDB->find(itsKeyId, keyEnt);
	if (keyEnt)
	return keyEnt->meta;
	else return NULL;
}


int keyToolKey_c :: getMetaSize(){
	keyEntity_c *keyEnt;
	itsManager->keyDB->find(itsKeyId, keyEnt);
	if (keyEnt)
	return keyEnt->metaSize;
	else return 0;
}

// Lock
keyToolKey_c::status_t keyToolKey_c :: permissionLock(char *password)
{
	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	if (!itsKeyEntity) return keyToolKey_c::FAILED;

	// If already locked check password
	if (itsKeyEntity->permissionLockStatus == keyEntity_c::LOCKED) {

		// If correct password then indicate already locked.
		if (strcmp(itsKeyEntity->permissionPassword, password) == 0) return keyToolKey_c::ALREADY_LOCKED;

		// Otherwise cannot unlock.
		else return keyToolKey_c::CANNOT_LOCK;
	} else {

		// If not already locked.

		// Check to make sure there is no null password
		if (password) {
			if (password[0] != '\0') {

				// Now lock it.
				itsKeyEntity->permissionLockStatus = keyEntity_c::LOCKED;
				strcpy(itsKeyEntity->permissionPassword, password);
				return keyToolKey_c::LOCKED;
			}
		}
	}

	// It seems password you passed in was bad so return failed.
	return keyToolKey_c::FAILED;

}

keyToolKey_c::status_t keyToolKey_c :: permissionForceLock(char *password)
{

	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	if (!itsKeyEntity) return keyToolKey_c::FAILED;

	// Check to make sure there is no null password
	if (password) {
		if (password[0] != '\0') {

			// Now lock it.
			itsKeyEntity->permissionLockStatus = keyEntity_c::LOCKED;
			strcpy(itsKeyEntity->permissionPassword, password);
			return keyToolKey_c::LOCKED;
		}
	}

	// It seems password you passed in was bad so return failed.
	return keyToolKey_c::FAILED;

}

keyToolKey_c::status_t keyToolKey_c :: getPermissionLockStatus(char *password, int size)
{
	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	password[0] = '\0';
	if (!itsKeyEntity) return keyToolKey_c::FAILED;

	if (itsKeyEntity->permissionLockStatus == keyEntity_c::LOCKED) {
		if (password)
			strncpy(password, itsKeyEntity->permissionPassword, size);
		return keyToolKey_c::LOCKED;
	}

	return keyToolKey_c::UNLOCKED;
}

keyToolKey_c::status_t keyToolKey_c :: permissionForceUnlock()
{
	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	if (!itsKeyEntity) return keyToolKey_c::FAILED;

	if (itsKeyEntity->permissionLockStatus == keyEntity_c::LOCKED) {
		itsKeyEntity->permissionLockStatus = keyEntity_c::UNLOCKED;
		itsKeyEntity->permissionPassword[0] = '\0';
		return keyToolKey_c::UNLOCKED;
	} else return keyToolKey_c::ALREADY_UNLOCKED;
}

keyToolKey_c::status_t keyToolKey_c :: permissionUnlock(char *password)
{
	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	if (!itsKeyEntity) return keyToolKey_c::FAILED;

	// If already locked check who holds lock
	if (itsKeyEntity->permissionLockStatus == keyEntity_c::LOCKED) {

		// If you have the right password then unlock it
		if (strcmp(itsKeyEntity->permissionPassword, password) == 0) {
			itsKeyEntity->permissionLockStatus = keyEntity_c::UNLOCKED;
			itsKeyEntity->permissionPassword[0] = '\0';
			return keyToolKey_c::UNLOCKED;
		}
		// Otherwise unlock failed
		else return keyToolKey_c::CANNOT_UNLOCK;
	} else 
		return keyToolKey_c::ALREADY_UNLOCKED;
}


keyToolKey_c::status_t keyToolKey_c :: setMeta(char* password, char *data, int msize)
{
	char thePassword[_kt_nameSize];
	int pass=1;
	if (getPermissionLockStatus(thePassword, _kt_nameSize) == keyToolKey_c::LOCKED) {
		if (password) {
			if (strcmp(password,thePassword)) pass=0;
		} else pass=0;
	}
	if (pass) return setMeta(data, msize);
	else return keyToolKey_c::LOCKED;
}

// Set meta data in key.
keyToolKey_c::status_t keyToolKey_c :: setMeta(char *data, int msize)
{
	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	if (!itsKeyEntity) return FAILED;

	if ((msize <= 0)|| (data == NULL)) {
		if (itsKeyEntity->meta) delete itsKeyEntity->meta;
		itsKeyEntity->meta = NULL;
		itsKeyEntity->metaSize = 0;
		return OK;
	}
		
	char *newMeta = new char[msize];
	if (newMeta == NULL) return FAILED;

	memcpy(newMeta,data,msize);
	if (itsKeyEntity->meta) delete itsKeyEntity->meta;

	itsKeyEntity->meta = newMeta;
	itsKeyEntity->metaSize = msize;
	return OK;
}

double keyToolKey_c :: getTimeStamp()
{
	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	if (itsKeyEntity)
	return itsKeyEntity->timeStamp;
	else return 0;
}

void keyToolKey_c :: stampTime() 
{
	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	if (itsKeyEntity)
	itsKeyEntity->stampTime();
}

void keyToolKey_c :: stampAuxTime() 
{
	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	if (itsKeyEntity)
	itsKeyEntity->stampAuxTime();
}

void keyToolKey_c :: setTimeStamp(double theTime)
{
	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	if (itsKeyEntity)
	itsKeyEntity->setTimeStamp(theTime);
}

void keyToolKey_c :: setAuxTimeStamp(double theTime)
{
	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	if (itsKeyEntity)
	itsKeyEntity->setAuxTimeStamp(theTime);
}

double keyToolKey_c :: getAuxTimeStamp()
{
	keyEntity_c *itsKeyEntity;
	itsManager->keyDB->find(itsKeyId, itsKeyEntity);
	if (itsKeyEntity)
	return itsKeyEntity->auxTimeStamp;
	else return 0;
}


double  keyEntity_c :: getTimeInSecs()
{
        const double oneMicroSec = 0.000001;
        static struct timeval timeSec;

#ifdef WIN32
        gettimeofday(&timeSec, NULL);
#else
        static struct timezone tzone;

        gettimeofday(&timeSec, &tzone);
#endif

        return (((double) timeSec.tv_sec) +
                (((double) timeSec.tv_usec) * oneMicroSec));
}

void keyEntity_c :: stampTime()
{
	timeStamp = getTimeInSecs();
}

void keyEntity_c :: stampAuxTime()
{
	auxTimeStamp = getTimeInSecs();
}

void keyEntity_c :: setTimeStamp(double theTime)
{
	timeStamp = theTime;
}

void keyEntity_c :: setAuxTimeStamp(double theTime)
{
	auxTimeStamp = theTime;
}

