/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#if defined(WIN32) && defined(QUANTA_USE_WINDOWSTHREADS)

/* This program is largely based on "Strategies for Implementing POSIX Condition
   Variables on Win32" written by Douglas C. Schmidt and Irfan Pyarali and published
   in C++ Report, Vol. 10, No. 5, June, 1998 issue. You can get the full article from
   "http://www.cs.wustl.edu/~schmidt/win32-cv-1.html". I fixed a few things from the
   code and added pthread_win_cond_destroy function.
*/

#include <pthread_win.hxx>
#include <stdio.h>

int pthread_win_cond_init(struct pthread_win_cond_t *cv)
{
    cv->waiters_count_ = 0;
    cv->wait_generation_count_ = 0;
    cv->release_count_ = 0;
	
	InitializeCriticalSection(&cv->waiters_count_lock_);

    // Create a manual-reset event.
    cv->event_ = CreateEvent (NULL,  // no security
                            TRUE,  // manual-reset
                            FALSE, // non-signaled initially
                            NULL); // unnamed
	if (cv->event_ == NULL) {
		printf("pthread_win_cond_init: cannot create event object.\n");
		return 1;
	}
	return 0;
}

int pthread_win_cond_wait(struct pthread_win_cond_t *cv,
                          QUANTAts_mutex_c *external_mutex)
{
	int my_generation;
    int last_waiter;
    // Avoid race conditions.
    EnterCriticalSection(&cv->waiters_count_lock_);

    // Increment count of waiters.
    cv->waiters_count_++;

    // Store current generation in our activation record.
    my_generation = cv->wait_generation_count_;

    LeaveCriticalSection(&cv->waiters_count_lock_);
    if (external_mutex) external_mutex->unlock();

    for (;;) {
		int wait_done;

        // Wait until the event is signaled.
        WaitForSingleObject(cv->event_, INFINITE);

        EnterCriticalSection(&cv->waiters_count_lock_);
        // Exit the loop when the <cv->event_> is signaled and
        // there are still waiting threads from this <wait_generation>
        // that haven't been released from this wait yet.
        wait_done = cv->release_count_ > 0
                    && cv->wait_generation_count_ != my_generation;
        LeaveCriticalSection(&cv->waiters_count_lock_);

        if (wait_done) break;
	}

    if (external_mutex) external_mutex->lock();
    EnterCriticalSection(&cv->waiters_count_lock_);
    cv->waiters_count_--;
    cv->release_count_--;
    last_waiter = cv->release_count_ == 0;
    LeaveCriticalSection(&cv->waiters_count_lock_);

	// We're the last waiter to be notified, so reset the manual event.
    if (last_waiter) ResetEvent(cv->event_);
	return 0;
}


int pthread_win_cond_signal(struct pthread_win_cond_t *cv)
{
    EnterCriticalSection(&cv->waiters_count_lock_);
    if (cv->waiters_count_ > cv->release_count_) {
        SetEvent (cv->event_); // Signal the manual-reset event.
        cv->release_count_++;
        cv->wait_generation_count_++;
	}
    LeaveCriticalSection(&cv->waiters_count_lock_);
	return 0;
}

int pthread_win_cond_broadcast(struct pthread_win_cond_t *cv)
{
    EnterCriticalSection(&cv->waiters_count_lock_);
    if (cv->waiters_count_ > 0) {  
        SetEvent(cv->event_);
        // Release all the threads in this generation.
        cv->release_count_ = cv->waiters_count_;

        // Start a new generation.
        cv->wait_generation_count_++;
	}
    LeaveCriticalSection(&cv->waiters_count_lock_);
	return 0;
}

int pthread_win_cond_destroy(struct pthread_win_cond_t *cv)
{
    cv->waiters_count_ = 0;
    cv->wait_generation_count_ = 0;
    cv->release_count_ = 0;

	DeleteCriticalSection(&cv->waiters_count_lock_);

    // Create a manual-reset event.
    if (cv->event_) CloseHandle(cv->event_);

	return 0;
}
#endif
