/******************************************************************************
 * QUANTA - A toolkit for High Performance Data Sharing
 * Copyright (C) 2003 Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Direct questions, comments etc about Quanta to cavern@evl.uic.edu
 *****************************************************************************/

#include <stdlib.h>
#ifndef WIN32
#include <signal.h>
#endif

#include <QUANTA.hxx>
#include <QUANTAnet_perfDaemon_c.hxx>
#include "getopt.hxx"

#define MAX_PATH_LENGTH 256

/** This program implements simple performance monitoring daemon program.
 Users can either save all performance messages into a log file using -f
 option and specifying the log file name or reflects all performance messages
 to the performance monitoring client such as QoSiMoto. 
 You can only use two options with this program. -h option will show you
 the descriptions of options that can be used with this program. If -f 
 option is used with a file name, all performance monitoring messages sent
 to this program will be saved into the log file named as the given file name.
 If no option is used at all, this program will send out all performance
 monitoring messages to the registered performance monitoring clients.
 If no performance monitoring client is connected to this program but several
 clients are connected to send out performance monitoring messages, the 
 messages will be consumed by this program and nothing will be sent out.
*/

char LogFileName[MAX_PATH_LENGTH] = "";
FILE* LogFile = NULL;

/** This function is going to be registered so that it must be called when
 this program exits normally. This must be used so that we can flush out all
 performance monitoring messages and save them into the log file. 
*/
void AtExit2(void)
{
    if (LogFile) {
        fflush(LogFile);
	fclose(LogFile);
        LogFile = NULL;
    }
}

#ifndef WIN32
void AtExit(int sigNo)
{
    printf("Exit\n");
    if (LogFile) {
        fflush(LogFile);
	fclose(LogFile);
	LogFile = NULL;
    }
    if (sigNo == SIGSEGV || sigNo == SIGBUS || sigNo == SIGQUIT)
        abort();

    exit(0);
}

void SetupSignals(void)
{
    struct sigaction newact, oldact;

    newact.sa_handler = AtExit;
    sigemptyset(&newact.sa_mask);
    newact.sa_flags = 0;
    sigaction(SIGABRT, &newact, &oldact);

    newact.sa_handler = AtExit;
    sigemptyset(&newact.sa_mask);
    newact.sa_flags = 0;
    sigaction(SIGBUS, &newact, &oldact);

    newact.sa_handler = AtExit;
    sigemptyset(&newact.sa_mask);
    newact.sa_flags = 0;
    sigaction(SIGKILL, &newact, &oldact);

    newact.sa_handler = AtExit;
    sigemptyset(&newact.sa_mask);
    newact.sa_flags = 0;
    sigaction(SIGSEGV, &newact, &oldact);
   
    newact.sa_handler = AtExit;
    sigemptyset(&newact.sa_mask);
    newact.sa_flags = 0;
    sigaction(SIGQUIT, &newact, &oldact);
    
    newact.sa_handler = AtExit;
    sigemptyset(&newact.sa_mask);
    newact.sa_flags = 0;
    sigaction(SIGTERM, &newact, &oldact);
    
    newact.sa_handler = AtExit;
    sigemptyset(&newact.sa_mask);
    newact.sa_flags = 0;
    sigaction(SIGINT, &newact, &oldact);
}
#endif

// Shows usage
void Usage(char* ProgramName)
{
	printf("%s [-h|-f FILE_NAME]\n", ProgramName);
	printf("Options:\n");
	printf("h: Help. Shows this message\n");
	printf("f: Specifies log file name. Performance daemon will save all performance\n");
	printf("   monitoring message into the log file.\n");
}

void ProcessCommandLine(int argc, char *argv[])
{
    char optionStr[] = "f:h";
    char opt;
    char* optarg;

    opt = GetOption(argc, argv, optionStr, &optarg);
    while (opt != (char) -1 && opt != (char) 0) {
	switch (opt) {
            case 'f': 
                if (optarg == NULL) {
                    Usage(argv[0]);
                    exit(0);
                }	
		strcpy(LogFileName, optarg);
		break;

            case 'h':
            case (char)-1:
            case (char) 1:
		Usage(argv[0]);
		exit(0);
	}		
	opt = GetOption(argc, argv, optionStr, &optarg);
    }
}

int main(int argc, char* argv[])
{
    QUANTAnet_perfDaemon_c* quantaPerfDaemon;

    if (argc > 1) ProcessCommandLine(argc, argv);

    QUANTAinit();
    if (LogFileName[0] == '\0')
	quantaPerfDaemon = new QUANTAnet_perfDaemon_c;
    else {
        LogFile = fopen(LogFileName, "w");

        if (LogFile) {
            quantaPerfDaemon = new QUANTAnet_perfDaemon_c(LogFile);
            atexit(AtExit2);
#ifndef WIN32
            SetupSignals();
#endif
        }
 	else {
            printf("Cannot open file %s\n", LogFileName);
            exit(1);
        } 
    }

    // Initialize quanta reflector
    if (quantaPerfDaemon->init() != QUANTAnet_perfDaemon_c::OK) {
        printf("Cannot initialize QUANTA perfdaemon server\n");
        exit(1);
    }

    while(1) {
        // Perform data reflection services.
        quantaPerfDaemon->process();
    }
    return 1;
}





